#Dis.g
#DisFuncs.g

/*
 * dis.d - basic Motorola 68000/68020/68851/68881 disassembler library.
 *
 * March 29, 1989
 * Chris Gray
 */

/*
 * da_disOne - top level of disassembling one instruction.
 */

#A
proc da_disOne()void:
    [4] char MOVE_LENGTH = ('?', 'b', 'l', 'w');
    *char SHIFT_OPS = "asr\e asl\e lsr\e lsl\e roxr\eroxl\erol\e ror";
    *char BF_OPS =
	"bftst\e bfextu\ebfchg\e bfexts\ebfclr\e bfffo\e bfset\e bfins";
    register proc(char ch)void putChar;
    register proc(*char st)void putString;
    register proc(*char opString; uint opCode)void putOpModed;
    register uint opCode, ea, reg;
    register char len;

    putChar := da_putChar;
    putString := da_putString;
    putOpModed := da_putOpModed;
    DS*.ds_illegal := false;
    DS*.ds_hadExtraWord := false;
    opCode := da_readWord();
    ea := opCode & 0o77;
    reg := (opCode & 0o007000) >> 9;

    /* do a case on the top nibble of the opCode */

    case opCode >> 12
    incase 0x0:
	da_topNibble0(opCode);
    incase 0x1:
    incase 0x2:
    incase 0x3:
	putString(if opCode & 0o000700 = 0o000100 then
			 "movea"
		     else
			 "move"
		     fi);
	len := MOVE_LENGTH[opCode >> 12];
	da_putLen(len);
	da_putEA(ea, len);
	putChar(',');
	opCode := (opCode >> 6) & 0o77;
	da_putEA((opCode >> 3) | ((opCode & 0o7) << 3), len);
    incase 0x4:
	da_topNibble4(opCode);
    incase 0x5:
	if opCode & 0o000300 = 0o000300 then
	    reg := (opCode & 0x0f00) >> 8;
	    case ea
	    incase 0o10 .. 0o17:
		putString("db");
		da_putCondition(reg);
		da_tab();
		da_putDReg(opCode & 0o7);
		putChar(',');
		da_putBranch(0);
	    incase 0o72 .. 0o74:
		DS*.ds_extended := true;
		DS*.ds_extendedNow := true;
		putString("trap");
		da_putCondition(reg);
		case opCode & 0o7
		incase 0o2:
		    putString(".w\t\#");
		    da_putWordDec(da_readWord());
		incase 0o3:
		    putString(".l\t\#");
		    da_putLongDec(da_readLong());
		default:
		    DS*.ds_illegal := true;
		    putString("???");
		esac;
	    default:
		putChar('s');
		DS*.ds_column := DS*.ds_column + 1;
		da_putCondition(reg);
		da_tab();
		da_putEA(ea, 'b');
	    esac;
	else
	    putString(if opCode & 0o000400 = 0 then "addq" else "subq" fi);
	    len := LENGTHS[(opCode & 0o000300) >> 6];
	    da_putLen(len);
	    putChar('\#');
	    da_putWordDec(
		if opCode & 0o007000 = 0 then
		    8
		else
		    (opCode & 0o007000) >> 9
		fi);
	    putChar(',');
	    da_putEA(ea, len);
	fi;
    incase 0x6:
	if opCode & 0x0f00 = 0x0100 then
	    putString("bsr\t");
	    da_putBranch(make(opCode & 0xff, short));
	else
	    if opCode & 0x0f00 = 0 then
		putString("bra\t");
	    else
		putChar('b');
		DS*.ds_column := DS*.ds_column + 1;
		da_putCondition((opCode & 0x0f00) >> 8);
		da_tab();
	    fi;
	    da_putBranch(make(opCode & 0xff, short));
	fi;
    incase 0x7:
	putString("moveq.l\t\#");
	da_putWordDec(make(opCode & 0xff, short));
	putChar(',');
	da_putDReg((opCode & 0o007000) >> 9);
    incase 0x8:
	case (opCode & 0o000700) >> 6
	incase 0o3:
	    da_putOpReg("divu", 'w', opCode);
	incase 0o7:
	    da_putOpReg("divs", 'w', opCode);
	incase 0o4:
	    if opCode & 0o000060 = 0 then
		putString("sbcd.b\t");
		da_putOpMult(opCode);
	    else
		putOpModed("or", opCode);
	    fi;
	incase 0o5:
	incase 0o6:
	    if opCode & 0o000060 = 0 then
		DS*.ds_extended := true;
		DS*.ds_extendedNow := true;
		putString(
		    if opCode & 0o000700 = 0o000500 then
			"pack\t"
		    else
			"unpack\t"
		    fi);
		da_putOpMult(opCode);
		putString(",\#");
		da_putWordDec(da_readWord());
	    else
		putOpModed("or", opCode);
	    fi;
	default:
	    putOpModed("or", opCode);
	esac;
    incase 0x9:
	if opCode & 0o000460 = 0o000400 and opCode & 0o000300 ~= 0o000300 then
	    putString("subx");
	    da_putLen(LENGTHS[(opCode & 0o000300) >> 6]);
	    da_putOpMult(opCode);
	else
	    putOpModed(if opCode & 0o000300 = 0o000300 then
			      "suba"
			  else
			      "sub"
			  fi, opCode);
	fi;
    incase 0xa:
	putString("line A");
	DS*.ds_illegal := true;
    incase 0xb:
	case (opCode & 0o000700) >> 6
	incase 0o0:
	incase 0o1:
	incase 0o2:
	    putOpModed("cmp", opCode);
	incase 0o3:
	incase 0o7:
	    putOpModed("cmpa", opCode);
	incase 0o4:
	incase 0o5:
	incase 0o6:
	    if opCode & 0o000070 = 0o000010 then
		putString("cmpm");
		da_putLen(LENGTHS[(opCode & 0o000300) >> 6]);
		putChar('(');
		da_putAReg(opCode & 0o7);
		putString(")+,(");
		da_putAReg((opCode & 0o007000) >> 9);
		putString(")+");
	    else
		putOpModed("eor", opCode);
	    fi;
	esac;
    incase 0xc:
	case (opCode & 0o000700) >> 6
	incase 0o3:
	    da_putOpReg("mulu", 'w', opCode);
	incase 0o7:
	    da_putOpReg("muls", 'w', opCode);
	default:
	    if opCode & 0o000060 = 0 then
		case (opCode & 0o000700) >> 6
		incase 0o4:
		    putString("abcd.b\t");
		    da_putOpMult(opCode);
		incase 0o5:
		incase 0o6:
		    putString("exg.l\t");
		    if opCode & 0o000710 = 0o000510 then
			da_putAReg(reg);
		    else
			da_putDReg(reg);
		    fi;
		    putChar(',');
		    reg := opCode & 0o7;
		    if opCode & 0o000710 = 0o000500 then
			da_putAReg(reg);
		    else
			da_putDReg(reg);
		    fi;
		default:
		    putOpModed("and", opCode);
		esac;
	    else
		putOpModed("and", opCode);
	    fi;
	esac;
    incase 0xd:
	if opCode & 0o000460 = 0o000400 and opCode & 0o000300 ~= 0o000300 then
	    putString("addx");
	    da_putLen(LENGTHS[(opCode & 0o000300) >> 6]);
	    da_putOpMult(opCode);
	else
	    putOpModed(if opCode & 0o000300 = 0o000300 then
			      "adda"
			  else
			      "add"
			  fi, opCode);
	fi;
    incase 0xe:
	if opCode & 0o000300 = 0o000300 then
	    if opCode & 0o004000 = 0 then
		putString(SHIFT_OPS + ((opCode & 0o003400) >> 8) * 5);
		putString(".w\t");
		da_putEA(ea, 'w');
	    else
		DS*.ds_extended := true;
		DS*.ds_extendedNow := true;
		putString(BF_OPS + ((opCode & 0o003400) >> 8) * 7);
		da_tab();
		reg := da_readWord();
		if opCode & 0o003400 = 0o003400 then
		    da_putDReg((opCode & 0o007000) >> 12);
		    putChar(',');
		fi;
		da_putEA(ea, ' ');
		putChar('{');
		if reg & 0o004000 = 0 then
		    da_putWordDec((reg & 0o003700) >> 6);
		else
		    da_putDReg((reg & 0o000700) >> 6);
		fi;
		putChar(':');
		if reg & 0o000040 = 0 then
		    da_putWordDec(reg & 0o000037);
		else
		    da_putDReg(reg & 0o7);
		fi;
		putChar('}');
		if opCode & 0o003400 = 0o000400 or
		    opCode & 0o003400 = 0o001400 or
		    opCode & 0o003400 = 0o002400
		then
		    putChar(',');
		    da_putDReg((reg & 0o070000) >> 12);
		fi;
	    fi;
	else
	    putString(
		SHIFT_OPS + (((opCode & 0o000030) >> 2) +
			     ((opCode & 0o000400) >> 8)) * 5);
	    da_putLen(LENGTHS[(opCode & 0o000300) >> 6]);
	    if opCode & 0o000040 = 0 then
		putChar('\#');
		da_putWordDec(if reg = 0 then 8 else reg fi);
	    else
		da_putDReg(reg);
	    fi;
	    putChar(',');
	    da_putDReg(opCode & 0o000007);
	fi;
    incase 0xf:

	/* co-processor instructions */

	DS*.ds_extended := true;
	DS*.ds_extendedNow := true;
	case reg
	incase 0o0:
	    da_MMU(opCode);
	incase 0o1:
	    da_FPU(opCode);
	default:
	    putString("coprocessor \#");
	    da_putWordDec(reg);
	    DS*.ds_illegal := true;
	esac;
    esac;

    if DS*.ds_illegal then
	putString("  {OPCODE: ");
	da_putWordHex(opCode);
	if DS*.ds_hadExtraWord then
	    putChar(' ');
	    da_putWordHex(DS*.ds_extraWord);
	fi;
	putChar('}');
    fi;
corp;

/*
 * da_putPosition - output a current position leader.
 */

proc da_putPosition()void:

    if DS*.ds_putPosition then
	if DS*.ds_absoluteAddress then
	    da_putLongHex(DS*.ds_address);
	else
	    da_putString("   +");
	    da_putWordHex(DS*.ds_address - DS*.ds_relativeBase);
	fi;
	da_putChar(':');
	DS*.ds_column := DS*.ds_column + 1;
    fi;
corp;

/*
 * internalDisassemble - top level of disassembly. Take the next word, treat
 *	it as an opCode and go from there. Any output line or error message
 *	will go through ds_putChar. Return 'true' if no error was noted,
 *	else return false.
 */

proc internalDisassemble(*DisassemblerState_t ds)bool:
    uint
	R_D0 = 0,
	R_FP = 6,
	OP_MOVEL = 0x2000,
	M_DDIR = 0,
	M_DISP = 5;
    register ulong addr;
    register *char label;
    ulong address;
    bool doneLabel;

    DS := ds;
    DS*.ds_errorMessage := nil;
    DS*.ds_extended := false;
    DS*.ds_extendedNow := false;
    addr := DS*.ds_address;
    address := addr;
    if DS*.ds_findLabel ~= nil then
	code(
	    OP_MOVEL | R_D0 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
	    address
	);
	label := DS*.ds_findLabel(/* address */);
	if label ~= nil then
	    da_putPosition();
	    da_putString(label);
	    da_putChar('\n');
	    DS*.ds_relativeBase := addr;
	fi;
    else
	label := nil;
    fi;
    doneLabel := false;
    DS*.ds_column := 1;
    if label = nil and DS*.ds_isLabel ~= nil then
	code(
	    OP_MOVEL | R_D0 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
	    address
	);
	if DS*.ds_isLabel(/* address */) then
	    da_putPosition();
	    da_putChar('L');
	    DS*.ds_column := DS*.ds_column + 1;
	    if DS*.ds_absoluteAddress then
		da_putLongHexRaw(addr);
		da_putChar('\n');
		DS*.ds_column := 1;
	    else
		da_putWordHexRaw(addr - DS*.ds_relativeBase);
		da_putString("   ");
		doneLabel := true;
	    fi;
	fi;
    fi;
    if DS*.ds_putChar ~= nil and not doneLabel then
	da_putPosition();
	da_putString("        ");
    fi;

    /* go ahead and disassemble one instruction */

    da_disOne();
    da_putChar('\n');
    if DS*.ds_errorMessage ~= nil and DS*.ds_putErrors then
	da_putString("*** ");
	da_putString(DS*.ds_errorMessage);
	da_putString(" ***\n");
    fi;
    DS*.ds_errorMessage = nil and not DS*.ds_illegal
corp;
