#Dis.g
#DisFuncs.g

/*
 * Dis0_4.d - disassemble instructions whose top 4 bits are 0x0 or 0x4.
 *
 * April 2, 1989
 * Chris Gray
 */

/*
 * da_topNibble0 - break out the instructions for which the high nibble = 0
 */

#A
proc da_topNibble0(register uint opCode)void:
    *char BIT_OPS = "btst\ebchg\ebclr\ebset";
    *char IMM_OPS = "ori\e andi\esubi\eaddi\eXXXi\eeori\ecmpi";
    register proc(char ch)void putChar;
    register proc(*char st)void putString;
    register proc(uint reg)void putDReg;
    register uint reg, ea1, wrd;
    ulong base;
    char len;

    putChar := da_putChar;
    putString := da_putString;
    putDReg := da_putDReg;
    reg := (opCode & 0o007000) >> 9;
    ea1 := opCode & 0o77;
    len := LENGTHS[(opCode & 0o000300) >> 6];
    if opCode & 0o000400 ~= 0 then
	if opCode & 0o000070 = 0o000010 then

	    /* movep */

	    putString("movep");
	    da_putLen(if opCode & 0o000100 = 0 then 'w' else 'l' fi);
	    if opCode & 0o000200 = 0 then
		putDReg(reg);
		putChar(',');
		da_putWordDec(da_readWord());
		putChar('(');
		da_putAReg(opCode & 0o7);
		putChar(')');
	    else
		da_putWordDec(da_readWord());
		putChar('(');
		da_putAReg(opCode & 0o7);
		putChar(')');
		putChar(',');
		putDReg(reg);
	    fi;
	else

	    /* dynamic bit */

	    putString(BIT_OPS + ((opCode >> 6) & 0o3) * 5);
	    da_putLen(if opCode & 0o000070 = 0 then 'l' else 'b' fi);
	    putDReg(reg);
	    putChar(',');
	    da_putEA(ea1, 'l');
	fi;
    elif opCode & 0o077400 = 0o004000 then

	/* static bit */

	putString(BIT_OPS + ((opCode >> 6) & 0o3) * 5);
	len := if opCode & 0o000070 = 0 then 'l' else 'b' fi;
	da_putLen(len);
	putChar('\#');
	da_putWordHex(da_readWord());
	putChar(',');
	da_putEA(ea1, len);
    elif opCode & 0o000300 = 0o000300 then

	/* 68020 fancy stuff */

	DS*.ds_extended := true;
	DS*.ds_extendedNow := true;
	if opCode & 0o004000 = 0 then
	    if opCode & 0o003000 = 0o003000 then
		if opCode & 0o000060 = 0 then
		    putString("retm\t");
		    if opCode & 0o000010 = 0 then
			putDReg(opCode & 0o7);
		    else
			da_putAReg(opCode & 0o7);
		    fi;
		else
		    putString("callm\t\#");
		    da_putWordDec(make(da_readWord() & 0xff, short));
		    putChar(',');
		    da_putEA(ea1, 'l');
		fi;
	    else
		wrd := da_readWord();
		putString(
		    if wrd & 0o004000 = 0 then "cmp2" else "chk2" fi);
		len := LENGTHS[reg & 0o3];
		da_putLen(len);
		da_putEA(ea1, len);
		putChar(',');
		reg := (wrd >> 12) & 0o7;
		if wrd & 0o100000 = 0 then
		    putDReg(reg);
		else
		    da_putAReg(reg);
		fi;
	    fi;
	else
	    putString(if ea1 = 0o74 then "cas2" else "cas" fi);
	    da_putLen(if reg & 1 = 0 then 'w' else 'l' fi);
	    wrd := da_readWord();
	    if ea1 = 0o74 then
		/* cas2 */
		opCode := da_readWord();
		putDReg(wrd & 0o7);
		putChar(':');
		putDReg(opCode & 0o7);
		putChar(',');
		putDReg((wrd & 0o000700) >> 6);
		putChar(':');
		putDReg((opCode & 0o000700) >> 6);
		putString(",(");
		reg := (wrd & 0o070000) >> 12;
		if wrd & 0o100000 = 0 then
		    putDReg(reg);
		else
		    da_putAReg(reg);
		fi;
		putString("):(");
		reg := (opCode & 0o070000) >> 12;
		if opCode & 0o100000 = 0 then
		    putDReg(reg);
		else
		    da_putAReg(reg);
		fi;
		putChar(')');
	    else
		/* cas */
		putDReg(wrd & 0o7);
		putChar(',');
		putDReg((wrd & 0o000700) >> 6);
		putChar(',');
		da_putEA(ea1, len);
	    fi;
	fi;
    elif opCode & 0o007400 = 0o007000 then

	/* moves */

	DS*.ds_extended := true;
	DS*.ds_extendedNow := true;
	putString("moves");
	da_putLen(len);
	wrd := da_readWord();
	reg := (wrd & 0o070000) >> 12;
	if wrd & 0o002000 ~= 0 then
	    if wrd & 0100000 = 0 then
		putDReg(reg);
	    else
		da_putAReg(reg);
	    fi;
	    putChar(',');
	    da_putEA(ea1, len);
	else
	    da_putEA(ea1, len);
	    putChar(',');
	    if wrd & 0100000 = 0 then
		putDReg(reg);
	    else
		da_putAReg(reg);
	    fi;
	fi;
    else

	/* immediate mode instruction */

	putString(IMM_OPS + ((opCode & 0o007000) >> 9) * 5);
	da_putLen(len);
	putChar('\#');
	if len = 'l' then
	    base := DS*.ds_address;
	    da_putAbsSymbol(da_readLong(), base, 'l', false);
	else
	    wrd := da_readWord();
	    if len = 'b' then
		da_putWordDec(make(wrd & 0xff, short));
	    else
		da_putWordDec(wrd);
	    fi;
	fi;
	putChar(',');
	if ea1 = 0o74 then
	    putString(if len = 'b' then "ccr" else "sr" fi);
	else
	    da_putEA(ea1, len);
	fi;
    fi;
corp;

/*
 * da_topNibble4 - break out the instructions for which the high nibble = 4
 */

#A
proc da_topNibble4(register uint opCode)void:
    *char SIMPLE_OPS =
	"reset\enop  \estop \erte  \ertd  \erts  \etrapv\ertr  ";
    register proc(char ch)void putChar;
    register proc(*char st)void putString;
    register proc(uint ea; char size)void putEA;
    register uint reg, ea;
    register char len;
    uint reg2;

    putChar := da_putChar;
    putString := da_putString;
    putEA := da_putEA;
    reg := (opCode & 0o007000) >> 9;
    ea := opCode & 0o77;
    len := LENGTHS[(opCode & 0o000300) >> 6];
    if opCode & 0o000400 = 0 then
	case reg
	incase 0o0:
	    if len = '?' then
		putString("move.w\tsr,");
		putEA(ea, 'w');
	    else
		da_putOpSingle("negx", len, ea);
	    fi;
	incase 0o1:
	    if len = '?' then
		DS*.ds_extended := true;
		DS*.ds_extendedNow := true;
		putString("move.w\tccr,");
		putEA(ea, 'w');
	    else
		da_putOpSingle("clr", len, ea);
	    fi;
	incase 0o2:
	    if len = '?' then
		putString("move.w\t");
		putEA(ea, 'w');
		putString(",ccr");
	    else
		da_putOpSingle("neg", len, ea);
	    fi;
	incase 0o3:
	    if len = '?' then
		putString("move.w\t");
		putEA(ea, 'w');
		putString(",sr");
	    else
		da_putOpSingle("not", len, ea);
	    fi;
	incase 0o4:
	    if len = 'b' then
		if ea & 0o70 = 0o10 then
		    DS*.ds_extended := true;
		    DS*.ds_extendedNow := true;
		    putString("link.l\t");
		    da_putAReg(ea & 0o7);
		    putString(",\#");
		    da_putLongDec(da_readLong());
		else
		    putString("nbcd\t");
		    putEA(ea, 'b');
		fi;
	    elif len = 'w' then
		if ea & 0o70 = 0o00 then
		    putString("swap.w\t");
		    da_putDReg(ea & 0o7);
		elif ea & 0o70 = 0o10 then
		    DS*.ds_extended := true;
		    DS*.ds_extendedNow := true;
		    putString("bkpt\t\#");
		    da_putWordDec(opCode & 0xf);
		else
		    putString("pea\t");
		    putEA(ea, 'l');
		fi;
	    elif ea & 0o70 = 0o00 then
		if len ~= '?' and len ~= 'l' then
		    DS*.ds_extended := true;
		    DS*.ds_extendedNow := true;
		fi;
		putString(if len = '?' then "extw" else "extb" fi);
		da_putLen(if len = 'l' then 'w' else 'l' fi);
		da_putDReg(ea & 0o7);
	    else
		putString("movem");
		da_putLen(if len = 'l' then 'w' else 'l' fi);
		da_putRegList(da_readWord(), ea & 0o70 ~= 0o40);
		putChar(',');
		putEA(ea, 'l');
	    fi;
	incase 0o5:
	    if len = '?' then
		if ea = 0o74 then
		    putString("illegal");
		else
		    putString("tas\t");
		    putEA(ea, 'b');
		fi;
	    else
		da_putOpSingle("tst", len, ea);
	    fi;
	incase 0o6:
	    case (opCode & 0o000300) >> 6
	    incase 0o0:
	    incase 0o1:
		DS*.ds_extended := true;
		DS*.ds_extendedNow := true;
		reg := da_readWord();
		if opCode & 0o000700 = 0 then
		    putString("mul");
		    putChar(
			if reg & 0o004000 = 0 then
			    if DS*.ds_capExtended then 'U' else 'u' fi
			else
			    if DS*.ds_capExtended then 'S' else 's' fi
			fi);
		    DS*.ds_column := DS*.ds_column + 1;
		    len := if reg & 0o002000 = 0 then 'l' else 'q' fi;
		else
		    len := 'l';
		    putString("div");
		    putChar(
			if reg & 0o004000 = 0 then
			    if DS*.ds_capExtended then 'U' else 'u' fi
			else
			    if DS*.ds_capExtended then 'S' else 's' fi
			fi);
		    DS*.ds_column := DS*.ds_column + 1;
		    if reg & 0o002000 ~= 0 then
			putChar(
			    if DS*.ds_capExtended then 'L' else 'l' fi);
			DS*.ds_column := DS*.ds_column + 1;
		    fi;
		fi;
		da_putLen(len);
		putEA(ea, 'l');
		putChar(',');
		if opCode & 0o000700 = 0 then
		    if reg & 0o002000 ~= 0 then
			da_putDReg(reg & 0o7);
			putChar(':');
		    fi;
		    da_putDReg((reg & 0o070000) >> 12);
		else
		    da_putDReg(reg & 0o7);
		    if (reg & 0o070000) >> 12 ~= reg & 0o7 then
			putChar(':');
			da_putDReg((reg & 0o070000) >> 12);
		    fi;
		fi;
	    incase 0o2:
	    incase 0o3:
		putString("movem");
		da_putLen(if len = 'l' then 'w' else 'l' fi);
		reg := da_readWord();
		putEA(ea, 'l');
		putChar(',');
		da_putRegList(reg, ea & 0o70 ~= 0o40);
	    esac;
	incase 0o7:
	    case (opCode & 0o000300) >> 6
	    incase 0o1:
		if opCode & 0o000060 = 0o000000 then
		    putString("trap\t\#");
		    da_putWordDec(opCode & 0xf);
		else
		    reg := ea & 0o7;
		    case ea >> 3
		    incase 0o2:
			putString("link.w\t");
			da_putAReg(reg);
			putString(",\#");
			da_putWordDec(da_readWord());
		    incase 0o3:
			putString("unlk\t");
			da_putAReg(reg);
		    incase 0o4:
			putString("move.l\t");
			da_putAReg(reg);
			putString(",usp");
		    incase 0o5:
			putString("move.l\tusp,");
			da_putAReg(reg);
		    incase 0o6:
			putString(SIMPLE_OPS + reg * 6);
			if reg = 0o2 then
			    putString("\t\#");
			    da_putWordHex(da_readWord());
			elif reg = 0o4 then
			    DS*.ds_extended := true;
			    DS*.ds_extendedNow := true;
			    putString("\t\#");
			    da_putWordDec(da_readWord());
			fi;
		    incase 0o7:
			DS*.ds_extended := true;
			DS*.ds_extendedNow := true;
			reg := da_readWord();
			putString("movec.l\t");
			reg2 := (reg & 0o070000) >> 12;
			if opCode & 0o000001 ~= 0 then
			    if reg & 0o100000 = 0 then
				da_putDReg(reg2);
			    else
				da_putAReg(reg2);
			    fi;
			    putChar(',');
			fi;
			putString(
			    case reg & 0x0fff
			    incase 0x000:
				"sfc"
			    incase 0x001:
				"dfc"
			    incase 0x002:
				"cacr"
			    incase 0x800:
				"usp"
			    incase 0x801:
				"vbr"
			    incase 0x802:
				"caar"
			    incase 0x803:
				"msp"
			    incase 0x804:
				"isp"
			    default:
				DS*.ds_illegal := true;
				DS*.ds_hadExtraWord := true;
				DS*.ds_extraWord := reg;
				"???"
			    esac);
			if opCode & 0o000001 = 0 then
			    putChar(',');
			    if reg & 0o100000 = 0 then
				da_putDReg(reg2);
			    else
				da_putAReg(reg2);
			    fi;
			fi;
		    esac;
		fi;
	    incase 0o2:
		putString("jsr\t");
		putEA(ea, 'c');
	    incase 0o3:
		putString("jmp\t");
		putEA(ea, 'c');
	    default:
		DS*.ds_illegal := true;
	    esac;
	esac;
    else
	if opCode & 0o000100 = 0 then
	    putString("chk\t");
	    putEA(ea, 'w');
	    putChar(',');
	    da_putDReg(reg);
	elif opCode & 0o007770 = 0o004700 then
	    DS*.ds_extended := true;
	    DS*.ds_extendedNow := true;
	    putString("extb.l\t");
	    da_putDReg(ea & 0o7);
	else
	    putString("lea\t");
	    putEA(ea, 'l');
	    putChar(',');
	    da_putAReg(reg);
	fi;
    fi;
corp;
