#Dis.g
#DisFuncs.g

/*
 * DisFPU.d - disassemble the FPU coprocessor instructions.
 *
 * April 2, 1989
 * Chris Gray
 */

/*
 * da_putFPReg - output the name of an FPU register.
 */

proc da_putFPReg(uint reg)void:

    da_putString("fp");
    da_putChar(reg + '0');
corp;

/*
 * da_putFPRegList - put out a register list for fmovem.
 */

proc da_putFPRegList(register uint w; bool goingUp)void:
    register uint mask, i;
    register bool needSlash;

    da_putChar('<');
    needSlash := false;
    mask := 0x80;
    for i from 0 upto 7 do
	if w & mask ~= 0 then
	    if needSlash then
		da_putChar('/');
	    fi;
	    if goingUp then
		da_putFPReg(7 - i);
	    else
		da_putFPReg(i);
	    fi;
	    needSlash := true;
	fi;
	mask := mask >> 1;
    od;
    da_putChar('>');
corp;

/*
 * da_putFPCntlReg - output the name of an FPU control register.
 */

proc da_putFPCntlReg(register uint mask)void:
    *char CNTL_NAMES = "fpcr\e fpsr\e fpiar";
    register uint i;
    register bool needSlash;

    needSlash := false;
    for i from 2 downto 0 do
	if mask & 0o1 ~= 0 then
	    if needSlash then
		da_putChar('/');
	    fi;
	    da_putString(CNTL_NAMES + i * 6);
	    needSlash := true;
	fi;
	mask := mask >> 1;
    od;
corp;

/*
 * da_putFPUCondition - output an FPU condition code.
 */

proc da_putFPUCondition(uint cond)void:
    *char FPU_COND =
	"f\e   eq\e  ogt\e oge\e olt\e ole\e ogl\e or\e  "
	"un\e  ueq\e ugt\e uge\e ult\e ule\e ne\e  t\e   "
	"sf\e  seq\e gt\e  ge\e  lt\e  le\e  gl\e  gle\e "
	"ngle\engl\e nle\e nlt\e nge\e ngt\e sne\e st";

    da_putString(FPU_COND + (cond & 0o37) * 5);
corp;

/*
 * da_FPU - top level of MMU disassembly.
 */

#A
proc da_FPU(uint opCode)void:
    [8] char FPU_LEN = ('l','s','x','p','w','d','b','k');
    *char FLOAT_OPS =
	"fmove\e  fint\e   fsinh\e  fintrz\e "
	"fsqrt\e  f???\e   flognp1\efetoxm1\e"
	"fetoxm1\eftanh\e  fatan\e  f???\e   "
	"fasin\e  fatanh\e fsin\e   ftan\e   "
	"fetox\e  ftwotox\eftentox\ef???\e   "
	"flogn\e  flog10\e flog2\e  f???\e   "
	"fabs\e   fcosh\e  fneg\e   f???\e   "
	"facos\e  fcos\e   fgetexp\efgetman\e"
	"fdiv\e   fmod\e   fadd\e   fmul\e   "
	"fsgldiv\efrem\e   fscale\e fsglmul\e"
	"fsub\e   f???\e   f???\e   f???\e   "
	"f???\e   f???\e   f???\e   f???\e   ";
    register proc(char ch)void putChar;
    register proc(*char st)void putString;
    register proc(uint reg)void putFPReg;
    register uint ea, wrd, fReg, fReg2, r_m;
    char len;
    bool flag;

    putChar := da_putChar;
    putString := da_putString;
    putFPReg := da_putFPReg;
    ea := opCode & 0o77;
    case (opCode & 0o000700) >> 6
    incase 0o0:
	wrd := da_readWord();
	fReg := (wrd & 0o001600) >> 7;
	case (wrd & 0o160000) >> 13
	incase 0o0:
	incase 0o2:
	    if wrd & 0o056000 = 0o056000 then
		putString("fmovecr.x\t");
		putChar('\#');
		da_putWordHex(wrd & 0o000177);
		putChar(',');
		putFPReg(fReg);
	    else
		r_m := wrd & 0o040000;
		fReg2 := (wrd & 0o016000) >> 10;
		len := if r_m ~= 0 then FPU_LEN[fReg2] else 'x' fi;
		case (wrd & 0o060) >> 4
		incase 0o0:
		incase 0o1:
		    /* unary operation */
		    putString(FLOAT_OPS + (wrd & 0o77) * 8);
		    da_putLen(len);
		    if r_m = 0 then
			/* source specifier is source float reg */
			if fReg2 ~= fReg and wrd & 0o17 ~= 0 then
			    putFPReg(fReg2);
			    putChar(',');
			fi;
		    else
			/* source specifier is len */
			da_putEA(ea, len);
			putChar(',');
		    fi;
		    putFPReg(fReg);
		incase 0o2:
		    /* binary operations */
		    putString(FLOAT_OPS + (wrd & 0o77) * 8);
		    da_putLen(len);
		    if r_m = 0 then
			putFPReg(fReg2);
		    else
			da_putEA(ea, len);
		    fi;
		    putChar(',');
		    putFPReg(fReg);
		incase 0o3:
		    /* miscellaneous */
		    if wrd & 0o000010 = 0 then
			putString("fsincos");
			da_putLen(len);
			if r_m = 0 then
			    putFPReg(fReg2);
			else
			    da_putEA(ea, len);
			fi;
			putChar(',');
			putFPReg(wrd & 0o7);
			putChar(':');
			putFPReg(fReg);
		    else
			case wrd & 0o7
			incase 0o0:
			    putString("fcmp");
			    da_putLen(len);
			    if r_m = 0 then
				putFPReg(fReg2);
			    else
				da_putEA(ea, len);
			    fi;
			    putChar(',');
			    putFPReg(fReg);
			incase 0o2:
			    putString("ftst");
			    da_putLen(len);
			    if r_m = 0 then
				putFPReg(fReg2);
			    else
				da_putEA(ea, len);
			    fi;
			default:
			    putString("f???");
			    DS*.ds_illegal := true;
			    DS*.ds_hadExtraWord := true;
			    DS*.ds_extraWord := wrd;
			esac;
		    fi;
		esac;
	    fi;
	incase 0o3:
	    /* FMOVE from FPn */
	    len := FPU_LEN[(wrd & 0o016000) >> 10];
	    putString("fmove");
	    da_putLen(if len = 'k' then 'p' else len fi);
	    putFPReg((wrd & 0o001600) >> 7);
	    putChar(',');
	    da_putEA(ea, len);
	    if len = 'k' then
		putChar('{');
		da_putDReg((wrd & 0o160) >> 4);
		putChar('}');
	    elif len = 'p' then
		putChar('{');
		da_putWordDec(wrd & 0o177);
		putChar('}');
	    fi;
	incase 0o4:
	incase 0o5:
	    if wrd & 0o001777 ~= 0 then
		putString("f???");
		DS*.ds_illegal := true;
		DS*.ds_hadExtraWord := true;
		DS*.ds_extraWord := wrd;
	    else
		putString("fmove.l\t");
		fReg := (wrd & 0o016000) >> 10;
		if wrd & 0o020000 = 0 then
		    da_putEA(ea, 'l');
		    putChar(',');
		    da_putFPCntlReg(fReg);
		else
		    da_putFPCntlReg(fReg);
		    putChar(',');
		    da_putEA(ea, 'l');
		fi;
	    fi;
	incase 0o6:
	incase 0o7:
	    if wrd & 0o001400 ~= 0 then
		putString("f???");
		DS*.ds_illegal := true;
		DS*.ds_hadExtraWord := true;
		DS*.ds_extraWord := wrd;
	    else
		putString("fmovem");
		if wrd & 0o377 = 0 then
		    da_putLen('l');
		    if wrd & 0o020000 = 0 then
			/* memory to FPCP */
			da_putEA(ea, 'l');
			putChar(',');
			da_putFPCntlReg((wrd & 0o016000) >> 10);
		    else
			/* FPCP to memory */
			da_putFPCntlReg((wrd & 0o016000) >> 10);
			putChar(',');
			da_putEA(ea, 'l');
		    fi;
		else
		    da_putLen('x');
		    if wrd & 0o020000 = 0 then
			/* memory to FPCP */
			da_putEA(ea, 'x');
			putChar(',');
			if wrd & 0o004000 = 0 then
			    da_putFPRegList(wrd & 0o377,
					    wrd & 0o010000 ~= 0);
			else
			    da_putDReg((wrd & 0o000160) >> 4);
			fi;
		    else
			/* FPCP to memory */
			if wrd & 0o004000 = 0 then
			    da_putFPRegList(wrd & 0o377,
					    wrd & 0o010000 ~= 0);
			else
			    da_putDReg((wrd & 0o000160) >> 4);
			fi;
			putChar(',');
			da_putEA(ea, 'x');
		    fi;
		fi;
	    fi;
	esac;
    incase 0o1:
	wrd := da_readWord();
	if ea & 0o70 = 0o10 then
	    putString("fdb");
	    da_putFPUCondition(wrd & 0o77);
	    da_tab();
	    da_putDReg(ea & 0o7);
	    putChar(',');
	    da_putBranch(0);
	elif ea & 0o70 = 0o70 and (ea & 0o07 >= 0o02 and ea & 0o07 <= 0o04)
	then
	    putString("ftrap");
	    da_putFPUCondition(wrd & 0o77);
	    len := ("wlX" + (ea & 0o7) - 2)*;
	    if len ~= 'X' then
		da_putLen(len);
		putChar('\#');
		if len = 'w' then
		    da_putWordDec(da_readWord());
		else
		    da_putLongDec(da_readLong());
		fi;
	    fi;
	else
	    putString("fs");
	    da_putFPUCondition(wrd & 0o77);
	    da_putLen('b');
	    da_putEA(ea, 'b');
	fi;
    incase 0o2:
    incase 0o3:
	flag := true;
	if opCode & 0o000100 = 0 then
	    if opCode & 0o000077 = 0 then
		wrd := da_readWord();
		if wrd = 0 then
		    putString("fnop  ");
		    flag := false;
		fi;
	    fi;
	fi;
	if flag then
	    putString("fb");
	    da_putFPUCondition(ea);
	    da_tab();
	    if opCode & 0o000100 = 0 then
		da_putBranch(0);
	    else
		da_putBranch(-1);
	    fi;
	fi;
    incase 0o4:
	putString("fsave\t");
	da_putEA(ea, 'l');
    incase 0o5:
	putString("frestore\t");
	da_putEA(ea, 'l');
    default:
	putString("f???");
	DS*.ds_illegal := true;
    esac;
corp;
