#Dis.g
#DisFuncs.g

/*
 * DisMMU.d - disassemble the MMU coprocessor instructions.
 *
 * April 2, 1989
 * Chris Gray
 */

/*
 * da_putFunctionCode - output an MMU function code.
 */

proc da_putFunctionCode(uint fc)void:

    if fc & 0x10 ~= 0x00 then
	da_putChar('\#');
	da_putChar(fc & 0o7 + '0');
    elif fc & 0x08 ~= 0x00 then
	da_putDReg(fc & 0o7);
    elif fc = 0x00 then
	da_putString("sfc");
    elif fc = 0x01 then
	da_putString("dfc");
    else
	da_putString("fc???");
    fi;
corp;

/*
 * da_putMMUCondition - output an MMU condition code.
 */

proc da_putMMUCondition(uint cond)void:

    if DS*.ds_capExtended then
	da_putChar(("BLSAWIGC" + (cond & 0o16) >> 1)*);
	da_putChar(if cond & 1 = 0 then 'S' else 'C' fi);
    else
	da_putChar(("blsawigc" + (cond & 0o16) >> 1)*);
	da_putChar(if cond & 1 = 0 then 's' else 'c' fi);
    fi;
    DS*.ds_column := DS*.ds_column + 2;
corp;

/*
 * da_putMMUReg - output the name of one of the 16 bit MMU registers.
 */

proc da_putMMUReg(uint mReg, data)void:
    *char REG_NAMES = "psr\e pcsr\eXXXX\eXXXX\ebad\e bac\e XXXX\eXXXX";

    da_putString(REG_NAMES + mReg * 5);
    if mReg & 0o40 ~= 0 then
	da_putChar(data + '0');
    fi;
corp;

/*
 * da_MMU - top level of MMU disassembly.
 */

#A
proc da_MMU(register uint opCode)void:
    *char MMU_REGS = "tc\e drp\esrp\ecrp\ecal\eval\escc\eac";
    register proc(char ch)void putChar;
    register proc(*char st)void putString;
    register proc(uint ea; char len)void putEA;
    register uint wrd, ea, mReg;
    char len;

    putChar := da_putChar;
    putString := da_putString;
    putEA := da_putEA;
    ea := opCode & 0o77;
    case (opCode & 0o000700) >> 6
    incase 0o0:
	wrd := da_readWord();
	mReg := (wrd & 0o016000) >> 10;
	case (wrd & 0o160000) >> 13
	incase 0o1:
	    case mReg
	    incase 0o0:
		putString("pload");
		putChar(
		    if wrd & 0o001000 = 0 then
			if DS*.ds_capExtended then 'W' else 'w' fi
		    else
			if DS*.ds_capExtended then 'R' else 'r' fi
		    fi);
		DS*.ds_column := DS*.ds_column + 1;
		da_tab();
		da_putFunctionCode(wrd & 0o000037);
		putChar(',');
		putEA(ea, 'l');
	    incase 0o1:
		if wrd & 0o000740 = 0o000040 then
		    putString("pflusha");
		else
		    putString("p???");
		    DS*.ds_illegal := true;
		    DS*.ds_hadExtraWord := true;
		    DS*.ds_extraWord := wrd;
		fi;
	    incase 0o2:
		putString("pvalid\tval,");
		putEA(ea, 'l');
	    incase 0o3:
		putString("pvalid\t");
		da_putAReg(wrd & 0o000007);
		putChar(',');
		putEA(ea, 'l');
	    incase 0o4 .. 0o7:
		putString("pflush");
		if wrd & 0o002000 = 0 then
		    putChar(if DS*.ds_capExtended then 'S' else 's' fi);
		fi;
		DS*.ds_column := DS*.ds_column + 1;
		da_tab();
		da_putFunctionCode(wrd & 0o000037);
		putString(",\#");
		da_putWordHex((wrd & 0o000740) >> 5);
		if wrd & 0o004000 = 0 then
		    putChar(',');
		    putEA(ea, 'l');
		fi;
	    default:
		putString("p???");
		DS*.ds_illegal := true;
		DS*.ds_hadExtraWord := true;
		DS*.ds_extraWord := wrd;
	    esac;
	incase 0o2:
	    putString("pmove");
	    len := ("lqqqbbbw" + mReg)*;
	    da_putLen(len);
	    if wrd & 0o001000 = 0 then
		putEA(ea, len);
		putChar(',');
		putString(MMU_REGS + mReg * 4);
	    else
		putString(MMU_REGS + mReg * 4);
		putChar(',');
		putEA(ea, len);
	    fi;
	incase 0o3:
	    putString("pmove.w\t");
	    if wrd & 0o001000 = 0 then
		putEA(ea, 'w');
		putChar(',');
		da_putMMUReg(mReg, wrd & 0o000007);
	    else
		da_putMMUReg(mReg, wrd & 0o000007);
		putChar(',');
		putEA(ea, 'w');
	    fi;
	incase 0o4:
	    putString("ptest");
	    putChar(
		if wrd & 0o001000 = 0 then
		    if DS*.ds_capExtended then 'W' else 'w' fi
		else
		    if DS*.ds_capExtended then 'R' else 'r' fi
		fi);
	    DS*.ds_column := DS*.ds_column + 1;
	    da_tab();
	    da_putFunctionCode(wrd & 0o000037);
	    putChar(',');
	    putEA(ea, 'l');
	    putString(",\#");
	    da_putWordDec(mReg);
	    if wrd & 0o000400 ~= 0 then
		putChar(',');
		da_putAReg((wrd & 0o000340) >> 5);
	    fi;
	incase 0o5:
	    putString("pflush\t");
	    putEA(ea, 'l');
	default:
	    putString("p???");
	    DS*.ds_illegal := true;
	    DS*.ds_hadExtraWord := true;
	    DS*.ds_extraWord := wrd;
	esac;
    incase 0o1:
	wrd := da_readWord() & 0o000077;
	case (opCode & 0o000070) >> 3
	incase 0o1:
	    putString("pdb");
	    da_putMMUCondition(wrd);
	    da_putLen('w');
	    da_putDReg(ea & 0o7);
	    putChar(',');
	    da_putBranch(0);
	incase 0o7:
	    putString("ptrap");
	    da_putMMUCondition(wrd);
	    if opCode & 0o000007 = 0o000002 then
		putString(".w\t\#");
		da_putWordDec(da_readWord());
	    elif opCode & 0o000007 = 0o000003 then
		putString(".l\t\#");
		da_putLongDec(da_readLong());
	    fi;
	default:
	    putString("ps");
	    da_putMMUCondition(wrd);
	    da_putLen('b');
	    putEA(ea, 'b');
	esac;
    incase 0o2:
    incase 0o3:
	putString("pb");
	da_putMMUCondition(ea);
	putString("\t");
	if wrd & 0o000100 = 0 then
	    da_putBranch(0);
	else
	    da_putBranch(-1);
	fi;
    incase 0o4:
	putString("psave\t");
	putEA(ea, 'l');
    incase 0o5:
	putString("prestore\t");
	putEA(ea, 'l');
    default:
	putString("p???");
	DS*.ds_illegal := true;
    esac;
corp;
