#Dis.g
#DisFuncs.g

/*
 * DisUtil.d - utility routines for Motorola 68000/68020/68851/68881
 *	disassembler library.
 *
 * April 1, 1989
 * Chris Gray
 */

uint
    R_D0 = 0,
    R_D1 = 1,
    R_A0 = 0,
    R_FP = 6,

    OP_MOVEB = 0x1000,
    OP_MOVEL = 0x2000,

    M_DDIR = 0,
    M_ADIR = 1,
    M_INDIR = 2,
    M_INC = 3,
    M_DEC = 4,
    M_DISP = 5,
    M_INDEX = 6,
    M_SPECIAL = 7,

    M_ABSSHORT = 0,
    M_ABSLONG = 1,
    M_PCDISP = 2,
    M_PCINDEX = 3,
    M_IMM = 4,

    R_PC = 8;

/*
 * da_readWord - read a word of code from the given "address".
 */

proc da_readWord()uint:
    ulong address;
    uint theWord;

    address := DS*.ds_address;
    code(
	OP_MOVEL | R_D0 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
	address
    );
    theWord := DS*.ds_readWord(/* address */);
    DS*.ds_address := DS*.ds_address + sizeof(uint);
    theWord
corp;

/*
 * da_readLong - read a long of code from the given "address".
 */

proc da_readLong()ulong:

    make(da_readWord(), ulong) << 16 | da_readWord()
corp;	

/*
 * da_putChar - interface as necessary to the character output routine.
 */

proc da_putChar(char ch)void:

    if DS*.ds_putChar ~= nil then
	code(
	    OP_MOVEB | R_D0 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
	    ch
	);
	DS*.ds_putChar(/* ch */);
    fi;
corp;

/*
 * da_tab - space out to the target operand column.
 */

proc da_tab()void:

    if DS*.ds_putChar ~= nil then
	while DS*.ds_column < DS*.ds_operandColumn do
	    da_putChar(' ');
	    DS*.ds_column := DS*.ds_column + 1;
	od;
    fi;
corp;

/*
 * da_putString - output a string through the character output routine.
 */

proc da_putString(register *char p)void:

    if DS*.ds_putChar ~= nil then
	if DS*.ds_capExtended and DS*.ds_extendedNow then
	    while p* ~= '\e' do
		if p* >= 'a' and p* <= 'z' then
		    da_putChar(p* - ('a' - 'A'));
		    DS*.ds_column := DS*.ds_column + 1;
		elif p* = '\t' then
		    da_tab();
		else
		    da_putChar(p*);
		    DS*.ds_column := DS*.ds_column + 1;
		fi;
		p := p + sizeof(char);
	    od;
	else
	    while p* ~= '\e' do
		if p* = '\t' then
		    da_tab();
		else
		    da_putChar(p*);
		    DS*.ds_column := DS*.ds_column + 1;
		fi;
		p := p + sizeof(char);
	    od;
	fi;
    fi;
corp;

/*
 * da_putLongHexRaw - output an 8 digit hex value without the '0x'.
 */

proc da_putLongHexRaw(register ulong n)void:
    register *char p;
    register uint i;
    register ushort digit;
    [9] char buffer;

    if DS*.ds_putChar ~= nil then
	p := &buffer[8];
	p* := '\e';
	for i from 7 downto 0 do
	    digit := n & 0xf;
	    p := p - sizeof(char);
	    p* := if digit >= 10 then digit + ('a' - 10) else digit + '0' fi;
	    n := n >> 4;
	od;
	da_putString(&buffer[0]);
    fi;
corp;

/*
 * da_putLongHex - output an 8 digit hex value.
 */

proc da_putLongHex(ulong n)void:

    if DS*.ds_putChar ~= nil then
	da_putString("0x");
	da_putLongHexRaw(n);
    fi;
corp;

/*
 * da_putWordHexRaw - output a 4 digit hex value without the '0x'.
 */

proc da_putWordHexRaw(register uint n)void:
    register *char p;
    register uint i;
    register ushort digit;
    [5] char buffer;

    if DS*.ds_putChar ~= nil then
	p := &buffer[4];
	p* := '\e';
	for i from 3 downto 0 do
	    digit := n & 0xf;
	    p := p - sizeof(char);
	    p* := if digit >= 10 then digit + ('a' - 10) else digit + '0' fi;
	    n := n >> 4;
	od;
	da_putString(&buffer[0]);
    fi;
corp;

/*
 * da_putWordHex - output a 4 digit hex value.
 */

proc da_putWordHex(uint n)void:

    if DS*.ds_putChar ~= nil then
	da_putString("0x");
	da_putWordHexRaw(n);
    fi;
corp;

/*
 * da_putLongDec - output a signed decimal 32 bit value.
 */

proc da_putLongDec(register long n)void:
    register *char p;
    [11] char buffer;

    if DS*.ds_putChar ~= nil then
	if n < 0 then
	    da_putChar('-');
	else
	    n := -n;
	fi;
	p := &buffer[10];
	p* := '\e';
	while
	    p := p - sizeof(char);
	    p* := - (n % 10) + '0';
	    n := n / 10;
	    n ~= 0
	do
	od;
	da_putString(p);
    fi;
corp;

/*
 * da_putWordDec - output a signed decimal 16 bit value.
 */

proc da_putWordDec(register int n)void:
    register *char p;
    [6] char buffer;

    if DS*.ds_putChar ~= nil then
	if n < 0 then
	    da_putChar('-');
	else
	    n := -n;
	fi;
	p := &buffer[5];
	p* := '\e';
	while
	    p := p - sizeof(char);
	    p* := - (n % 10) + '0';
	    n := n / 10;
	    n ~= 0
	do
	od;
	da_putString(p);
    fi;
corp;

/*
 * da_putLen - output the '.', a len character and a tab.
 */

proc da_putLen(char len)void:

    if DS*.ds_putChar ~= nil then
	da_putChar('.');
	if DS*.ds_capExtended and DS*.ds_extendedNow and len ~= '?' then
	    da_putChar(len - ('a' - 'A'));
	else
	    da_putChar(len);
	fi;
	DS*.ds_column := DS*.ds_column + 2;
	da_tab();
    fi;
corp;

/*
 * da_putAbsSymbol - put out a symbolic form of an absolute address if one
 *	is available. Use relative to it if possible.
 */

proc da_putAbsSymbol(register ulong addr; ulong refAt;
		     char len; bool isBranch)void:
    ulong trueAddr;
    register *char symbol;
    *ulong pTrueAddr;

    if DS*.ds_findAbsSymbol ~= nil then
	trueAddr := addr;
	if isBranch then
	    refAt := refAt | 0x80000000;
	fi;
	pTrueAddr := &trueAddr;
	code(
	    OP_MOVEL | R_D0 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
	    trueAddr,
	    OP_MOVEL | R_D1 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
	    refAt,
	    OP_MOVEL | R_A0 << 9 | M_ADIR << 6 | M_DISP << 3 | R_FP,
	    pTrueAddr
	);
	symbol := DS*.ds_findAbsSymbol(/* addr, refAt, &trueAddr */);
	if symbol ~= nil then
	    da_putString(symbol);
	    if trueAddr ~= addr then
		da_putChar('+');
		if addr - trueAddr > 0xffff then
		    da_putLongHex(addr - trueAddr);
		else
		    da_putWordHex(addr - trueAddr);
		fi;
	    fi;
	    if DS*.ds_putAddress then
		da_putChar('[');
		if len = 'l' then
		    da_putLongHex(addr);
		else
		    da_putWordHex(addr);
		fi;
		da_putChar(']');
	    fi;
	fi;
    else
	symbol := nil;
    fi;
    if symbol = nil then
	if len = 'l' then
	    da_putLongHex(addr);
	else
	    da_putWordHex(addr);
	fi;
    fi;
corp;

/*
 * da_putRelCode - output any symbolic label for PC-relative mode.
 *	Return 'true' if we did any symbolic or label output.
 */

proc da_putRelCode(long offset; ulong refAt)bool:
    *char symbol;
    register ulong addr;

    addr := refAt + offset;
    if DS*.ds_findRelCode ~= nil then
	code(
	    OP_MOVEL | R_D0 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
	    offset,
	    OP_MOVEL | R_D1 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
	    refAt
	);
	symbol := DS*.ds_findRelCode(/* offset, refAt */);
	if symbol ~= nil then
	    if DS*.ds_putAddress then
		da_putChar('[');
		if DS*.ds_absoluteAddress then
		    da_putLongHex(addr);
		else
		    da_putWordHex(addr - DS*.ds_relativeBase);
		fi;
		da_putChar(']');
	    fi;
	    if DS*.ds_putRelForm then
		da_putChar('[');
		da_putLongDec(offset);
		da_putString("(pc)");
		da_putChar(']');
	    fi;
	    da_putString(symbol);
	fi;
    else
	symbol := nil;
    fi;
    refAt := addr;
    if symbol = nil then
	if DS*.ds_isLabel ~= nil then
	    code(
		OP_MOVEL | R_D0 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
		refAt
	    );
	    if DS*.ds_isLabel(/* refAt */) then
		if DS*.ds_putAddress then
		    da_putChar('[');
		    if DS*.ds_absoluteAddress then
			da_putLongHex(addr);
		    else
			da_putWordHex(addr - DS*.ds_relativeBase);
		    fi;
		    da_putChar(']');
		fi;
		if DS*.ds_putRelForm then
		    da_putChar('[');
		    da_putLongDec(offset);
		    da_putString("(pc)");
		    da_putChar(']');
		fi;
		da_putChar('L');
		if DS*.ds_absoluteAddress then
		    da_putLongHexRaw(addr);
		else
		    da_putWordHexRaw(addr - DS*.ds_relativeBase);
		fi;
		symbol := "";
	    fi;
	fi;
	if symbol = nil then
	    da_putChar('[');
	    if DS*.ds_absoluteAddress then
		da_putLongHex(addr);
	    else
		da_putWordHex(addr - DS*.ds_relativeBase);
	    fi;
	    da_putChar(']');
	    da_putLongDec(offset);
	fi;
    fi;
    if DS*.ds_labelAt ~= nil then
	code(
	    OP_MOVEL | R_D0 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
	    refAt
	);
	DS*.ds_labelAt(/* refAt */);
    fi;
    symbol ~= nil
corp;

/*
 * da_putRelData - output any symbolic label for A4-relative mode.
 *	Return 'true' if we could do so, else false.
 */

proc da_putRelData(long offset; ulong refAt)bool:
    *char symbol;
    long trueOffset;
    *long pTrueOffset;

    if DS*.ds_findRelData ~= nil then
	pTrueOffset := &trueOffset;
	code(
	    OP_MOVEL | R_D0 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
	    offset,
	    OP_MOVEL | R_D1 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
	    refAt,
	    OP_MOVEL | R_A0 << 9 | M_ADIR << 6 | M_DISP << 3 | R_FP,
	    pTrueOffset
	);
	symbol := DS*.ds_findRelData(/* offset, refAt, pTrueOffset */);
	if symbol ~= nil then
	    if DS*.ds_putRelForm then
		da_putChar('[');
		da_putLongDec(offset);
		da_putChar(']');
	    fi;
	    da_putString(symbol);
	    if trueOffset ~= offset then
		da_putChar('+');
		da_putLongDec(offset - trueOffset);
	    fi;
	    true
	else
	    false
	fi
    else
	false
    fi
corp;

/*
 * da_putSimpleRelData - common routine for simple forms of this.
 */

proc da_putSimpleRelData(long offset; ulong refAt)void:

    if not da_putRelData(offset, refAt) then
	da_putLongDec(offset);
    fi;
    da_putString("(a4)");
corp;

/*
 * da_putAReg - output an A-reg name.
 */

proc da_putAReg(uint reg)void:
    [9 * 3] char REG_NAMES = "a0\ea1\ea2\ea3\ea4\ea5\ea6\esp\epc";

    if DS*.ds_putChar ~= nil then
	da_putString(&REG_NAMES[reg * 3]);
    fi;
corp;

/*
 * da_putDReg - output a D-reg name.
 */

proc da_putDReg(uint reg)void:

    if DS*.ds_putChar ~= nil then
	da_putChar(
	    if DS*.ds_capExtended and DS*.ds_extendedNow then 'D' else 'd' fi);
	da_putChar(reg + '0');
    fi;
corp;

/*
 * da_putScaledIndex - output a possibly scaled index register name.
 */

proc da_putScaledIndex(register uint wrd)void:
    register uint reg;

    reg := (wrd & 0o070000) >> 12;
    if wrd & 0o100000 = 0 then
	da_putDReg(reg);
    else
	da_putAReg(reg);
    fi;
    /* can't quite use da_putLen for this - don't want the tab */
    da_putChar('.');
    if DS*.ds_capExtended and DS*.ds_extendedNow then
	da_putChar(if wrd & 0o004000 = 0 then 'W' else 'L' fi);
    else
	da_putChar(if wrd & 0o004000 = 0 then 'w' else 'l' fi);
    fi;
    case (wrd >> 9) & 0o3
    incase 1:
	da_putString("*2");
    incase 2:
	da_putString("*4");
    incase 3:
	da_putString("*8");
    esac;
corp;

/*
 * da_putIndexMode - output an indexed address mode. Fancy for 68020.
 */

proc da_putIndexMode(uint aReg)void:
    register proc(char ch)void putChar;
    register proc(uint wrd)void putScaledIndex;
    register ulong base;
    register long bdValue, odValue;
    register uint wrd;
    bool hasBase, hasIndex, putSym;

    putChar := da_putChar;
    putScaledIndex := da_putScaledIndex;
    base := DS*.ds_address;
    wrd := da_readWord();
    putSym := false;
    if wrd & 0o000400 = 0 then
	/* brief format extension word (as on 68000) */
	if aReg = R_PC then
	    putSym := da_putRelCode(make(wrd & 0xff, short), base);
	elif aReg ~= 4 or
	    not da_putRelData(make(wrd & 0xff, short), base)
	then
	    da_putWordDec(make(wrd & 0xff, short));
	fi;
	if DS*.ds_putChar ~= nil then
	    if wrd & 0o003000 ~= 0 then
		DS*.ds_extended := true;
		DS*.ds_extendedNow := true;
	    fi;
	    putChar('(');
	    if not putSym then
		da_putAReg(aReg);
		putChar(',');
	    fi;
	    putScaledIndex(wrd);
	    putChar(')');
	fi;
    else
	/* full format extension word */
	DS*.ds_extended := true;
	DS*.ds_extendedNow := true;
	case (wrd & 0o000060) >> 4
	incase 0:
	incase 1:
	    bdValue := 0;
	incase 2:
	    bdValue := make(da_readWord(), int);
	incase 3:
	    bdValue := da_readLong();
	esac;
	hasBase := wrd & 0o000200 = 0;
	hasIndex := wrd & 0o000100 = 0;
	if wrd & 0o3 = 0 then
	    /* no memory indirect */
	    if wrd & 0o000300 ~= 0o000300 then
		if hasBase and aReg = R_PC then
		    putSym := da_putRelCode(bdValue, base);
		elif not hasBase or aReg ~= 4 or
		    not da_putRelData(bdValue, base)
		then
		    da_putLongDec(bdValue);
		fi;
		if putSym then
		    if hasIndex then
			putChar('(');
			putScaledIndex(wrd);
			putChar(')');
		    fi;
		else
		    putChar('(');
		    if hasBase then
			da_putAReg(aReg);
			if hasIndex then
			    putChar(',');
			fi;
		    fi;
		    if hasIndex then
			putScaledIndex(wrd);
		    fi;
		    putChar(')');
		fi;
	    else
		da_putLongDec(bdValue);
	    fi;
	else
	    /* memory indirect of some kind */
	    case wrd & 0o3
	    incase 1:
		odValue := 0;
	    incase 2:
		odValue := make(da_readWord(), int);
	    incase 3:
		odValue := da_readLong();
	    esac;
	    da_putLongDec(odValue);
	    putChar('(');
	    if hasBase and aReg = R_PC then
		putSym := da_putRelCode(bdValue, base);
	    elif not hasBase or aReg ~= 4 or
		not da_putRelData(bdValue, base)
	    then
		da_putLongDec(bdValue);
	    fi;
	    if putSym then
		if hasIndex then
		    if wrd & 0o000004 = 0 then
			/* pre-indexed */
			putChar('(');
			putScaledIndex(wrd);
			putChar(')');
		    else
			/* post-indexed */
			putChar(',');
			putScaledIndex(wrd);
		    fi;
		fi;
	    else
		putChar('(');
		if hasBase then
		    da_putAReg(aReg);
		else
		    putChar('0');
		fi;
		if wrd & 0o000004 = 0 and hasIndex then
		    /* pre-indexed */
		    putChar(',');
		    putScaledIndex(wrd);
		fi;
		putChar(')');
		if wrd & 0o000004 ~= 0 and hasIndex then
		    /* post-indexed */
		    putChar(',');
		    putScaledIndex(wrd);
		fi;
	    fi;
	    putChar(')');
	fi;
    fi;
corp;

/*
 * da_putEA - output an effective address and all its baggage.
 */

proc da_putEA(uint ea; register char len)void:
    register proc(char ch)void putChar;
    register ulong base;
    register long offset;
    register uint reg;
    ulong addr;
    bool isBranch, saveExtended;

    putChar := da_putChar;
    if len = 'c' then
	isBranch := true;
	len := 'l';
    else
	isBranch := false;
    fi;
    base := DS*.ds_address;
    saveExtended := DS*.ds_extendedNow;
    reg := ea & 0o7;
    case ea >> 3
    incase M_DDIR:
	da_putDReg(reg);
    incase M_ADIR:
	da_putAReg(reg);
    incase M_INDIR:
	/* change: do not use 'da_putSimpleRelData' here. The compilers that
	   use a4 addressing for variables will not use M_INDIR, since they
	   do not know the true offset until after linking. */
	if DS*.ds_putChar ~= nil then
	    putChar('(');
	    da_putAReg(reg);
	    putChar(')');
	fi;
    incase M_INC:
	if DS*.ds_putChar ~= nil then
	    putChar('(');
	    da_putAReg(reg);
	    putChar(')');
	    putChar('+');
	fi;
    incase M_DEC:
	if DS*.ds_putChar ~= nil then
	    putChar('-');
	    putChar('(');
	    da_putAReg(reg);
	    putChar(')');
	fi;
    incase M_DISP:
	if reg = 4 then
	    da_putSimpleRelData(make(da_readWord(), int), base);
	else
	    da_putWordDec(da_readWord());
	    putChar('(');
	    da_putAReg(reg);
	    putChar(')');
	fi;
    incase M_INDEX:
	da_putIndexMode(reg);
    incase M_SPECIAL:
	case reg
	incase M_ABSSHORT:
	    da_putAbsSymbol(da_readWord(), base, 'w', isBranch);
	incase M_ABSLONG:
	    da_putAbsSymbol(da_readLong(), base, 'l', isBranch);
	incase M_PCDISP:
	    offset := make(da_readWord(), int);
	    if not da_putRelCode(offset, base) then
		da_putString("(pc)");
	    fi;
	    if isBranch and DS*.ds_branchTo ~= nil then
		addr := base + offset;
		code(
		    OP_MOVEL | R_D0 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
		    addr
		);
		DS*.ds_branchTo(/* addr */);
	    fi;
	incase M_PCINDEX:
	    da_putIndexMode(R_PC);
	incase M_IMM:
	    putChar('\#');
	    case len
	    incase 'b':
		da_putWordDec(make(da_readWord() & 0xff, short));
	    incase 'w':
		da_putWordDec(da_readWord());
	    incase 's':
	    incase 'x':
	    incase 'p':
	    incase 'd':
		reg := if len = 's' then 1 elif len = 'd' then 2 else 3 fi;
		da_putString("0x");
		while reg ~= 0 do
		    reg := reg - 1;
		    da_putLongHexRaw(da_readLong());
		od;
	    default:
		da_putAbsSymbol(da_readLong(), base, 'l', false);
	    esac;
	default:
	    da_putString("???");
	    DS*.ds_illegal := true;
	esac;
    esac;
    DS*.ds_extendedNow := saveExtended;
corp;

/*
 * da_putBranch - output a PC-relative branch offset.
 */

proc da_putBranch(register int offset)void:
    register ulong base;
    ulong addr;

    base := DS*.ds_address;
    if offset = 0 then
	offset := da_readWord();
    elif offset = -1 then
	offset := da_readLong();
    fi;
    if not da_putRelCode(offset, base) then
	da_putString("(pc)");
    fi;
    if DS*.ds_branchTo ~= nil then
	addr := base + offset;
	code(
	    OP_MOVEL | R_D0 << 9 | M_DDIR << 6 | M_DISP << 3 | R_FP,
	    addr
	);
	DS*.ds_branchTo(/* addr */);
    fi;
corp;

/*
 * da_putCondition - output a condition for a condition branch, etc.
 */

proc da_putCondition(uint condition)void:
    *char COND =
	"t \ef \ehi\els\ehs\elo\ene\eeq\evc\evs\epl\emi\ege\elt\egt\ele";

    da_putString(COND + condition * 3);
corp;

/*
 * da_putOpReg - output an instruction that is a register operation.
 */

proc da_putOpReg(*char op; char len; uint opCode)void:

    da_putString(op);
    da_putLen(len);
    da_putEA(opCode & 0o77, len);
    da_putChar(',');
    da_putDReg((opCode & 0o007000) >> 9);
corp;

/*
 * da_putOpMult - output part of an instruction with the multi-operand format.
 */

proc da_putOpMult(register uint opCode)void:

    if opCode & 0o10 = 0 then
	da_putDReg(opCode & 0o7);
	da_putChar(',');
	da_putDReg((opCode & 0o007000) >> 9);
    else
	da_putString("-(");
	da_putAReg(opCode & 0o7);
	da_putString("),-(");
	da_putAReg((opCode & 0o007000) >> 9);
	da_putChar(')');
    fi;
corp;

/*
 * da_putOpModed - output one of the common instructions with an op-mode.
 */

proc da_putOpModed(*char op; register uint opCode)void:
    register char len;
    register uint reg, ea;

    da_putString(op);
    reg := (opCode & 0o007000) >> 9;
    ea := opCode & 0o77;
    if opCode & 0o000300 = 0o000300 then
	/* special for adda, suba, cmpa */
	len := if opCode & 0o000700 = 0o000300 then 'w' else 'l' fi;
	da_putLen(len);
	da_putEA(ea, len);
	da_putChar(',');
	da_putAReg(reg);
    else
	len := LENGTHS[(opCode & 0o000300) >> 6];
	da_putLen(len);
	if opCode & 0o000400 = 0 then
	    da_putEA(ea, len);
	    da_putChar(',');
	    da_putDReg(reg);
	else
	    da_putDReg(reg);
	    da_putChar(',');
	    da_putEA(ea, len);
	fi;
    fi;
corp;

/*
 * da_putOpSingle - output one of the single operand instructions.
 */

proc da_putOpSingle(*char op; char len; uint ea)void:

    da_putString(op);
    da_putLen(len);
    da_putEA(ea, len);
corp;

/*
 * da_putRegList - output a register list for movem.
 */

proc da_putRegList(register uint w; bool goingUp)void:
    register uint mask, i;
    register bool needSlash;

    if DS*.ds_putChar ~= nil then
	da_putChar('<');
	needSlash := false;
	mask := 0x8000;
	for i from 0 upto 7 do
	    if w & mask ~= 0 then
		if needSlash then
		    da_putChar('/');
		fi;
		if goingUp then
		    da_putAReg(7 - i);
		else
		    da_putDReg(i);
		fi;
		needSlash := true;
	    fi;
	    mask := mask >> 1;
	od;
	for i from 0 upto 7 do
	    if w & mask ~= 0 then
		if needSlash then
		    da_putChar('/');
		fi;
		if goingUp then
		    da_putDReg(7 - i);
		else
		    da_putAReg(i);
		fi;
		needSlash := true;
	    fi;
	    mask := mask >> 1;
	od;
	da_putChar('>');
    fi;
corp;
