/*
 * Nodes.g - definition of structures used in the language libraries.
 */

type
    Construct_t = struct {
	uint con_token; 	/* this token starts the construct */
	proc(*ParserState_t ps; *Construct_t con)void con_builder;
	proc(*ParserState_t ps; *ExpressionNode_t exn)void con_destructor;
	proc(*InterpreterState_t ex; *ExpressionNode_t exn)void
	    con_interpreter;
	proc(*PrintState_t prs; *ExpressionNode_t exn)void con_printer;
    },

    OperationKind_t = enum {
	opk_prefix,		/* prefix operation, eg. unary '-', C '*' */
	opk_postfix,		/* postfix op, e.g. '[', Draco '*', C '->' */
	opk_binary		/* regular binary operation */
    },

    Operation_t = struct {
	uint oper_token;
	OperationKind_t oper_kind;
	proc(*ParserState_t ps; uint token)void oper_checker1;
	proc(*ParserState_t ps; uint token)void oper_checker2;
    },

    ExpressionNodeList_t = struct {
	*ExpressionNodeList_t exnl_next;
	*ExpressionNode_t exnl_this;
    },

    ExpressionKind_t = enum {
	exk_prefix,			/* node is a prefix operation */
	exk_postfix,			/* node is a postfix operation */
	exk_binary,			/* node is a binary operation */
	exk_special,			/* node is a special construct */
	exk_string,			/* node is a string constant */
	exk_id, 			/* node is an identifier */
	exk_ulong,			/* node is an integral constant */
	exk_float,			/* node is a floating point constant */
	exk_char,			/* node is a character constant */
	exk_list,			/* node is an expression node list */
	exk_comment,			/* a comment preserved */
	exk_paren,			/* a parenthesized sub-expression */
	exk_error			/* node represents an error */
    },

    ExpressionNode_t = struct {
	ExpressionKind_t exn_kind;
	uint exn_code;
	TypeValue_t exn_type;		/* application-specific */
	union {
	    struct {	/* prefix or postfix operation */
		*Operation_t exu_operation;
		*ExpressionNode_t exu_operand;
	    } exu_un;
	    struct {	/* binary operation */
		*Operation_t exb_operation;
		*ExpressionNode_t exb_left;
		*ExpressionNode_t exb_right;
	    } exu_bi;
	    struct {	/* construct of some kind */
		*Construct_t exc_construct;
		arbptr exc_tree;
	    } exu_con;
	    struct {	/* a string constant or a comment */
		*char exs_string;
		ulong exs_length;
	    } exu_string;
	    *Symbol_t exu_id;
	    ulong exu_ulong;
	    float exu_float;
	    char exu_char;
	    *ExpressionNode_t exu_paren;
	    *ExpressionNodeList_t exu_list;
	} exn_;
    },

    SymbolList_t = struct {
	*SymbolList_t syl_next;
	*Symbol_t syl_this;
    },

    Procedure_t = struct {
	*SymbolList_t pr_parameters;
	TypeValue_t pr_resultType;
	*SymbolList_t pr_locals;
	*ExpressionNode_t pr_body;
    };
