#LexInternal.g

ulong MAXIMUM_INTEGER = 0xffffffff;

/* very special register declaration */

register *TokenizerState_t TK;

/*
 * tk_addChar - private entry to add a character to our local buffer.
 */

proc tk_addChar(char ch)void:
    register ulong position;

    if not TK*.tk_bufferOverflow then
	position := TK*.tk_bufferPosition;
	if position = TK*.tk_bufferLength then
	    /* expand buffer */
	    if not TK*.tk_doubleBuffer(TK) then
		TK*.tk_bufferOverflow := true;
		TK*.tk_stringError("string/identifier buffer overflow");
	    fi;
	fi;
	(TK*.tk_buffer + position * sizeof(char))* := ch;
	TK*.tk_bufferPosition := position + 1;
    fi;
corp;

/*
 * tk_skipWhiteSpace - skip past white space characters and comments.
 */

proc tk_skipWhiteSpace()void:
    register uint commentLevel;
    register char nextChar, currentChar;
    register CharacterClass_t cc;
    register proc()char getChar;

    currentChar := TK*.tk_currentChar;
    nextChar := TK*.tk_nextChar;
    getChar := TK*.tk_getChar;
    TK*.tk_gotComment := false;

    commentLevel := 0;
    while
	cc := TK*.tk_characterClass[currentChar];
	if cc = cc_whiteSpace or cc = cc_newLine and not TK*.tk_reportNewLines
	then
	    if TK*.tk_preserveComments and TK*.tk_gotComment then
		tk_addChar(currentChar);
	    fi;
	    currentChar := nextChar;
	    nextChar := getChar();
	    true
	elif commentLevel = 0 and currentChar = TK*.tk_lineComment[0] and
	    (not TK*.tk_inString or not TK*.tk_preserveComments) and
	    (TK*.tk_lineComment[1] = ' ' or TK*.tk_lineComment[1] = nextChar)
	then
	    /* an 'upto-end-of-line' comment */
	    if TK*.tk_preserveComments then
		TK*.tk_bufferPosition := 0;
		TK*.tk_bufferOverflow := false;
		TK*.tk_gotComment := true;
	    fi;
	    while cc ~= cc_newLine and cc ~= cc_endOfFile do
		if TK*.tk_preserveComments then
		    tk_addChar(currentChar);
		fi;
		currentChar := nextChar;
		cc := TK*.tk_characterClass[currentChar];
		nextChar := getChar();
	    od;
	    currentChar := nextChar;
	    nextChar := getChar();
	    not TK*.tk_preserveComments
	elif currentChar = TK*.tk_commentStart[0] and
	    (not TK*.tk_inString or not TK*.tk_preserveComments) and
	    (TK*.tk_commentsNest or commentLevel = 0) and
	    (TK*.tk_commentStart[1] = ' ' or TK*.tk_commentStart[1] = nextChar)
	then
	    if TK*.tk_preserveComments then
		if commentLevel = 0 then
		    TK*.tk_bufferPosition := 0;
		    TK*.tk_bufferOverflow := false;
		    TK*.tk_gotComment := true;
		fi;
		tk_addChar(currentChar);
		tk_addChar(nextChar);
	    fi;
	    commentLevel := commentLevel + 1;
	    currentChar := getChar();
	    nextChar := getChar();
	    true
	elif commentLevel ~= 0 then
	    if cc = cc_endOfFile then
		false
	    else
		if currentChar = TK*.tk_commentEnd[0] and
		    (TK*.tk_commentEnd[1] = ' ' or
		     TK*.tk_commentEnd[1] = nextChar)
		then
		    if TK*.tk_preserveComments then
			tk_addChar(currentChar);
			tk_addChar(nextChar);
		    fi;
		    commentLevel := commentLevel - 1;
		    currentChar := getChar();
		    nextChar := getChar();
		else
		    if TK*.tk_preserveComments then
			tk_addChar(currentChar);
		    fi;
		    currentChar := nextChar;
		    nextChar := getChar();
		fi;
		commentLevel ~= 0 or not TK*.tk_preserveComments
	    fi
	else
	    false
	fi
    do
    od;

    TK*.tk_currentChar := currentChar;
    TK*.tk_nextChar := nextChar;
corp;

/*
 * tk_getNumber - get a numeric constant.
 */

proc tk_getNumber()void:
    register ulong valu, digit;
    ulong base;
    bool overFlow;
    register char currentChar;
    char nextChar;

    currentChar := TK*.tk_currentChar;
    nextChar := TK*.tk_nextChar;

    overFlow := false;
    valu := 0;
    base := 10;
    TK*.tk_bufferPosition := 0;
    if currentChar = '0' then
	tk_addChar('0');
	currentChar := nextChar;
	nextChar := TK*.tk_getChar();
	if currentChar = 'x' or currentChar = 'X' then
	    base := 16;
	    tk_addChar(currentChar);
	    currentChar := nextChar;
	    nextChar := TK*.tk_getChar();
	elif currentChar = 'o' or currentChar = 'O' then
	    base := 8;
	    tk_addChar(currentChar);
	    currentChar := nextChar;
	    nextChar := TK*.tk_getChar();
	elif currentChar = 'b' or currentChar = 'B' then
	    base := 2;
	    tk_addChar(currentChar);
	    currentChar := nextChar;
	    nextChar := TK*.tk_getChar();
	fi;
    fi;

    while currentChar >= '0' and currentChar <= '9' or base = 16 and
	(currentChar >= 'a' and currentChar <= 'f' or
	 currentChar >= 'A' and currentChar <= 'F')
    do
	tk_addChar(currentChar);
	digit :=
	    if currentChar >= '0' and currentChar <= '9' then
		currentChar - '0'
	    elif currentChar >= 'a' then
		currentChar - ('a' - 10)
	    else
		currentChar - ('A' - 10)
	    fi;
	if digit >= base then
	    TK*.tk_stringError("illegal digit in integer constant");
	fi;
	if valu > MAXIMUM_INTEGER / base or
	    valu = MAXIMUM_INTEGER / base and digit > MAXIMUM_INTEGER % base
	then
	    if not overFlow then
		overFlow := true;
		TK*.tk_stringError("overflow in integer constant");
	    fi;
	else
	    valu := valu * base + digit;
	fi;
	currentChar := nextChar;
	nextChar := TK*.tk_getChar();
    od;

    if currentChar = '.' then
	TK*.tk_kind := tk_floatConst;
    else
	TK*.tk_ulongValue := valu;
	TK*.tk_kind := tk_ulongConst;
    fi;

    TK*.tk_currentChar := currentChar;
    TK*.tk_nextChar := nextChar;
corp;

/*
 * tk_getCharacter - get a single character for a char or string constant.
 */

proc tk_getCharacter(char delim)void:
    *char saveBuffer;
    ulong saveLength, savePosition;
    register ulong valu;
    register char currentChar;

    TK*.tk_kind := tk_charConst;
    currentChar := TK*.tk_currentChar;
    if currentChar = TK*.tk_stringEscape then
	/* handle an escaped character */
	currentChar := TK*.tk_nextChar;
	TK*.tk_nextChar := TK*.tk_getChar();
	if currentChar = TK*.tk_stringEscape or currentChar = delim then
	    TK*.tk_charValue := currentChar;
	    TK*.tk_currentChar := TK*.tk_nextChar;
	    TK*.tk_nextChar := TK*.tk_getChar();
	else
	    case currentChar
	    incase 'e':
		TK*.tk_charValue := '\e';
		TK*.tk_currentChar := TK*.tk_nextChar;
		TK*.tk_nextChar := TK*.tk_getChar();
	    incase 't':
		TK*.tk_charValue := '\t';
		TK*.tk_currentChar := TK*.tk_nextChar;
		TK*.tk_nextChar := TK*.tk_getChar();
	    incase 'n':
		TK*.tk_charValue := '\n';
		TK*.tk_currentChar := TK*.tk_nextChar;
		TK*.tk_nextChar := TK*.tk_getChar();
	    incase 'r':
		TK*.tk_charValue := '\r';
		TK*.tk_currentChar := TK*.tk_nextChar;
		TK*.tk_nextChar := TK*.tk_getChar();
	    incase 'b':
		TK*.tk_charValue := '\b';
		TK*.tk_currentChar := TK*.tk_nextChar;
		TK*.tk_nextChar := TK*.tk_getChar();
	    incase '(':
		TK*.tk_currentChar := TK*.tk_nextChar;
		TK*.tk_nextChar := TK*.tk_getChar();
		saveBuffer := TK*.tk_buffer;
		saveLength := TK*.tk_bufferLength;
		savePosition := TK*.tk_bufferPosition;
		TK*.tk_buffer := nil;
		TK*.tk_bufferLength := 0;
		valu := TK*.tk_getExpression();
		TK*.tk_buffer := saveBuffer;
		TK*.tk_bufferLength := saveLength;
		TK*.tk_bufferPosition := savePosition;
		TK*.tk_kind := tk_charConst;
		if valu > 0xff then
		    TK*.tk_stringError(
			"overflow in escaped character expression");
		    valu := valu & 0xff;
		fi;
		TK*.tk_charValue := valu + '\e';
		if TK*.tk_currentChar = ')' then
		    TK*.tk_currentChar := TK*.tk_nextChar;
		    TK*.tk_nextChar := TK*.tk_getChar();
		else
		    TK*.tk_stringError(
			"missing ')' in escaped character expression");
		fi;
	    default:
		TK*.tk_stringError("illegal escaped character");
		TK*.tk_charValue := currentChar;
		TK*.tk_currentChar := TK*.tk_nextChar;
		TK*.tk_nextChar := TK*.tk_getChar();
	    esac;
	fi;
    else
	TK*.tk_charValue := currentChar;
	TK*.tk_currentChar := TK*.tk_nextChar;
	TK*.tk_nextChar := TK*.tk_getChar();
    fi;
corp;

/*
 * tk_getString - get a quoted string.
 */

proc tk_getString()void:
    register CharacterClass_t cc;
    register char delim;
    register proc()char getChar;

    TK*.tk_bufferPosition := 0;
    TK*.tk_bufferOverflow := false;
    TK*.tk_inString := true;
    getChar := TK*.tk_getChar;
    delim := TK*.tk_currentChar;

    /* outer loop - collect all pieces for a string break */

    while

	/* skip the opening quote */

	TK*.tk_currentChar := TK*.tk_nextChar;
	TK*.tk_nextChar := getChar();

	/* inner loop - go until the closing quote (or end of line) */

	while
	    cc := TK*.tk_characterClass[TK*.tk_currentChar];
	    cc ~= cc_newLine and cc ~= cc_endOfFile and
	    TK*.tk_currentChar ~= delim
	do
	    tk_getCharacter(delim);
	    tk_addChar(TK*.tk_charValue);
	od;

	/* we are now at a closing quote or end of line - see if we should do
	   a string break */

	if TK*.tk_currentChar = delim then
	    TK*.tk_currentChar := TK*.tk_nextChar;
	    TK*.tk_nextChar := getChar();
	else
	    TK*.tk_stringError("missing closing quote for string constant");
	fi;
	if TK*.tk_stringBreakAllowed then
	    tk_skipWhiteSpace();
	    TK*.tk_currentChar = delim
	else
	    false
	fi
    do
    od;
    tk_addChar('\e');

    TK*.tk_kind := tk_stringConst;
corp;

/*
 * tk_GetNextToken - top level of tokenizer - get the next input token.
 */

proc tk_GetNextToken(*TokenizerState_t tk)void:
    *ReservedWord_t rw;
    register *char p1, p2;
    register *Operator_t op @ p1;
    register uint hi, lo, mid, count @ mid;
    register char nextChar @ hi, currentChar @ lo, delim @ mid;

    TK := tk;

    /* skip past any leading whitespace and comments */

    TK*.tk_inString := false;
    tk_skipWhiteSpace();
    if TK*.tk_gotComment then
	tk_addChar('\e');
	TK*.tk_kind := tk_comment;
	return;
    fi;

    /* get the token */

    currentChar := TK*.tk_currentChar;

    case TK*.tk_characterClass[currentChar]
    incase cc_identifier:

	/* identifier or reserved word - first collect the letters */

	nextChar := TK*.tk_nextChar;

	TK*.tk_kind := tk_id;
	TK*.tk_bufferPosition := 0;
	TK*.tk_bufferOverflow := false;
	while TK*.tk_characterClass[currentChar] = cc_identifier or
	    TK*.tk_characterClass[currentChar] = cc_digit
	do
	    tk_addChar(currentChar);
	    currentChar := nextChar;
	    nextChar := TK*.tk_getChar();
	od;
	tk_addChar('\e');

	TK*.tk_currentChar := currentChar;
	TK*.tk_nextChar := nextChar;

	/* now check to see if it is a reserved word */

	hi := TK*.tk_reservedWordCount + 1;
	if hi ~= 1 then
	    lo := 0;
	    while lo + 1 < hi do
		mid := (hi + lo) / 2;
		rw := TK*.tk_reservedWords + (mid -1) * sizeof(ReservedWord_t);
		p1 := TK*.tk_buffer;
		p2 := rw*.rw_text;
		while p1* = p2* and p1* ~= '\e' do
		    p1 := p1 + sizeof(char);
		    p2 := p2 + sizeof(char);
		od;
		if p1* > p2* then
		    lo := mid;
		else
		    hi := mid;
		fi;
	    od;
	    if lo ~= TK*.tk_reservedWordCount then
		rw := TK*.tk_reservedWords + lo * sizeof(ReservedWord_t);
		p1 := TK*.tk_buffer;
		p2 := rw*.rw_text;
		while p1* = p2* and p1* ~= '\e' do
		    p1 := p1 + sizeof(char);
		    p2 := p2 + sizeof(char);
		od;
		if p1* = p2* then
		    TK*.tk_kind := rw*.rw_code;
		fi;
	    fi;
	fi;

    incase cc_digit:

	/* a numeric constant */

	tk_getNumber();

    incase cc_string:

	/* a string constant */

	tk_getString();

    incase cc_char:

	/* a character constant */

	delim := TK*.tk_currentChar;
	currentChar := TK*.tk_nextChar;
	TK*.tk_currentChar := currentChar;
	TK*.tk_nextChar := TK*.tk_getChar();
	TK*.tk_kind := tk_charConst;
	if TK*.tk_characterClass[currentChar] ~= cc_newLine and
	    TK*.tk_characterClass[currentChar] ~= cc_endOfFile
	then
	    tk_getCharacter(delim);
	fi;
	if TK*.tk_currentChar = delim then
	    TK*.tk_currentChar := TK*.tk_nextChar;
	    TK*.tk_nextChar := TK*.tk_getChar();
	else
	    TK*.tk_stringError("missing closing quote for character constant");
	fi;

    incase cc_operator:

	/* an operator character */

	nextChar := TK*.tk_nextChar;
	op := TK*.tk_operators;
	count := TK*.tk_operatorCount;
	while count ~= 0 and
	    (op*.op_text[0] ~= currentChar or
	     op*.op_text[1] ~= ' ' and op*.op_text[1] ~= nextChar)
	do
	    count := count - 1;
	    op := op + sizeof(Operator_t);
	od;
	if count = 0 then
	    TK*.tk_stringError("unknown operator");
	    TK*.tk_kind := tk_illegal;
	    TK*.tk_charValue := currentChar;
	else
	    TK*.tk_kind := op*.op_code;
	    if op*.op_text[1] ~= ' ' then
		TK*.tk_currentChar := nextChar;
		TK*.tk_nextChar := TK*.tk_getChar();
	    fi;
	fi;
	TK*.tk_currentChar := TK*.tk_nextChar;
	TK*.tk_nextChar := TK*.tk_getChar();

    incase cc_whiteSpace:

	/* this cannot happen - we have already skipped whitespace */

	;

    incase cc_newLine:

	/* a newline character. If we are not reporting newlines, then we
	   would not get here since the newlines would be absorbed as
	   whitespace. */

	TK*.tk_kind := tk_newLine;
	TK*.tk_currentChar := TK*.tk_nextChar;
	TK*.tk_nextChar := TK*.tk_getChar();

    incase cc_illegal:

	/* illegal character(s) */

	TK*.tk_stringError("illegal characters encountered");
	TK*.tk_kind := tk_illegal;
	TK*.tk_bufferPosition := 0;
	TK*.tk_bufferOverflow := false;
	while TK*.tk_characterClass[currentChar] = cc_illegal do
	    tk_addChar(currentChar);
	    currentChar := TK*.tk_nextChar;
	    TK*.tk_nextChar := TK*.tk_getChar();
	od;
	TK*.tk_currentChar := currentChar;

    incase cc_endOfFile:

	/* end of file */

	TK*.tk_kind := tk_endOfFile;

    esac;
corp;
