type
    CharacterClass_t = enum {
	cc_identifier,
	cc_digit,
	cc_string,
	cc_char,
	cc_operator,
	cc_whiteSpace,
	cc_newLine,
	cc_illegal,
	cc_endOfFile
    },

    ReservedWord_t = struct {
	*char rw_text;
	uint rw_code;
    },

    Operator_t = struct {
	[2] char op_text;
	uint op_code;
    },

    TokenizerState_t = struct {

	/* proc which supplies characters for tokenization */

	proc()char tk_getChar;

	/* proc used to go get a constant expression for escaped character */

	proc()ulong tk_getExpression;

	/* proc to call to emit a single-string error message */

	proc(*char message)void tk_stringError;

	/* proc to call to double the size of tk_buffer/tk_bufferLength and
	   copy tk_bufferPosition bytes from the old one to the new one */

	proc(*TokenizerState_t tk)bool tk_doubleBuffer;

	/* fields which describe the tokenization desired */

	*ReservedWord_t tk_reservedWords;
	uint tk_reservedWordCount;
	*Operator_t tk_operators;
	uint tk_operatorCount;
	[256] CharacterClass_t tk_characterClass;
	[2] char tk_lineComment;
	[2] char tk_commentStart, tk_commentEnd;
	char tk_stringEscape;
	bool tk_commentsNest;
	bool tk_stringBreakAllowed;
	bool tk_reportNewLines;
	bool tk_preserveComments;

	/* the resulting token */

	float tk_floatValue;
	ulong tk_ulongValue;
	char tk_charValue;
	uint tk_kind;
	*char tk_buffer;
	ulong tk_bufferLength, tk_bufferPosition;

	/* internal fields */

	char tk_currentChar;
	char tk_nextChar;
	bool tk_bufferOverflow;
	bool tk_inString;
	bool tk_gotComment;
    };

/* codes for the standard token types */

uint
    tk_endOfFile	= 0,
    tk_illegal		= tk_endOfFile + 1,
    tk_id		= tk_illegal + 1,
    tk_ulongConst	= tk_id + 1,
    tk_floatConst	= tk_ulongConst + 1,
    tk_stringConst	= tk_floatConst + 1,
    tk_charConst	= tk_stringConst + 1,
    tk_newLine		= tk_charConst + 1,
    tk_comment		= tk_newLine + 1,

    /* these are provided for use by users of the parser library */

    tk_semicolon	= tk_comment + 1,
    tk_comma		= tk_semicolon + 1,
    tk_openParen	= tk_comma + 1,
    tk_closeParen	= tk_openParen + 1;

extern
    OpenlexLibrary(ulong version)*Library_t,
    CloselexLibrary()void,
    GetNextToken(*TokenizerState_t tk)void;
