#LexInternal.g
#LexStandardInternal.g
#ansi.g

uint
    R_D0 = 0,
    R_D1 = 1,
    R_A0 = 0,
    R_A1 = 1,
    R_FP = 6,
    OP_MOVEL = 0x2000,
    M_DDIR = 0,
    M_ADIR = 1,
    M_DISP = 5;

[STANDARD_RESERVED_WORD_COUNT] ReservedWord_t StandardReservedWords := (
    ("and"	  , tk_and	  ),
    ("arbptr"	  , tk_arbptr	  ),
    ("binary"	  , tk_binary	  ),
    ("bool"	  , tk_bool	  ),
    ("byte"	  , tk_byte	  ),
    ("case"	  , tk_case	  ),
    ("channel"	  , tk_channel	  ),
    ("char"	  , tk_char	  ),
    ("chip"	  , tk_chip	  ),
    ("close"	  , tk_close	  ),
    ("corp"	  , tk_corp	  ),
    ("default"	  , tk_default	  ),
    ("dim"	  , tk_dim	  ),
    ("do"	  , tk_do	  ),
    ("downto"	  , tk_downto	  ),
    ("elif"	  , tk_elif	  ),
    ("else"	  , tk_else	  ),
    ("enum"	  , tk_enum	  ),
    ("error"	  , tk_error	  ),
    ("esac"	  , tk_esac	  ),
    ("export"	  , tk_export	  ),
    ("extern"	  , tk_extern	  ),
    ("fallthrough", tk_fallthrough),
    ("fi"	  , tk_fi	  ),
    ("fix"	  , tk_fix	  ),
    ("float"	  , tk_float	  ),
    ("flt"	  , tk_flt	  ),
    ("for"	  , tk_for	  ),
    ("free"	  , tk_free	  ),
    ("from"	  , tk_from	  ),
    ("if"	  , tk_if	  ),
    ("ignore"	  , tk_ignore	  ),
    ("import"	  , tk_import	  ),
    ("incase"	  , tk_incase	  ),
    ("input"	  , tk_input	  ),
    ("int"	  , tk_int	  ),
    ("ioerror"	  , tk_ioerror	  ),
    ("long"	  , tk_long	  ),
    ("make"	  , tk_make	  ),
    ("module"	  , tk_module	  ),
    ("new"	  , tk_new	  ),
    ("nil"	  , tk_nil	  ),
    ("not"	  , tk_not	  ),
    ("od"	  , tk_od	  ),
    ("open"	  , tk_open	  ),
    ("or"	  , tk_or	  ),
    ("output"	  , tk_output	  ),
    ("private"	  , tk_private	  ),
    ("proc"	  , tk_proc	  ),
    ("public"	  , tk_public	  ),
    ("range"	  , tk_range	  ),
    ("read"	  , tk_read	  ),
    ("readln"	  , tk_readln	  ),
    ("register"   , tk_register   ),
    ("return"	  , tk_return	  ),
    ("short"	  , tk_short	  ),
    ("signed"	  , tk_signed	  ),
    ("sizeof"	  , tk_sizeof	  ),
    ("step"	  , tk_step	  ),
    ("string"	  , tk_string	  ),
    ("struct"	  , tk_struct	  ),
    ("text"	  , tk_text	  ),
    ("then"	  , tk_then	  ),
    ("type"	  , tk_type	  ),
    ("uint"	  , tk_uint	  ),
    ("ulong"	  , tk_ulong	  ),
    ("union"	  , tk_union	  ),
    ("unknown"	  , tk_unknown	  ),
    ("unsigned"   , tk_unsigned   ),
    ("upto"	  , tk_upto	  ),
    ("ushort"	  , tk_ushort	  ),
    ("void"	  , tk_void	  ),
    ("while"	  , tk_while	  ),
    ("write"	  , tk_write	  ),
    ("writeln"	  , tk_writeln	  )
);

[STANDARD_OPERATOR_COUNT] Operator_t StandardOperators := (
    (('<', '<'), tk_shiftLeft	  ),
    (('>', '>'), tk_shiftRight	  ),
    (('>', '<'), tk_bitXor	  ),
    ((':', '='), tk_assign	  ),
    (('<', '='), tk_lessOrEqual   ),
    (('>', '='), tk_greaterOrEqual),
    (('~', '='), tk_notEqual	  ),
    (('.', '.'), tk_dotDot	  ),
    ((';', ' '), tk_semicolon	  ),
    ((',', ' '), tk_comma	  ),
    (('(', ' '), tk_openParen	  ),
    ((')', ' '), tk_closeParen	  ),
    (('+', ' '), tk_plus	  ),
    (('-', ' '), tk_minus	  ),
    (('*', ' '), tk_star	  ),
    (('/', ' '), tk_slash	  ),
    (('%', ' '), tk_percent	  ),
    (('&', ' '), tk_bitAnd	  ),
    (('|', ' '), tk_bitOr	  ),
    (('~', ' '), tk_bitNot	  ),
    (('<', ' '), tk_lessThan	  ),
    (('>', ' '), tk_greaterThan   ),
    (('=', ' '), tk_equal	  ),
    (('.', ' '), tk_dot 	  ),
    ((':', ' '), tk_colon	  ),
    (('[', ' '), tk_openSquare	  ),
    ((']', ' '), tk_closeSquare   ),
    (('{', ' '), tk_openBrace	  ),
    (('}', ' '), tk_closeBrace	  ),
    (('$', ' '), tk_dollar	  )
);

/*
 * LexStandard - fill in a tokenizer state structure with a set of standard
 *	values. Note that the reserved word strings are in the shared
 *	library, so the caller must not change or free them.
 */

proc LexStandard(/* register *TokenizerState_t tk */)void:
    *TokenizerState_t tkState;
    register *TokenizerState_t tk;

    code (
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | R_A0,
	tkState
    );
    tk := tkState;
    if tk*.tk_reservedWords ~= nil then
	pretend(tk*.tk_reservedWords,
		*[STANDARD_RESERVED_WORD_COUNT] ReservedWord_t)* :=
	    StandardReservedWords;
	tk*.tk_reservedWordCount := STANDARD_RESERVED_WORD_COUNT;
    fi;
    if tk*.tk_operators ~= nil then
	pretend(tk*.tk_operators,
		*[STANDARD_OPERATOR_COUNT] Operator_t)* := StandardOperators;
	tk*.tk_operatorCount := STANDARD_OPERATOR_COUNT;
    fi;
    tk*.tk_lineComment [0] := '-'; tk*.tk_lineComment [1] := '-';
    tk*.tk_commentStart[0] := '/'; tk*.tk_commentStart[1] := '*';
    tk*.tk_commentEnd  [0] := '*'; tk*.tk_commentEnd  [1] := '/';
    tk*.tk_stringEscape := '\\';
    tk*.tk_commentsNest := true;
    tk*.tk_stringBreakAllowed := true;
    tk*.tk_reportNewLines := false;
    tk*.tk_preserveComments := true;
    tk*.tk_characterClass := ANSI_MAP;
    tk*.tk_currentChar := ' ';
    tk*.tk_nextChar := ' ';
corp;

/*
 * LexDeleteOperator - delete an operator from the operator table.
 */

proc LexDeleteOperator(/* *TokenizerState_t tk; register uint token */)void:
    *TokenizerState_t tk;
    ulong parToken;
    register *Operator_t op;
    register *byte p, q;
    register uint token, cnt;

    if sizeof(Operator_t) ~= 4 then
	/* compile time check */
	error("Operator_t is not 4 bytes long");
    fi;

    code (
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | R_A0,
	tk,
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D0,
	parToken
    );
    token := parToken;

    op := tk*.tk_operators;
    cnt := tk*.tk_operatorCount;
    while cnt ~= 0 and op*.op_code ~= token do
	cnt := cnt - 1;
	op := op + sizeof(Operator_t);
    od;
    if cnt ~= 0 then
	tk*.tk_operatorCount := tk*.tk_operatorCount - 1;
	p := pretend(op, *byte);
	q := p + sizeof(Operator_t);
	while
	    cnt := cnt - 1;
	    cnt ~= 0
	do
	    p* := q*;
	    p := p + 1;
	    q := q + 1;
	    p* := q*;
	    p := p + 1;
	    q := q + 1;
	    p* := q*;
	    p := p + 1;
	    q := q + 1;
	    p* := q*;
	    p := p + 1;
	    q := q + 1;
	od;
    fi;
corp;

/*
 * LexInsertOperator - insert an operator into the operator table.
 *	Note: the table is assumed to be large enough.
 */

proc LexInsertOperator(/* *TokenizerState_t tk; uint token;[2]char str */)void:
    *TokenizerState_t tk;
    ulong token;
    *[2] char pStr;
    register *Operator_t op;
    register *byte p, q;
    register uint cnt;

    if sizeof(Operator_t) ~= 4 then
	/* compile time check */
	error("Operator_t is not 4 bytes long");
    fi;

    code (
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | R_A0,
	tk,
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D0,
	token,
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | R_A1,
	pStr
    );

    cnt := tk*.tk_operatorCount + 1;
    tk*.tk_operatorCount := cnt;
    op := tk*.tk_operators;
    p := pretend(op, *byte) + cnt * sizeof(Operator_t);
    q := p - sizeof(Operator_t);
    if pStr*[1] = ' ' then
	/* insert it after all of the two-character operators */
	while cnt ~= 0 and op*.op_text[1] ~= ' ' do
	    op := op + sizeof(Operator_t);
	    cnt := cnt - 1;
	od;
    fi;
    while cnt ~= 0 do
	cnt := cnt - 1;
	p := p - 1;
	q := q - 1;
	p* := q*;
	p := p - 1;
	q := q - 1;
	p* := q*;
	p := p - 1;
	q := q - 1;
	p* := q*;
	p := p - 1;
	q := q - 1;
	p* := q*;
    od;
    op*.op_code := token;
    op*.op_text[0] := pStr*[0];
    op*.op_text[1] := pStr*[1];
corp;

/*
 * LexDeleteReservedWord - delete a reserved word reference from the
 *	reserved word table. No freeing, etc. of strings is done.
 */

proc LexDeleteReservedWord(/* *TokenizerState_t tk; uint token */)void:
    *TokenizerState_t tk;
    ulong parToken;
    register *ReservedWord_t rw;
    register *byte p, q;
    register uint token, cnt;

    if sizeof(ReservedWord_t) ~= 6 then
	error("ReservedWord_t is not 6 bytes long");
    fi;

    code (
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | R_A0,
	tk,
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D0,
	parToken
    );
    token := parToken;

    rw := tk*.tk_reservedWords;
    cnt := tk*.tk_reservedWordCount;
    while cnt ~= 0 and rw*.rw_code ~= token do
	rw := rw + sizeof(ReservedWord_t);
	cnt := cnt - 1;
    od;
    if cnt ~= 0 then
	tk*.tk_reservedWordCount := tk*.tk_reservedWordCount - 1;
	p := pretend(rw, *byte);
	q := p + sizeof(ReservedWord_t);
	while
	    cnt := cnt - 1;
	    cnt ~= 0
	do
	    p* := q*;
	    p := p + 1;
	    q := q + 1;
	    p* := q*;
	    p := p + 1;
	    q := q + 1;
	    p* := q*;
	    p := p + 1;
	    q := q + 1;
	    p* := q*;
	    p := p + 1;
	    q := q + 1;
	    p* := q*;
	    p := p + 1;
	    q := q + 1;
	    p* := q*;
	    p := p + 1;
	    q := q + 1;
	od;
    fi;
corp;

/*
 * LexInsertReservedWord - insert a new reserved word into the reserved word
 *	table. Note: the table is assumed to be large enough.
 */

proc LexInsertReservedWord(/* *TokenizerState_t tk; uint token;
			      *char str */)void:
    *TokenizerState_t tk;
    ulong token;
    *char str;
    register *ReservedWord_t rw;
    register *byte p, q;
    register *char cp1 @ p, cp2 @ q;
    register uint cnt;

    if sizeof(ReservedWord_t) ~= 6 then
	error("ReservedWord_t is not 6 bytes long");
    fi;

    code (
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | R_A0,
	tk,
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_DDIR << 3 | R_D0,
	token,
	OP_MOVEL | R_FP << 9 | M_DISP << 6 | M_ADIR << 3 | R_A1,
	str
    );

    cnt := tk*.tk_reservedWordCount;
    tk*.tk_reservedWordCount := cnt + 1;
    rw := tk*.tk_reservedWords;
    while
	if cnt = 0 then
	    false
	else
	    cp1 := rw*.rw_text;
	    cp2 := str;
	    while cp1* = cp2* and cp1* ~= '\e' do
		cp1 := cp1 + sizeof(char);
		cp2 := cp2 + sizeof(char);
	    od;
	    cp1* < cp2*
	fi
    do
	rw := rw + sizeof(ReservedWord_t);
	cnt := cnt - 1;
    od;
    p := pretend(tk*.tk_reservedWords, *byte) +
	    tk*.tk_reservedWordCount * sizeof(ReservedWord_t);
    q := p - sizeof(ReservedWord_t);
    while cnt ~= 0 do
	cnt := cnt - 1;
	p := p - 1;
	q := q - 1;
	p* := q*;
	p := p - 1;
	q := q - 1;
	p* := q*;
	p := p - 1;
	q := q - 1;
	p* := q*;
	p := p - 1;
	q := q - 1;
	p* := q*;
	p := p - 1;
	q := q - 1;
	p* := q*;
	p := p - 1;
	q := q - 1;
	p* := q*;
    od;
    rw*.rw_code := token;
    rw*.rw_text := str;
corp;
