/*
 * LexStandard.g -
 *	definitions and functions by which the Lex library provides a standard
 *	set of tokens structures to the caller.
 */

uint

    /* these values are understood by the standard declarers */

    tk_proc		= 100,
    tk_corp		= tk_proc + 1,
    tk_type		= tk_corp + 1,
    tk_struct		= tk_type + 1,
    tk_union		= tk_struct + 1,
    tk_enum		= tk_union + 1,
    tk_signed		= tk_enum + 1,
    tk_unsigned 	= tk_signed + 1,
    tk_unknown		= tk_unsigned + 1,
    tk_channel		= tk_unknown + 1,
    tk_input		= tk_channel + 1,
    tk_output		= tk_input + 1,
    tk_text		= tk_output + 1,
    tk_binary		= tk_text + 1,
    tk_import		= tk_binary + 1,
    tk_export		= tk_import + 1,
    tk_module		= tk_export + 1,
    tk_extern		= tk_module + 1,
    tk_private		= tk_extern + 1,
    tk_public		= tk_private + 1,
    tk_register 	= tk_public + 1,
    tk_chip		= tk_register + 1,

    /* these are provided for the normal set of types */

    tk_void		= 200,
    tk_bool		= tk_void + 1,
    tk_byte		= tk_bool + 1,
    tk_char		= tk_byte + 1,
    tk_string		= tk_char + 1,
    tk_ushort		= tk_string + 1,
    tk_short		= tk_ushort + 1,
    tk_uint		= tk_short + 1,
    tk_int		= tk_uint + 1,
    tk_ulong		= tk_int + 1,
    tk_long		= tk_ulong + 1,
    tk_float		= tk_long + 1,
    tk_arbptr		= tk_float + 1,

    /* these values are understood by the construct handlers */

    tk_if		= 300,
    tk_then		= tk_if + 1,
    tk_elif		= tk_then + 1,
    tk_else		= tk_elif + 1,
    tk_fi		= tk_else + 1,
    tk_while		= tk_fi + 1,
    tk_do		= tk_while + 1,
    tk_od		= tk_do + 1,
    tk_for		= tk_od + 1,
    tk_from		= tk_for + 1,
    tk_step		= tk_from + 1,
    tk_upto		= tk_step + 1,
    tk_downto		= tk_upto + 1,
    tk_case		= tk_downto + 1,
    tk_incase		= tk_case + 1,
    tk_default		= tk_incase + 1,
    tk_fallthrough	= tk_default + 1,
    tk_esac		= tk_fallthrough + 1,
    tk_read		= tk_esac + 1,
    tk_write		= tk_read + 1,
    tk_readln		= tk_write + 1,
    tk_writeln		= tk_readln + 1,
    tk_ioerror		= tk_writeln + 1,
    tk_open		= tk_ioerror + 1,
    tk_close		= tk_open + 1,
    tk_return		= tk_close + 1,
    tk_error		= tk_return + 1,
    tk_ignore		= tk_error + 1,
    tk_fix		= tk_ignore + 1,
    tk_flt		= tk_fix + 1,
    tk_make		= tk_flt + 1,
    tk_and		= tk_make + 1,
    tk_or		= tk_and + 1,
    tk_not		= tk_or + 1,
    tk_nil		= tk_not + 1,
    tk_dim		= tk_nil + 1,
    tk_sizeof		= tk_dim + 1,
    tk_range		= tk_sizeof + 1,
    tk_new		= tk_range + 1,
    tk_free		= tk_new + 1,

    /* these are provided for the normal set of operators */

    tk_plus		= 400,
    tk_minus		= tk_plus + 1,
    tk_star		= tk_minus + 1,
    tk_slash		= tk_star + 1,
    tk_percent		= tk_slash + 1,
    tk_shiftLeft	= tk_percent + 1,
    tk_shiftRight	= tk_shiftLeft + 1,
    tk_bitXor		= tk_shiftRight + 1,
    tk_bitAnd		= tk_bitXor + 1,
    tk_bitOr		= tk_bitAnd + 1,
    tk_bitNot		= tk_bitOr + 1,
    tk_assign		= tk_bitNot + 1,
    tk_lessThan 	= tk_assign + 1,
    tk_greaterThan	= tk_lessThan + 1,
    tk_equal		= tk_greaterThan + 1,
    tk_lessOrEqual	= tk_equal + 1,
    tk_greaterOrEqual	= tk_lessOrEqual + 1,
    tk_notEqual 	= tk_greaterOrEqual + 1,

    /* and the remaining few for various constructs */

    tk_dot		= 500,
    tk_dotDot		= tk_dot + 1,
    tk_colon		= tk_dotDot + 1,
    tk_openSquare	= tk_colon + 1,
    tk_closeSquare	= tk_openSquare + 1,
    tk_openBrace	= tk_closeSquare + 1,
    tk_closeBrace	= tk_openBrace + 1,
    tk_dollar		= tk_closeBrace + 1;

uint
    STANDARD_RESERVED_WORD_COUNT = 75,
    STANDARD_OPERATOR_COUNT = 30;

extern
    LexStandard(/* *TokenizerState_t tk */)void,
    LexDeleteOperator(/* *TokenizerState_t tk; uint token */)void,
    LexInsertOperator(/* *TokenizerState_t tk; uint token; [2]char str */)void,
    LexDeleteReservedWord(/**TokenizerState_t tk; uint token */)void,
    LexInsertReservedWord(/**TokenizerState_t tk; uint token;*char str */)void;
