#//Lex/Src/Lex.g
#//Symbol/Src/Symbol.g
#//Common/Types.g
#//Common/Nodes.g

#Parse.g

/*
 * ParseType - parse a type, and return a descriptor of it.
 */

proc ParseType(register *ParserState_t ps; *TypeValue_t pType)void:
    register *BuiltInList_t bil;
    register uint token;

    token := ps*.ps_tk.tk_kind;
    bil := ps*.ps_builtInTypes;
    while bil ~= nil do
	if bil*.bil_token = token then
	    pType*.ty_typeNo := bil*.bil_typeNo;
	    GetNextToken(&ps*.ps_tk);
	    return;
	fi;
	bil := bil*.bil_next;
    od;
    ps*.ps_stringError("unknown type");
    GetNextToken(&ps*.ps_tk);
    pType*.ty_typeNo := ty_error;
corp;

/*
 * Declare - declare variables, etc. of the given type.
 */

proc Declare(*ParserState_t ps; *TypeValue_t pType; **SymbolList_t pSyl;
	     uint terminator)void:
    register *SymbolList_t syl;
    register **SymbolList_t pTail;
    register bool noList;

    noList := pSyl = nil;
    if not noList then
	pTail := pSyl;
	while pTail* ~= nil do
	    pTail := &pTail**.syl_next;
	od;
    fi;
    while ps*.ps_tk.tk_kind ~= tk_semicolon and ps*.ps_tk.tk_kind ~= terminator
    do
	if ps*.ps_tk.tk_kind = tk_id then
	    if noList then
		ignore ps*.ps_idDeclarer(ps*.ps_tk.tk_buffer,
					 ps*.ps_tk.tk_bufferPosition, pType);
	    else
		syl := ps*.ps_alloc(sizeof(SymbolList_t));
		if syl ~= nil then
		    pTail* := syl;
		    pTail := &syl*.syl_next;
		    syl*.syl_this :=
			ps*.ps_idDeclarer(ps*.ps_tk.tk_buffer,
					  ps*.ps_tk.tk_bufferPosition, pType);
		fi;
	    fi;
	    GetNextToken(&ps*.ps_tk);
	    if ps*.ps_tk.tk_kind = tk_comma then
		GetNextToken(&ps*.ps_tk);
	    elif ps*.ps_tk.tk_kind ~= tk_semicolon and
		ps*.ps_tk.tk_kind ~= terminator
	    then
		ps*.ps_stringError("expecting separator in id list");
		while ps*.ps_tk.tk_kind ~= tk_id and
		    ps*.ps_tk.tk_kind ~= tk_comma and
		    ps*.ps_tk.tk_kind ~= tk_semicolon and
		    ps*.ps_tk.tk_kind ~= terminator
		do
		    GetNextToken(&ps*.ps_tk);
		od;
		if ps*.ps_tk.tk_kind = tk_comma then
		    GetNextToken(&ps*.ps_tk);
		fi;
	    fi;
	elif ps*.ps_tk.tk_kind = tk_comma then
	    ps*.ps_stringError("expecting id in declaration");
	fi;
    od;
    if not noList then
	pTail* := nil;
    fi;
corp;

/*
 * ParseProcHeader - parse a procedure header and return the completed
 *	structure for it. This can be used for actual procedure definitions,
 *	procedure declarations and procedure types. Note that in an actual
 *	definition or a declaration, the procedure name identifier is
 *	assumed to have already been consumed.
 */

proc ParseProcHeader(register *ParserState_t ps)*Procedure_t:
    register *Procedure_t pr;
    TypeValue_t typ;

    pr := ps*.ps_alloc(sizeof(Procedure_t));
    if pr = nil then
	return nil;
    fi;
    pr*.pr_parameters := nil;
    pr*.pr_locals := nil;
    pr*.pr_body := nil;
    if ps*.ps_tk.tk_kind = tk_openParen then
	GetNextToken(&ps*.ps_tk);
    else
	ps*.ps_stringError("expecting '(' to start procedure header");
    fi;
    while ps*.ps_tk.tk_kind ~= tk_closeParen and
	not ps*.ps_isNonDeclaration(ps*.ps_tk.tk_kind)
    do
	ParseType(ps, &typ);
	Declare(ps, &typ, &pr*.pr_parameters, tk_closeParen);
    od;
    if ps*.ps_tk.tk_kind = tk_closeParen then
	GetNextToken(&ps*.ps_tk);
    else
	ps*.ps_stringError("expecting ')' to end procedure header");
    fi;
    ParseType(ps, &pr*.pr_resultType);
    pr
corp;
