#//Lex/Src/Lex.g
#//Symbol/Src/Symbol.g
#//Common/Types.g
#//Common/Nodes.g

#Parse.g

#drinc:util.g

/*
 * AllocError - no space to build nodes, etc.
 */

proc AllocError(register *ParserState_t ps)void:

    if not ps*.ps_hadAllocError then
	ps*.ps_hadAllocError := true;
	ps*.ps_stringError("no space to allocate expression node");
    fi;
corp;

/*
 * ps_parseLevel - parse an expression at the given level of priority.
 */

proc ps_parseLevel(register *ParserState_t ps; *PriorityList_t prl)void:
    register *ConstructList_t conl;
    register *OperationList_t opl @ conl;
    register *ExpressionNode_t exn @ conl;
    register *Construct_t con;
    register *Operation_t oper @ con;
    *ExpressionNode_t leftHandSide;
    register uint token;

    token := ps*.ps_tk.tk_kind;
    if prl = nil then
	/* gone past highest level of operations. Try for constructs and
	   simple forms. We try for a construct first, so that the user
	   can override our default actions for these tokens. */
	conl := ps*.ps_constructs;
	while conl ~= nil and conl*.conl_this*.con_token ~= token do
	    conl := conl*.conl_next;
	od;
	if conl = nil then
	    exn := ps*.ps_alloc(sizeof(ExpressionNode_t));
	    if exn ~= nil then
		if token = tk_openParen then
		    exn*.exn_kind := exk_paren;
		    GetNextToken(&ps*.ps_tk);
		    ps_parseLevel(ps, ps*.ps_operations);
		    exn*.exn_.exu_paren := ps*.ps_currentNode;
		    if ps*.ps_currentNode ~= nil then
			exn*.exn_type.ty_typeNo :=
			    ps*.ps_currentNode*.exn_type.ty_typeNo;
		    else
			exn*.exn_type.ty_typeNo := ty_unknown;
		    fi;
		    if ps*.ps_tk.tk_kind = tk_closeParen then
			GetNextToken(&ps*.ps_tk);
		    else
			ps*.ps_stringError("missing close parenthesis");
		    fi;
		else
		    case token
		    incase tk_ulongConst:
			exn*.exn_kind := exk_ulong;
			exn*.exn_type.ty_typeNo := ty_ulong;
			exn*.exn_.exu_ulong := ps*.ps_tk.tk_ulongValue;
			GetNextToken(&ps*.ps_tk);
		    incase tk_charConst:
			exn*.exn_kind := exk_char;
			exn*.exn_type.ty_typeNo := ty_char;
			exn*.exn_.exu_char := ps*.ps_tk.tk_charValue;
			GetNextToken(&ps*.ps_tk);
		    incase tk_stringConst:
		    incase tk_comment:
			exn*.exn_.exu_string.exs_string :=
			    ps*.ps_alloc(ps*.ps_tk.tk_bufferPosition);
			if exn*.exn_.exu_string.exs_string = nil then
			    ps*.ps_free(exn, sizeof(ExpressionNode_t));
			    exn := nil;
			    AllocError(ps);
			else
			    if token = tk_stringConst then
				exn*.exn_kind := exk_string;
				exn*.exn_type.ty_typeNo := ty_string;
			    else
				exn*.exn_kind := exk_comment;
				exn*.exn_type.ty_typeNo := ty_unknown;
			    fi;
			    BlockCopy(exn*.exn_.exu_string.exs_string,
				      ps*.ps_tk.tk_buffer,
				      ps*.ps_tk.tk_bufferPosition);
			    exn*.exn_.exu_string.exs_length :=
				ps*.ps_tk.tk_bufferPosition;
			fi;
			GetNextToken(&ps*.ps_tk);
		    incase tk_floatConst:
			exn*.exn_kind := exk_float;
			exn*.exn_type.ty_typeNo := ty_float;
			exn*.exn_.exu_float := ps*.ps_tk.tk_floatValue;
			GetNextToken(&ps*.ps_tk);
		    incase tk_id:
			exn*.exn_kind := exk_id;
			exn*.exn_.exu_id :=
			    ps*.ps_idHandler(ps*.ps_tk.tk_buffer,
					     ps*.ps_tk.tk_bufferPosition);
			if exn*.exn_.exu_id ~= nil then
			    exn*.exn_type.ty_typeNo :=
				exn*.exn_.exu_id*.sy_type.syu_typeNo;
			else
			    exn*.exn_type.ty_typeNo := ty_error;
			fi;
			GetNextToken(&ps*.ps_tk);
		    default:
			ps*.ps_stringError("unidentified syntax error");
			ps*.ps_currentNode := nil;
			exn*.exn_kind := exk_error;
			exn*.exn_type.ty_typeNo := ty_error;
			GetNextToken(&ps*.ps_tk);
		    esac;
		fi;
	    else
		AllocError(ps);
	    fi;
	    ps*.ps_currentNode := exn;
	else
	    con := conl*.conl_this;
	    con*.con_builder(ps, con);
	fi;
    else
	opl := prl*.prl_operations;
	while
	    if opl = nil then
		false
	    else
		oper := opl*.opl_this;
		oper*.oper_token ~= token or oper*.oper_kind ~= opk_prefix
	    fi
	do
	    opl := opl*.opl_next;
	od;
	if opl ~= nil then
	    /* there is a prefix operation at this priority level */
	    token := ps*.ps_tk.tk_kind;
	    GetNextToken(&ps*.ps_tk);
	    ps_parseLevel(ps, prl*.prl_next);
	    exn := ps*.ps_alloc(sizeof(ExpressionNode_t));
	    if exn ~= nil then
		exn*.exn_kind := exk_prefix;
		exn*.exn_.exu_un.exu_operation := oper;
		exn*.exn_.exu_un.exu_operand := ps*.ps_currentNode;
		ps*.ps_currentNode := exn;
		oper*.oper_checker2(ps, token);
	    else
		AllocError(ps);
	    fi;
	else
	    ps_parseLevel(ps, prl*.prl_next);
	    token := ps*.ps_tk.tk_kind;
	    opl := prl*.prl_operations;
	    while
		if opl = nil then
		    false
		else
		    oper := opl*.opl_this;
		    oper*.oper_token ~= token or oper*.oper_kind = opk_prefix
		fi
	    do
		opl := opl*.opl_next;
	    od;
	    if opl ~= nil then
		if oper*.oper_kind = opk_postfix then
		    /* found a postfix operation to apply */
		    exn := ps*.ps_alloc(sizeof(ExpressionNode_t));
		    if exn ~= nil then
			exn*.exn_kind := exk_postfix;
			exn*.exn_.exu_un.exu_operation := oper;
			exn*.exn_.exu_un.exu_operand := ps*.ps_currentNode;
			ps*.ps_currentNode := exn;
			oper*.oper_checker2(ps, token);
		    else
			AllocError(ps);
		    fi;
		    GetNextToken(&ps*.ps_tk);
		else
		    /* found a binary operation to apply */
		    oper*.oper_checker1(ps, token);
		    leftHandSide := ps*.ps_currentNode;
		    token := ps*.ps_tk.tk_kind;
		    GetNextToken(&ps*.ps_tk);
		    ps_parseLevel(ps, prl*.prl_next);
		    exn := ps*.ps_alloc(sizeof(ExpressionNode_t));
		    if exn ~= nil then
			exn*.exn_kind := exk_binary;
			exn*.exn_.exu_bi.exb_operation := oper;
			exn*.exn_.exu_bi.exb_left := leftHandSide;
			exn*.exn_.exu_bi.exb_right := ps*.ps_currentNode;
			ps*.ps_currentNode := exn;
			oper*.oper_checker2(ps, token);
		    else
			/* just leave the RHS there */
			AllocError(ps);
			DestroyTree(ps, leftHandSide);
		    fi;
		fi;
	    fi;
	fi;
    fi;
corp;

/*
 * ParseExpression - top level entry point for parser routines. Will return,
 *	pointed to by ps_currentNode, parse tree of what it finds.
 */

proc ParseExpression(*ParserState_t ps)void:

    ps_parseLevel(ps, ps*.ps_operations);
corp;

/*
 * ps_appendStatement - local routine to add a node to a statement list.
 */

proc ps_appendStatement(register *ParserState_t ps; *ExpressionNode_t exn;
			***ExpressionNodeList_t pTail)void:
    register *ExpressionNodeList_t exnl;

    if pTail* ~= nil then
	exnl := ps*.ps_alloc(sizeof(ExpressionNodeList_t));
	if exnl ~= nil then
	    exnl*.exnl_this := exn;
	    pTail** := exnl;
	    pTail* := &exnl*.exnl_next;
	else
	    AllocError(ps);
	fi;
    fi;
corp;

/*
 * ps_handleComments - add any comments to the statement list being built.
 */

proc ps_handleComments(register *ParserState_t ps;
		       register ***ExpressionNodeList_t pTail)void:

    while ps*.ps_tk.tk_kind = tk_comment do
	ps_parseLevel(ps, nil);
	ps_appendStatement(ps, ps*.ps_currentNode, pTail);
    od;
corp;

/*
 * ParseStatementList - parse a statement list. Yields a exk_list node.
 */

proc ParseStatementList(register *ParserState_t ps)void:
    register *ExpressionNode_t exn, exnTemp;
    **ExpressionNodeList_t exnlTail;

    exnlTail := nil;
    exn := ps*.ps_alloc(sizeof(ExpressionNode_t));
    if exn ~= nil then
	exn*.exn_kind := exk_list;
	exnlTail := &exn*.exn_.exu_list;
    else
	AllocError(ps);
    fi;
    while
	ps_handleComments(ps, &exnlTail);
	if ps*.ps_isTerminator(ps*.ps_tk.tk_kind) or
	    ps*.ps_tk.tk_kind = tk_semicolon
	then
	    exnTemp := nil;
	else
	    ps_parseLevel(ps, ps*.ps_operations);
	    exnTemp := ps*.ps_currentNode;
	    ps_handleComments(ps, &exnlTail);
	fi;
	ps_appendStatement(ps, exnTemp, &exnlTail);
	ps*.ps_tk.tk_kind = tk_semicolon
    do
	if exnTemp ~= nil and
	    exnTemp*.exn_type.ty_typeNo ~= ty_void and
	    exnTemp*.exn_type.ty_typeNo ~= ty_error and
	    exnTemp*.exn_type.ty_typeNo ~= ty_unknown
	then
	    ps*.ps_stringError("statements must be of type void");
	fi;
	GetNextToken(&ps*.ps_tk);
    od;
    if exnlTail ~= nil then
	exnlTail* := nil;
	if exn*.exn_.exu_list = nil then
	    ps*.ps_free(exn, sizeof(ExpressionNode_t));
	    exn := nil;
	elif exn*.exn_.exu_list*.exnl_next = nil then
	    exnTemp := exn*.exn_.exu_list*.exnl_this;
	    ps*.ps_free(exn*.exn_.exu_list, sizeof(ExpressionNodeList_t));
	    ps*.ps_free(exn, sizeof(ExpressionNode_t));
	    exn := exnTemp;
	else
	    if exnTemp = nil then
		/* sequence ended in a statement separator */
		exn*.exn_type.ty_typeNo := ty_void;
	    else
		exn*.exn_type.ty_typeNo := exnTemp*.exn_type.ty_typeNo;
	    fi;
	fi;
    fi;
    ps*.ps_currentNode := exn;
corp;

/*
 * DestroyTree - a standard tree destructor routine.
 */

proc DestroyTree(register *ParserState_t ps;
		 register *ExpressionNode_t exn)void:
    register *ExpressionNodeList_t exnl, exnlTemp;

    if exn ~= nil then
	case exn*.exn_kind
	incase exk_prefix:
	incase exk_postfix:
	    DestroyTree(ps, exn*.exn_.exu_un.exu_operand);
	incase exk_binary:
	    DestroyTree(ps, exn*.exn_.exu_bi.exb_right);
	    DestroyTree(ps, exn*.exn_.exu_bi.exb_left);
	incase exk_special:
	    exn*.exn_.exu_con.exc_construct*.con_destructor(ps, exn);
	incase exk_string:
	incase exk_comment:
	    ps*.ps_free(exn*.exn_.exu_string.exs_string,
			exn*.exn_.exu_string.exs_length);
	incase exk_list:
	    exnl := exn*.exn_.exu_list;
	    while exnl ~= nil do
		exnlTemp := exnl;
		DestroyTree(ps, exnl*.exnl_this);
		exnl := exnl*.exnl_next;
		ps*.ps_free(exnlTemp, sizeof(ExpressionNodeList_t));
	    od;
	incase exk_paren:
	    DestroyTree(ps, exn*.exn_.exu_paren);
	esac;
	ps*.ps_free(exn, sizeof(ExpressionNode_t));
    fi;
corp;

/*
 * Assignable - this checker-type routine is here because of the include
 *	of Symbol/Src/Symbol.g
 */

proc Assignable(*ExpressionNode_t exn)bool:
    register uint kind;

    if exn = nil then
	return(true);
    fi;
    if exn*.exn_kind = exk_id then
	if exn*.exn_.exu_id = nil then
	    return(true);
	fi;
	kind := exn*.exn_.exu_id*.sy_kind;
	return(kind = syk_globalVariable or kind = syk_fileVariable or
	       kind = syk_modulePublic or kind = syk_modulePrivate or
	       kind = syk_localVariable or kind = syk_parameter or
	       kind = syk_systemVar or kind = syk_absoluteVar or
	       kind = syk_externVar);
    fi;
    false
corp;
