/*
 * Parse.g - definitions for creating a parser.
 */

type
    ConstructList_t = struct {
	*ConstructList_t conl_next;
	*Construct_t conl_this;
    },

    OperationList_t = struct {
	*OperationList_t opl_next;
	*Operation_t opl_this;
    },

    PriorityList_t = struct {
	*PriorityList_t prl_next;
	*OperationList_t prl_operations;
    },

    BuiltInList_t = struct {
	*BuiltInList_t bil_next;
	uint bil_token;
	uint bil_typeNo;
    },

    ParserState_t = struct {
	TokenizerState_t ps_tk;
	proc(*char message)void ps_stringError;
	proc(*char message1, message2, message3)void ps_stringError3;
	proc(ulong length)arbptr ps_alloc;
	proc(arbptr p; ulong length)void ps_free;
	proc(*char name; ulong nameLength;
	     *TypeValue_t pType)*Symbol_t ps_idDeclarer;
	proc(*char name; ulong nameLength)*Symbol_t ps_idHandler;
	proc(uint token)bool ps_isNonDeclaration;
	proc(uint token)bool ps_isTerminator;
	proc(*ParserState_t ps; *TypeValue_t pOldType;
	     **ExpressionNode_t newNode)bool ps_ifCaseCompatible;
	proc(*ExpressionNode_t exn)bool ps_assignable;
	proc(*ExpressionNode_t exn; *char format)bool ps_checkIO;

	*ConstructList_t ps_constructs;
	*PriorityList_t ps_operations;
	*BuiltInList_t ps_builtInTypes;

	*ExpressionNode_t ps_currentNode;
	bool ps_hadAllocError;

	arbptr ps_userData;		/* whatever a user may want */
    };

extern
    AllocError(*ParserState_t ps)void,
    ParseExpression(*ParserState_t ps)void,
    ParseStatementList(*ParserState_t ps)void,
    DestroyTree(*ParserState_t ps; *ExpressionNode_t exn)void,
    Assignable(*ExpressionNode_t exn)bool,

    NumericLeftTester(*ParserState_t ps; uint token)bool,
    NumericLeftChecker(*ParserState_t ps; uint token)void,
    NumericRightTester(*ParserState_t ps; uint token)bool,
    NumericRightChecker(*ParserState_t ps; uint token)void,
    NumericPrefixTester(*ParserState_t ps; uint token)bool,
    NumericPrefixChecker(*ParserState_t ps; uint token)void,
    BoolLeftTester(*ParserState_t ps; uint token)bool,
    BoolLeftChecker(*ParserState_t ps; uint token)void,
    BoolRightTester(*ParserState_t ps; uint token)bool,
    BoolRightChecker(*ParserState_t ps; uint token)void,
    BoolPrefixTester(*ParserState_t ps; uint token)bool,
    BoolPrefixChecker(*ParserState_t ps; uint token)void,
    AssignLeftTester(*ParserState_t ps; uint token)bool,
    AssignLeftChecker(*ParserState_t ps; uint token)void,
    AssignRightTester(*ParserState_t ps; uint token)bool,
    AssignRightChecker(*ParserState_t ps; uint token)void,
    CheckIO(*ExpressionNode_t exn; *char pFormat)bool,
    IfCaseCompatible(*ParserState_t ps; *TypeValue_t pOldType;
		     **ExpressionNode_t pNewNode)bool,

    BuildIf(*ParserState_t ps; *Construct_t con)void,
    DestroyIf(*ParserState_t ps; *ExpressionNode_t exn)void,
    BuildWhile(*ParserState_t ps; *Construct_t con)void,
    DestroyWhile(*ParserState_t ps; *ExpressionNode_t exn)void,
    BuildFor(*ParserState_t ps; *Construct_t con)void,
    DestroyFor(*ParserState_t ps; *ExpressionNode_t exn)void,
    BuildCase(*ParserState_t ps; *Construct_t con)void,
    DestroyCase(*ParserState_t ps; *ExpressionNode_t exn)void,
    BuildIO(*ParserState_t ps; *Construct_t con)void,
    DestroyIO(*ParserState_t ps; *ExpressionNode_t exn)void,
    ConstructTerminator(uint token)bool,
    IsNonDeclaration(uint token)bool,

    ParseType(*ParserState_t ps; *TypeValue_t pType)void,
    Declare(*ParserState_t ps; *TypeValue_t pType; **SymbolList_t pSyl;
	    uint terminator)void,
    ParseProcHeader(*ParserState_t ps)*Procedure_t;
