#//Lex/Src/Lex.g
#//Lex/Src/LexStandard.g
#//Common/Types.g
#//Common/Nodes.g
#//Common/NodeTypes.g

#Parse.g

/*
 * NumericLeftTester - a standard numeric type compatibility tester.
 */

proc NumericLeftTester(*ParserState_t ps; uint token)bool:
    register *ExpressionNode_t exn;
    register ulong typeNo;

    exn := ps*.ps_currentNode;
    if exn = nil then
	return(true);
    fi;
    typeNo := exn*.exn_type.ty_typeNo;
    if (token = tk_shiftLeft or token = tk_shiftRight) and
	typeNo ~= ty_ulong and typeNo ~= ty_uint and typeNo ~= ty_ushort
    then
	return(false);
    fi;
    typeNo = ty_error or typeNo = ty_float or
	typeNo = ty_ushort or typeNo = ty_short or
	typeNo = ty_uint or typeNo = ty_int or
	typeNo = ty_ulong or typeNo = ty_long
corp;

/*
 * NumericLeftChecker - call the above and complain if it fails.
 */

proc NumericLeftChecker(*ParserState_t ps; uint token)void:
    register *ExpressionNode_t exn;

    if not NumericLeftTester(ps, token) then
	if token = tk_shiftLeft or token = tk_shiftRight then
	    ps*.ps_stringError("unsigned left operand required for shift");
	else
	    ps*.ps_stringError("numeric left operand required for binary");
	fi;
	exn := ps*.ps_currentNode;
	if exn ~= nil then
	    exn*.exn_type.ty_typeNo := ty_error;
	fi;
    fi;
corp;

/*
 * insertConversion - insert a conversion node into a tree.
 */

proc insertConversion(*ParserState_t ps; **ExpressionNode_t pExn;
		      int nodeCode, nodeType)void:
    register *ExpressionNode_t exnNew;

    exnNew := ps*.ps_alloc(sizeof(ExpressionNode_t));
    if exnNew = nil then
	AllocError(ps);
	return;
    fi;
    exnNew*.exn_kind := exk_prefix;
    exnNew*.exn_code := nodeCode;
    exnNew*.exn_type.ty_typeNo := nodeType;
    exnNew*.exn_.exu_un.exu_operation := nil;
    exnNew*.exn_.exu_un.exu_operand := pExn*;
    pExn* := exnNew;
corp;

/*
 * expandToLong - local routine to expand a short or int node to a long,
 *	paying attention to any signedness.
 */

proc expandToLong(*ParserState_t ps; **ExpressionNode_t pExn)void:
    *ExpressionNode_t exn;

    exn := pExn*;
    if exn = nil then
	return;
    fi;
    if exn*.exn_type.ty_typeNo = ty_ushort then
	insertConversion(ps, &exn, exp_ushortToUint, ty_uint);
    elif exn*.exn_type.ty_typeNo = ty_short then
	insertConversion(ps, &exn, exp_shortToInt, ty_int);
    fi;
    if exn*.exn_type.ty_typeNo = ty_uint then
	insertConversion(ps, &exn, exp_uintToUlong, ty_ulong);
    elif exn*.exn_type.ty_typeNo = ty_int then
	insertConversion(ps, &exn, exp_intToLong, ty_long);
    fi;
    pExn* := exn;
corp;

/*
 * expandToInt - similar to above.
 */

proc expandToInt(*ParserState_t ps; **ExpressionNode_t pExn)void:
    *ExpressionNode_t exn;

    exn := pExn*;
    if exn = nil then
	return;
    fi;
    if exn*.exn_type.ty_typeNo = ty_ushort then
	insertConversion(ps, pExn, exp_ushortToUint, ty_uint);
    elif exn*.exn_type.ty_typeNo = ty_short then
	insertConversion(ps, pExn, exp_shortToInt, ty_int);
    fi;
corp;

/*
 * NumericRightTester - tester/fixer for binary numeric operations.
 */

proc NumericRightTester(*ParserState_t ps; uint token)bool:
    register *ExpressionNode_t exn;
    *ExpressionNode_t exnLeft, exnRight;
    register ulong typeLeft, typeRight;
    register uint eCode, eType;
    bool leftArb, rightArb;

    exn := ps*.ps_currentNode;
    if exn = nil then
	return(true);
    fi;
    eType := ty_error;
    case token
    incase tk_plus:
	eCode := exp_ulongPlus;
    incase tk_minus:
	eCode := exp_ulongMinus;
    incase tk_star:
	eCode := exp_ulongTimes;
    incase tk_slash:
	eCode := exp_ulongDivide;
    incase tk_percent:
	eCode := exp_ulongRemainder;
    incase tk_bitAnd:
	eCode := exp_ulongAnd;
    incase tk_bitOr:
	eCode := exp_ulongOr;
    incase tk_bitXor:
	eCode := exp_ulongXor;
    incase tk_shiftLeft:
	eCode := exp_ulongShiftLeft;
    incase tk_shiftRight:
	eCode := exp_ulongShiftRight;
    incase tk_lessThan:
	eCode := exp_ulongLessThan;
	eType := ty_bool;
    incase tk_lessOrEqual:
	eCode := exp_ulongLessOrEqual;
	eType := ty_bool;
    incase tk_greaterThan:
	eCode := exp_ulongGreaterThan;
	eType := ty_bool;
    incase tk_greaterOrEqual:
	eCode := exp_ulongGreaterOrEqual;
	eType := ty_bool;
    incase tk_equal:
	eCode := exp_ulongEqual;
	eType := ty_bool;
    incase tk_notEqual:
	eCode := exp_ulongNotEqual;
	eType := ty_bool;
    default:
	eCode := exp_error;
    esac;
    exn*.exn_code := eCode;
    exn*.exn_type.ty_typeNo := eType;
    exnLeft := exn*.exn_.exu_bi.exb_left;
    if exnLeft = nil then
	typeLeft := ty_unknown;
    else
	typeLeft := exnLeft*.exn_type.ty_typeNo;
    fi;
    exnRight := exn*.exn_.exu_bi.exb_right;
    if exnRight = nil then
	typeRight := ty_unknown;
    else
	typeRight := exnRight*.exn_type.ty_typeNo;
    fi;
    leftArb := typeLeft = ty_error or typeLeft = ty_unknown;
    rightArb := typeRight = ty_error or typeRight = ty_unknown;
    if typeLeft = ty_float or typeRight = ty_float then
	if eType ~= ty_bool then
	    exn*.exn_type.ty_typeNo := ty_float;
	fi;
	if eCode >= exp_ulongLessThan then
	    eCode := eCode + (exp_floatLessThan - exp_ulongLessThan);
	else
	    eCode := eCode + (exp_floatPlus - exp_ulongPlus);
	fi;
	exn*.exn_code := eCode;
	if typeLeft = ty_float then
	    return(typeRight = ty_float or rightArb);
	fi;
	if typeRight = ty_float then
	    return(leftArb);
	fi;
    fi;
    if typeLeft = ty_ulong or typeLeft = ty_long or
	typeRight = ty_ulong or typeRight = ty_long
    then
	if typeLeft = ty_long or typeRight = ty_long then
	    /* do signed operation */
	    if eCode >= exp_ulongLessThan then
		eCode := eCode - (exp_ulongLessThan - exp_longLessThan);
	    else
		eCode := eCode - (exp_ulongPlus - exp_longPlus);
	    fi;
	fi;
	exn*.exn_code := eCode;
	if typeLeft = ty_ulong or typeLeft = ty_long then
	    if typeRight = ty_ulong or typeRight = ty_long or
		typeRight = ty_uint or typeRight = ty_int or
		typeRight = ty_ushort or typeRight = ty_short
	    then
		expandToLong(ps, &exnRight);
		if exnRight ~= nil and exnRight*.exn_type.ty_typeNo = ty_long
		then
		    typeLeft := ty_long;
		fi;
		exn*.exn_.exu_bi.exb_right := exnRight;
		if eType ~= ty_bool then
		    exn*.exn_type.ty_typeNo := typeLeft;
		fi;
		return(true);
	    fi;
	    if eType ~= ty_bool then
		exn*.exn_type.ty_typeNo := typeLeft;
	    fi;
	    return(rightArb);
	else
	    if typeLeft = ty_uint or typeLeft = ty_int or
		typeLeft = ty_ushort or typeLeft = ty_short
	    then
		expandToLong(ps, &exnLeft);
		if exnLeft ~= nil and exnLeft*.exn_type.ty_typeNo = ty_long
		then
		    typeRight := ty_long;
		fi;
		exn*.exn_.exu_bi.exb_left := exnLeft;
		if eType ~= ty_bool then
		    exn*.exn_type.ty_typeNo := typeRight;
		fi;
		return(true);
	    fi;
	    if eType ~= ty_bool then
		exn*.exn_type.ty_typeNo := typeRight;
	    fi;
	    return(leftArb);
	fi;
    fi;
    if typeLeft = ty_uint or typeLeft = ty_int or
	typeRight = ty_uint or typeRight = ty_int
    then
	if typeLeft = ty_int or typeRight = ty_int then
	    /* do signed operation */
	    if eCode >= exp_uintLessThan then
		eCode := eCode + (exp_intLessThan - exp_ulongLessThan);
	    else
		eCode := eCode + (exp_intPlus - exp_ulongPlus);
	    fi;
	else
	    /* do unsigned operation */
	    eCode := eCode + (exp_uintNegate - exp_ulongNegate);
	fi;
	exn*.exn_code := eCode;
	if typeLeft = ty_uint or typeLeft = ty_int then
	    if typeRight = ty_uint or typeRight = ty_int or
		typeRight = ty_ushort or typeRight = ty_short
	    then
		expandToInt(ps, &exnRight);
		if exnRight ~= nil and exnRight*.exn_type.ty_typeNo = ty_int
		then
		    typeLeft := ty_int;
		fi;
		exn*.exn_.exu_bi.exb_right := exnRight;
		if eType ~= ty_bool then
		    exn*.exn_type.ty_typeNo := typeLeft;
		fi;
		return(true);
	    fi;
	    if eType ~= ty_bool then
		exn*.exn_type.ty_typeNo := typeLeft;
	    fi;
	    return(rightArb);
	fi;
	if typeRight = ty_uint or typeRight = ty_int then
	    if typeLeft = ty_ushort or typeLeft = ty_short then
		expandToInt(ps, &exnLeft);
		if exnLeft ~=nil and exnLeft*.exn_type.ty_typeNo = ty_int
		then
		    typeRight := ty_int;
		fi;
		exn*.exn_.exu_bi.exb_left := exnLeft;
		if eType ~= ty_bool then
		    exn*.exn_type.ty_typeNo := typeRight;
		fi;
		return(true);
	    fi;
	    if eType ~= ty_bool then
		exn*.exn_type.ty_typeNo := typeRight;
	    fi;
	    return(leftArb);
	fi;
    fi;
    if typeLeft = ty_ushort or typeLeft = ty_short or
	typeRight = ty_ushort or typeRight = ty_short
    then
	if typeLeft = ty_short or typeRight = ty_short then
	    /* do signed operation */
	    if eCode >= exp_ushortLessThan then
		eCode := eCode + (exp_shortLessThan - exp_ulongLessThan);
	    else
		eCode := eCode + (exp_shortPlus - exp_ulongPlus);
	    fi;
	else
	    /* do unsigned operation */
	    eCode := eCode + (exp_ushortNegate - exp_ulongNegate);
	fi;
	exn*.exn_code := eCode;
	if typeLeft = ty_ushort or typeLeft = ty_short then
	    if eType ~= ty_bool then
		exn*.exn_type.ty_typeNo := typeLeft;
	    fi;
	    return(rightArb);
	fi;
	if typeRight = ty_ushort or typeRight = ty_short then
	    if eType ~= ty_bool then
		exn*.exn_type.ty_typeNo := typeRight;
	    fi;
	    return(leftArb);
	fi;
    fi;
    exn*.exn_code := exp_error;
    if typeLeft = ty_error and typeRight = ty_error then
	if eType ~= ty_bool then
	    exn*.exn_type.ty_typeNo := ty_error;
	fi;
	return(true);
    fi;
    if typeLeft = ty_unknown and typeRight = ty_unknown then
	if eType ~= ty_bool then
	    exn*.exn_type.ty_typeNo := ty_unknown;
	fi;
	return(true);
    fi;
    if eType ~= ty_bool then
	exn*.exn_type.ty_typeNo := ty_error;
    fi;
    false
corp;

/*
 * NumericRightChecker - complain if the above fails.
 */

proc NumericRightChecker(*ParserState_t ps; uint token)void:

    if not NumericRightTester(ps, token) then
	ps*.ps_stringError("numeric right operand required for binary");
    fi;
corp;

/*
 * NumericPrefixTester - check operand for numeric prefix operation.
 */

proc NumericPrefixTester(*ParserState_t ps; uint token)bool:
    register *ExpressionNode_t exn, exnUpper;
    register uint typ, eCode;

    exnUpper := ps*.ps_currentNode;
    if exnUpper = nil then
	return(true);
    fi;
    exn := exnUpper*.exn_.exu_un.exu_operand;
    if exn = nil then
	exnUpper*.exn_type.ty_typeNo := ty_unknown;
	exnUpper*.exn_code := exp_error;
	return(true);
    fi;
    typ := exn*.exn_type.ty_typeNo;
    exnUpper*.exn_type.ty_typeNo := typ;
    case token
    incase tk_plus:
	eCode := exp_longUnPlus;
    incase tk_minus:
	eCode := exp_longNegate;
    incase tk_bitOr:
	eCode := exp_longAbs;
    incase tk_bitNot:
	eCode := exp_longNot;
    default:
	eCode := exp_error;
    esac;
    if typ = ty_long then
	exn*.exn_code := eCode;
	return(true);
    fi;
    if typ = ty_ulong then
	exn*.exn_code := eCode + (exp_ulongUnPlus - exp_longUnPlus);
	return(true);
    fi;
    if typ = ty_int then
	exn*.exn_code := eCode + (exp_intUnPlus - exp_longUnPlus);
	return(true);
    fi;
    if typ = ty_uint then
	exn*.exn_code := eCode + (exp_uintUnPlus - exp_longUnPlus);
	return(true);
    fi;
    if typ = ty_short then
	exn*.exn_code := eCode + (exp_shortUnPlus - exp_longUnPlus);
	return(true);
    fi;
    if typ = ty_ushort then
	exn*.exn_code := eCode + (exp_ushortUnPlus - exp_longUnPlus);
	return(true);
    fi;
    if typ = ty_float and eCode ~= exp_longNot then
	exn*.exn_code := eCode + (exp_floatUnPlus - exp_longUnPlus);
	return(true);
    fi;
    exn*.exn_code := eCode;
    false
corp;

/*
 * NumericPrefixChecker - complain if the above fails.
 */

proc NumericPrefixChecker(*ParserState_t ps; uint token)void:

    if not NumericPrefixTester(ps, token) then
	ps*.ps_stringError("numeric operand required for prefix operator");
    fi;
corp;

/*
 * BoolLeftTester - check for a bool operand.
 */

proc BoolLeftTester(*ParserState_t ps; uint token)bool:
    register *ExpressionNode_t exn;
    register uint typ;

    exn := ps*.ps_currentNode;
    if exn = nil then
	return(true);
    fi;
    typ := exn*.exn_type.ty_typeNo;
    typ = ty_bool or typ = ty_error or typ = ty_unknown
corp;

/*
 * BoolLeftChecker - complain if the above fails.
 */

proc BoolLeftChecker(*ParserState_t ps; uint token)void:

    if not BoolLeftTester(ps, token) then
	ps*.ps_stringError("'bool' left operand required for binary");
    fi;
corp;

/*
 * BoolRightTester - simple tester for boolean operands.
 */

proc BoolRightTester(*ParserState_t ps; uint token)bool:
    register *ExpressionNode_t exn, exnArg;
    register ulong typeLeft, typeRight;

    exn := ps*.ps_currentNode;
    if exn = nil then
	return(true);
    fi;
    exnArg := exn*.exn_.exu_bi.exb_left;
    if exnArg = nil then
	typeLeft := ty_unknown;
    else
	typeLeft := exnArg*.exn_type.ty_typeNo;
    fi;
    exnArg := exn*.exn_.exu_bi.exb_right;
    if exnArg = nil then
	typeRight := ty_unknown;
    else
	typeRight := exnArg*.exn_type.ty_typeNo;
    fi;
    if token = tk_and then
	exn*.exn_code := exp_boolAnd;
    elif token = tk_or then
	exn*.exn_code := exp_boolOr;
    elif token = tk_equal then
	exn*.exn_code := exp_boolEqual;
    elif token = tk_notEqual then
	exn*.exn_code := exp_boolNotEqual;
    else
	exn*.exn_code := exp_error;
    fi;
    exn*.exn_type.ty_typeNo := ty_bool;
    if typeLeft = ty_bool then
	return(typeRight = ty_bool or
	       typeRight = ty_error or typeRight = ty_unknown);
    fi;
    if typeRight = ty_bool then
	return(typeLeft = ty_error or typeLeft = ty_unknown);
    fi;
    if typeLeft = ty_error and typeRight = ty_error then
	return(true);
    fi;
    if typeLeft = ty_unknown and typeRight = ty_unknown then
	return(true);
    fi;
    false
corp;

/*
 * BoolRightChecker - complain if the above fails.
 */

proc BoolRightChecker(*ParserState_t ps; uint token)void:

    if not BoolRightTester(ps, token) then
	ps*.ps_stringError("'bool' right operand required for binary");
    fi;
corp;

/*
 * BoolPrefixTester - test for a boolean operand to a unary operation.
 */

proc BoolPrefixTester(*ParserState_t ps; uint token)bool:
    register *ExpressionNode_t exn;
    register uint typ;

    exn := ps*.ps_currentNode;
    if exn = nil then
	return(true);
    fi;
    exn*.exn_type.ty_typeNo := ty_bool;
    exn*.exn_code := exp_boolNot;
    exn := exn*.exn_.exu_un.exu_operand;
    if exn = nil then
	return(true);
    fi;
    typ := exn*.exn_type.ty_typeNo;
    typ = ty_bool or typ = ty_error or typ = ty_unknown
corp;

/*
 * BoolPrefixChecker - complain if the above fails.
 */

proc BoolPrefixChecker(*ParserState_t ps; uint token)void:

    if not BoolPrefixTester(ps, token) then
	ps*.ps_stringError("'bool' operand required for unary");
    fi;
corp;

/*
 * AssignRightTester - check a numeric, string or bool assignment,
 *	insert any coercions that are needed, return 'true' if fully
 *	handled, else false.
 */

proc AssignRightTester(*ParserState_t ps; uint token)bool:
    register *ExpressionNode_t exn;
    *ExpressionNode_t exnLeft, exnRight;
    register ulong typeLeft, typeRight;
    register bool leftArb, rightArb;

    exn := ps*.ps_currentNode;
    if exn = nil then
	return(true);
    fi;
    exn*.exn_type.ty_typeNo := ty_void;
    exnLeft := exn*.exn_.exu_bi.exb_left;
    if exnLeft = nil then
	typeLeft := ty_unknown;
    else
	typeLeft := exnLeft*.exn_type.ty_typeNo;
    fi;
    exnRight := exn*.exn_.exu_bi.exb_right;
    if exnRight = nil then
	typeRight := ty_unknown;
    else
	typeRight := exnRight*.exn_type.ty_typeNo;
    fi;
    leftArb := typeLeft = ty_error or typeLeft = ty_unknown;
    rightArb := typeRight = ty_error or typeRight = ty_unknown;
    if typeLeft = ty_float then
	exn*.exn_code := exp_floatAssign;
	return(typeRight = ty_float or rightArb);
    fi;
    if typeRight = ty_float then
	exn*.exn_code := exp_floatAssign;
	return(leftArb);
    fi;
    if typeLeft = ty_string then
	exn*.exn_code := exp_stringAssign;
	return(typeRight = ty_string or rightArb);
    fi;
    if typeRight = ty_string then
	exn*.exn_code := exp_stringAssign;
	return(leftArb);
    fi;
    if typeLeft = ty_bool then
	exn*.exn_code := exp_boolAssign;
	return(typeRight = ty_bool or rightArb);
    fi;
    if typeRight = ty_bool then
	exn*.exn_code := exp_boolAssign;
	return(leftArb);
    fi;
    if typeLeft = ty_char then
	exn*.exn_code := exp_charAssign;
	return(typeRight = ty_char or rightArb);
    fi;
    if typeRight = ty_char then
	exn*.exn_code := exp_charAssign;
	return(leftArb);
    fi;
    if typeLeft = ty_ulong or typeLeft = ty_long then
	if typeLeft = ty_ulong then
	    exn*.exn_code := exp_ulongAssign;
	else
	    exn*.exn_code := exp_longAssign;
	fi;
	if typeRight = ty_ulong or typeRight = ty_long or
	    typeRight = ty_uint or typeRight = ty_int or
	    typeRight = ty_ushort or typeRight = ty_short
	then
	    expandToLong(ps, &exn*.exn_.exu_bi.exb_right);
	    return(true);
	fi;
	return(rightArb);
    fi;
    if typeLeft = ty_uint or typeLeft = ty_int then
	if typeLeft = ty_uint then
	    exn*.exn_code := exp_uintAssign;
	else
	    exn*.exn_code := exp_intAssign;
	fi;
	if typeRight = ty_uint or typeRight = ty_int or
	    typeRight = ty_ushort or typeRight = ty_short
	then
	    expandToInt(ps, &exn*.exn_.exu_bi.exb_right);
	    return(true);
	fi;
	if typeRight = ty_ulong then
	    insertConversion(ps, &exn*.exn_.exu_bi.exb_right,
			     exp_ulongToUint, ty_uint);
	    return(true);
	fi;
	if typeRight = ty_long then
	    insertConversion(ps, &exn*.exn_.exu_bi.exb_right,
			     exp_longToInt, ty_int);
	    return(true);
	fi;
	return(rightArb);
    fi;
    if typeLeft = ty_ushort or typeLeft = ty_short then
	if typeLeft = ty_ushort then
	    exn*.exn_code := exp_ushortAssign;
	else
	    exn*.exn_code := exp_shortAssign;
	fi;
	if typeRight = ty_ulong then
	    insertConversion(ps, &exn*.exn_.exu_bi.exb_right,
			     exp_ulongToUint, ty_uint);
	    typeRight := ty_uint;
	elif typeRight = ty_long then
	    insertConversion(ps, &exn*.exn_.exu_bi.exb_right,
			     exp_longToInt, ty_int);
	    typeRight := ty_int;
	fi;
	if typeRight = ty_uint then
	    insertConversion(ps, &exn*.exn_.exu_bi.exb_right,
			     exp_uintToUshort, ty_ushort);
	    return(true);
	fi;
	if typeRight = ty_int then
	    insertConversion(ps, &exn*.exn_.exu_bi.exb_right,
			     exp_intToShort, ty_short);
	    return(true);
	fi;
	return(rightArb);
    fi;
    exn*.exn_code := exp_error;
    if leftArb and
	(rightArb or
	 typeRight = ty_ulong or typeRight = ty_long or
	 typeRight = ty_uint or typeRight = ty_int or
	 typeRight = ty_ushort or typeRight = ty_short)
    then
	return(true);
    fi;
    false
corp;

/*
 * AssignRightChecker - complain if the above fails.
 */

proc AssignRightChecker(*ParserState_t ps; uint token)void:

    if not AssignRightTester(ps, token) then
	ps*.ps_stringError("assignment incompatibility");
    fi;
corp;

/*
 * AssignLeftTester - validate that a LHS is a type that can be assigned to.
 */

proc AssignLeftTester(*ParserState_t ps; uint token)bool:
    register *ExpressionNode_t exn;
    register ulong typ;

    exn := ps*.ps_currentNode;
    if exn = nil then
	return(true);
    fi;
    if not Assignable(exn) then
	return(false);
    fi;
    typ := exn*.exn_type.ty_typeNo;
    typ = ty_error or typ = ty_unknown or typ = ty_float or typ = ty_char or
    typ = ty_string or typ = ty_bool or typ = ty_ulong or typ = ty_long or
    typ = ty_uint or typ = ty_int or typ = ty_ushort or typ = ty_short
corp;

/*
 * AssignLeftChecker - complain if the above fails.
 */

proc AssignLeftChecker(*ParserState_t ps; uint token)void:

    if not AssignLeftTester(ps, token) then
	ps*.ps_stringError("cannot assign to this");
    fi;
corp;

/*
 * CheckIO - check an expression against a format character.
 */

proc CheckIO(*ExpressionNode_t exn; *char pFormat)bool:
    register ulong typ;
    register char format;

    if exn = nil then
	return(true);
    fi;
    typ := exn*.exn_type.ty_typeNo;
    if typ = ty_error or typ = ty_unknown then
	return(true);
    fi;
    format := pFormat*;
    if typ = ty_string or typ = ty_bool then
	return(format = ' ');
    fi;
    if typ = ty_float then
	if format = ' ' then
	    pFormat* := 'g';
	    return(true);
	fi;
	return(format = 'e' or format = 'f' or format = 'g');
    fi;
    if typ = ty_ulong or typ = ty_long or typ = ty_uint or typ = ty_int or
	typ = ty_ushort or typ = ty_short
    then
	if format = ' ' then
	    pFormat* :=
		if typ = ty_ulong or typ = ty_uint or typ = ty_ushort then
		    'u'
		else
		    'd'
		fi;
	    return(true);
	fi;
	return(format = 'x' or format = 'o' or format = 'b' or
	       format = 'u' or format = 'd');
    fi;
    false
corp;

/*
 * IfCaseCompatible - handle the standard types as appearing as the body
 *	part of an if-expression or case-expression.
 */

proc IfCaseCompatible(register *ParserState_t ps; *TypeValue_t pOldType;
		      register **ExpressionNode_t pNewNode)bool:
    *TypeValue_t pNewType;
    register ulong oldType, newType;

    if pNewNode* = nil then
	return(true);
    fi;
    pNewType := &pNewNode**.exn_type;
    if pOldType*.ty_typeNo = pNewType*.ty_typeNo then
	return(true);
    fi;
    if pNewType*.ty_typeNo = ty_error then
	return(true);
    fi;
    if pOldType*.ty_typeNo = ty_unknown or pOldType*.ty_typeNo = ty_error then
	pOldType*.ty_typeNo := pNewType*.ty_typeNo;
	return(true);
    fi;
    oldType := pOldType*.ty_typeNo;
    newType := pNewType*.ty_typeNo;
    if oldType = ty_ulong or oldType = ty_long then
	if newType = ty_ulong or newType = ty_long or
	    newType = ty_uint or newType = ty_int or
	    newType = ty_ushort or newType = ty_short
	then
	    expandToLong(ps, pNewNode);
	    return(true);
	fi;
	return(false);
    fi;
    if oldType = ty_uint or oldType = ty_int then
	if newType = ty_uint or newType = ty_int or
	    newType = ty_ushort or newType = ty_short
	then
	    expandToInt(ps, pNewNode);
	    return(true);
	fi;
	if newType = ty_ulong then
	    insertConversion(ps, pNewNode, exp_ulongToUint, ty_uint);
	    return(true);
	fi;
	if newType = ty_long then
	    insertConversion(ps, pNewNode, exp_longToInt, ty_int);
	    return(true);
	fi;
	return(false);
    fi;
    if oldType = ty_ushort or oldType = ty_short then
	if newType = ty_ulong then
	    insertConversion(ps, pNewNode, exp_ulongToUint, ty_uint);
	    newType := ty_uint;
	elif newType = ty_long then
	    insertConversion(ps, pNewNode, exp_longToInt, ty_int);
	    newType := ty_int;
	fi;
	if newType = ty_uint then
	    insertConversion(ps, pNewNode, exp_uintToUshort, ty_ushort);
	    return(true);
	fi;
	if newType = ty_int then
	    insertConversion(ps, pNewNode, exp_intToShort, ty_short);
	    return(true);
	fi;
	return(false);
    fi;
    false
corp;
