#//Lex/Src/Lex.g
#//Lex/Src/LexStandard.g
#//Common/Types.g
#//Common/Nodes.g
#//Common/Constructs.g

#Parse.g

/*
 * Constructs.d - parser handler routines for standard constructs.
 */

/*
 * NOTE: even if allocations for structures fail, we still want to parse
 *	everything, so that we don't leave an upper-level caller stranded.
 */

/*
 * BuildIf - parse and build nodes for an 'if' statement/expression.
 */

proc BuildIf(register *ParserState_t ps; *Construct_t con)void:
    *ExpressionNode_t exn, exn1, exn2;
    register *IfNode_t ifn;
    *IfPart_t ifp;
    **IfPart_t ifpTail;
    TypeValue_t oldType;

    ifpTail := nil;
    ifp := nil;
    exn := ps*.ps_alloc(sizeof(ExpressionNode_t));
    if exn ~= nil then
	ifn := ps*.ps_alloc(sizeof(IfNode_t));
	if ifn = nil then
	    ps*.ps_free(exn, sizeof(ExpressionNode_t));
	    exn := nil;
	    AllocError(ps);
	else
	    exn*.exn_kind := exk_special;
	    exn*.exn_.exu_con.exc_construct := con;
	    exn*.exn_.exu_con.exc_tree := ifn;
	    ifpTail := &ifn*.ifn_alternatives;
	    ifn*.ifn_else := nil;
	fi;
    else
	AllocError(ps);
    fi;
    oldType.ty_typeNo := ty_unknown;
    while
	GetNextToken(&ps*.ps_tk);
	ParseExpression(ps);
	exn1 := ps*.ps_currentNode;
	if exn1 ~= nil and exn1*.exn_type.ty_typeNo ~= ty_bool and
	    exn1*.exn_type.ty_typeNo ~= ty_error
	then
	    ps*.ps_stringError("'if' condition must be of type 'bool'");
	fi;
	if ps*.ps_tk.tk_kind = tk_then then
	    GetNextToken(&ps*.ps_tk);
	else
	    ps*.ps_stringError("expecting 'then' in 'if'");
	fi;
	ParseStatementList(ps);
	exn2 := ps*.ps_currentNode;
	if exn2 ~= nil and not ps*.ps_ifCaseCompatible(ps,&oldType, &exn2) then
	    ps*.ps_stringError("incompatible alternatives in 'if'");
	fi;
	ifp := nil;
	if ifpTail ~= nil then
	    ifp := ps*.ps_alloc(sizeof(IfPart_t));
	    if ifp ~= nil then
		ifp*.ifp_condition := exn1;
		ifp*.ifp_statement := exn2;
		ifpTail* := ifp;
		ifpTail := &ifp*.ifp_next;
	    else
		AllocError(ps);
	    fi;
	fi;
	if ifp = nil then
	    DestroyTree(ps, exn2);
	    DestroyTree(ps, exn1);
	fi;
	ps*.ps_tk.tk_kind = tk_elif
    do
    od;
    if ifpTail ~= nil then
	ifpTail* := nil;
    fi;
    if ps*.ps_tk.tk_kind = tk_else then
	GetNextToken(&ps*.ps_tk);
	ParseStatementList(ps);
	exn2 := ps*.ps_currentNode;
	if exn2 ~= nil and not ps*.ps_ifCaseCompatible(ps,&oldType, &exn2) then
	    ps*.ps_stringError("incompatible alternatives in 'if'");
	fi;
	if ifpTail ~= nil then
	    ifn*.ifn_else := exn2;
	fi;
    else
	if oldType.ty_typeNo ~= ty_void and oldType.ty_typeNo ~= ty_error
	then
	    ps*.ps_stringError("'if' expression must have 'else' part");
	fi;
    fi;
    if ps*.ps_tk.tk_kind = tk_fi then
	GetNextToken(&ps*.ps_tk);
    else
	ps*.ps_stringError("expecting 'fi' in 'if'");
    fi;
    ps*.ps_currentNode := exn;
    if exn ~= nil then
	exn*.exn_type.ty_typeNo := oldType.ty_typeNo;
    fi;
corp;

/*
 * DestroyIf - destroy an 'if' node.
 */

proc DestroyIf(register *ParserState_t ps; *ExpressionNode_t exn)void:
    *IfNode_t ifn;
    register *IfPart_t ifp, ifpTemp;

    ifn := exn*.exn_.exu_con.exc_tree;
    if ifn ~= nil then
	ifp := ifn*.ifn_alternatives;
	while ifp ~= nil do
	    DestroyTree(ps, ifp*.ifp_statement);
	    DestroyTree(ps, ifp*.ifp_condition);
	    ifpTemp := ifp;
	    ifp := ifp*.ifp_next;
	    ps*.ps_free(ifpTemp, sizeof(IfPart_t));
	od;
	DestroyTree(ps, ifn*.ifn_else);
	ps*.ps_free(ifn, sizeof(IfNode_t));
    fi;
corp;

/*
 * BuildWhile - parse and build nodes for a 'while' statement.
 */

proc BuildWhile(register *ParserState_t ps; *Construct_t con)void:
    register *ExpressionNode_t exn, exn1;
    register *WhileNode_t whn;

    whn := nil;
    exn := ps*.ps_alloc(sizeof(ExpressionNode_t));
    if exn ~= nil then
	whn := ps*.ps_alloc(sizeof(WhileNode_t));
	if whn = nil then
	    ps*.ps_free(exn, sizeof(ExpressionNode_t));
	    exn := nil;
	    AllocError(ps);
	else
	    exn*.exn_kind := exk_special;
	    exn*.exn_type.ty_typeNo := ty_void;
	    exn*.exn_.exu_con.exc_construct := con;
	    exn*.exn_.exu_con.exc_tree := whn;
	    whn*.whn_body := nil;
	fi;
    else
	AllocError(ps);
    fi;
    GetNextToken(&ps*.ps_tk);
    ParseStatementList(ps);
    exn1 := ps*.ps_currentNode;
    if exn1 ~= nil and exn1*.exn_type.ty_typeNo ~= ty_bool and
	exn1*.exn_type.ty_typeNo ~= ty_error
    then
	ps*.ps_stringError("'while' condition must be of type 'bool'");
    fi;
    if whn ~= nil then
	whn*.whn_condition := exn1;
    else
	DestroyTree(ps, exn1);
    fi;
    if ps*.ps_tk.tk_kind = tk_do then
	GetNextToken(&ps*.ps_tk);
    else
	ps*.ps_stringError("expecting 'do' in 'while'");
    fi;
    if ps*.ps_tk.tk_kind ~= tk_od then
	ParseStatementList(ps);
	exn1 := ps*.ps_currentNode;
	if exn1 ~= nil and exn1*.exn_type.ty_typeNo ~= ty_void and
	    exn1*.exn_type.ty_typeNo ~= ty_error
	then
	    ps*.ps_stringError("'while' body must be of type 'void'");
	fi;
	if whn ~= nil then
	    whn*.whn_body := exn1;
	else
	    DestroyTree(ps, exn1);
	fi;
    fi;
    if ps*.ps_tk.tk_kind = tk_od then
	GetNextToken(&ps*.ps_tk);
    else
	ps*.ps_stringError("expecting 'od' in 'while'");
    fi;
    ps*.ps_currentNode := exn;
corp;

/*
 * DestroyWhile - destroy a 'while' node.
 */

proc DestroyWhile(register *ParserState_t ps; *ExpressionNode_t exn)void:
    register *WhileNode_t whn;

    whn := exn*.exn_.exu_con.exc_tree;
    if whn ~= nil then
	DestroyTree(ps, whn*.whn_body);
	DestroyTree(ps, whn*.whn_condition);
	ps*.ps_free(whn, sizeof(WhileNode_t));
    fi;
corp;

/*
 * ps_getLen - helper routine for BuildIO.
 */

proc ps_getLen(register *ParserState_t ps; **ExpressionNode_t pLen, pPrec)void:

    ParseExpression(ps);
    pLen* := ps*.ps_currentNode;
    if ps*.ps_tk.tk_kind = tk_colon then
	GetNextToken(&ps*.ps_tk);
	ParseExpression(ps);
	pPrec* := ps*.ps_currentNode;
    fi;
corp;

/*
 * BuildIO - build a read/write node.
 */

proc BuildIO(register *ParserState_t ps; *Construct_t con)void:
    register *ExpressionNode_t exn;
    register *FormatList_t fml;
    register **FormatList_t fmlTail;
    *IOStatement_t ios;
    *ExpressionNode_t expression, length, precision;
    bool isRead, isLine, hadChannel, hadError;
    char format;

    hadChannel := false;
    case ps*.ps_tk.tk_kind
    incase tk_read:
	isRead := true;
	isLine := false;
    incase tk_readln:
	isRead := true;
	isLine := true;
    incase tk_write:
	isRead := false;
	isLine := false;
    incase tk_writeln:
	isRead := false;
	isLine := true;
    esac;
    fmlTail := nil;
    fml := nil;
    exn := ps*.ps_alloc(sizeof(ExpressionNode_t));
    if exn ~= nil then
	ios := ps*.ps_alloc(sizeof(IOStatement_t));
	if ios = nil then
	    ps*.ps_free(exn, sizeof(ExpressionNode_t));
	    exn := nil;
	    AllocError(ps);
	else
	    exn*.exn_kind := exk_special;
	    exn*.exn_type.ty_typeNo := ty_void;
	    exn*.exn_.exu_con.exc_construct := con;
	    exn*.exn_.exu_con.exc_tree := ios;
	    ios*.ios_channel := nil;
	    ios*.ios_isRead := isRead;
	    ios*.ios_isLine := isLine;
	    fmlTail := &ios*.ios_expressions;
	fi;
    else
	AllocError(ps);
    fi;
    GetNextToken(&ps*.ps_tk);
    if ps*.ps_tk.tk_kind = tk_openParen then
	GetNextToken(&ps*.ps_tk);
    else
	ps*.ps_stringError("expecting '(' for read/write");
    fi;
    while ps*.ps_tk.tk_kind ~= tk_closeParen and
	not ConstructTerminator(ps*.ps_tk.tk_kind)
    do
	ParseExpression(ps);
	if ps*.ps_tk.tk_kind = tk_semicolon then
	    if hadChannel then
		ps*.ps_stringError("already had channel for read/write");
		DestroyTree(ps, ps*.ps_currentNode);
	    else
		hadChannel := true;
/* %%% */	if true then
		    ps*.ps_stringError("value not a channel");
		else
		    if ios ~= nil then
			ios*.ios_channel := ps*.ps_currentNode;
		    fi;
		fi;
	    fi;
	    GetNextToken(&ps*.ps_tk);
	else
	    expression := ps*.ps_currentNode;
	    hadError := false;
	    if isRead and not ps*.ps_assignable(expression) then
		ps*.ps_stringError("can't assign to this");
		hadError := true;
	    fi;
	    format := ' ';
	    length := nil;
	    precision := nil;
	    if fmlTail ~= nil then
		fml := ps*.ps_alloc(sizeof(FormatList_t));
		if fml = nil then
		    AllocError(ps);
		fi;
	    fi;
	    if fml ~= nil then
		fmlTail* := fml;
		fmlTail := &fml*.fml_next;
	    fi;
	    if ps*.ps_tk.tk_kind = tk_colon then
		GetNextToken(&ps*.ps_tk);
		if ps*.ps_tk.tk_kind = tk_id and
		    ps*.ps_tk.tk_bufferPosition = 2
		then
		    /* single-character id - assume it is a format code */
		    format := ps*.ps_tk.tk_buffer*;
		    if not hadError and not ps*.ps_checkIO(expression, &format)
		    then
			ps*.ps_stringError(
			    "invalid format code for this type");
		    fi;
		    GetNextToken(&ps*.ps_tk);
		    if ps*.ps_tk.tk_kind = tk_colon then
			GetNextToken(&ps*.ps_tk);
			ps_getLen(ps, &length, &precision);
		    fi;
		else
		    ps_getLen(ps, &length, &precision);
		fi;
	    fi;
	    if fml ~= nil then
		fml*.fml_expression := expression;
		fml*.fml_format := format;
		fml*.fml_fieldLength := length;
		fml*.fml_fieldPrecision := precision;
	    else
		DestroyTree(ps, precision);
		DestroyTree(ps, length);
		DestroyTree(ps, expression);
	    fi;
	    if ps*.ps_tk.tk_kind = tk_comma then
		GetNextToken(&ps*.ps_tk);
	    fi;
	fi;
    od;
    if fmlTail ~= nil then
	fmlTail* := nil;
    fi;
    if ps*.ps_tk.tk_kind = tk_closeParen then
	GetNextToken(&ps*.ps_tk);
    else
	ps*.ps_stringError("expecting ')' to end read/write");
    fi;
    ps*.ps_currentNode := exn;
corp;

/*
 * DestroyIO - destroy a read/write construct.
 */

proc DestroyIO(register *ParserState_t ps; *ExpressionNode_t exn)void:
    *IOStatement_t ios;
    register *FormatList_t fml, fmlTemp;

    ios := exn*.exn_.exu_con.exc_tree;
    if ios ~= nil then
	DestroyTree(ps, ios*.ios_channel);
	fml := ios*.ios_expressions;
	while fml ~= nil do
	    fmlTemp := fml;
	    fml := fml*.fml_next;
	    DestroyTree(ps, fmlTemp*.fml_fieldPrecision);
	    DestroyTree(ps, fmlTemp*.fml_fieldLength);
	    DestroyTree(ps, fmlTemp*.fml_expression);
	    ps*.ps_free(fmlTemp, sizeof(FormatList_t));
	od;
	ps*.ps_free(ios, sizeof(IOStatement_t));
    fi;
corp;

/*
 * ConstructTerminator - return 'true' if the given token type is one which
 *	can terminate a statement list in a construct supported here.
 */

proc ConstructTerminator(register uint token)bool:

    token = tk_fi or token = tk_elif or token = tk_else or token = tk_od or
    token = tk_case or token = tk_default or token = tk_esac or
    token = tk_endOfFile
corp;

/*
 * IsNonDeclaration - token cannot start a declaration - used to tell when
 *	to stop parsing declarations and switch to executable code.
 */

proc IsNonDeclaration(register uint token)bool:

    token = tk_if or token = tk_while or token = tk_case or token = tk_for or
    token = tk_read or token = tk_write or token = tk_readln or
    token = tk_writeln or token = tk_openParen or token = tk_closeParen or
    ConstructTerminator(token)
corp;
