#//Common/Types.g
#//Common/NodeTypes.g
#//Common/Nodes.g
#//Lex/Src/Lex.g
#//Lex/Src/LexStandard.g
#//Print/Src/Print.g
#Parse.g
#ParseStandard.g

[STANDARD_CONSTRUCT_COUNT] Construct_t Constructs := (
    (tk_if     , BuildIf   , DestroyIf	 , nil, PrintIf   ),
    (tk_while  , BuildWhile, DestroyWhile, nil, PrintWhile),
    (tk_read   , BuildIO   , DestroyIO	 , nil, PrintIO   ),
    (tk_readln , BuildIO   , DestroyIO	 , nil, PrintIO   ),
    (tk_write  , BuildIO   , DestroyIO	 , nil, PrintIO   ),
    (tk_writeln, BuildIO   , DestroyIO	 , nil, PrintIO   )
);

[STANDARD_CONSTRUCT_COUNT] ConstructList_t ConstructList := (
    (&ConstructList[1], &Constructs[0]),
    (&ConstructList[2], &Constructs[1]),
    (&ConstructList[3], &Constructs[2]),
    (&ConstructList[4], &Constructs[3]),
    (&ConstructList[5], &Constructs[4]),
    (nil	      , &Constructs[5]),
);

/* maintained in order of increasing priority */

[STANDARD_OPERATION_COUNT] Operation_t Operations := (
    (tk_assign	      , opk_binary, AssignLeftChecker , AssignRightChecker  ),
    (tk_or	      , opk_binary, BoolLeftChecker   , BoolRightChecker    ),
    (tk_and	      , opk_binary, BoolLeftChecker   , BoolRightChecker    ),
    (tk_not	      , opk_prefix, nil 	      , BoolPrefixChecker   ),
    (tk_lessThan      , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_lessOrEqual   , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_greaterThan   , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_greaterOrEqual, opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_equal	      , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_notEqual      , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_star	      , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_slash	      , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_percent       , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_plus	      , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_minus	      , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_plus	      , opk_prefix, nil 	      , NumericPrefixChecker),
    (tk_minus	      , opk_prefix, nil 	      , NumericPrefixChecker),
    (tk_bitOr	      , opk_prefix, nil 	      , NumericPrefixChecker),
    (tk_bitOr	      , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_shiftLeft     , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_shiftRight    , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_bitXor	      , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_bitAnd	      , opk_binary, NumericLeftChecker, NumericRightChecker ),
    (tk_bitNot	      , opk_prefix, nil 	      , NumericPrefixChecker)
);

/* TRUE -> end of a given priority level in the above table */

[STANDARD_OPERATION_COUNT] bool OperationBreak := (
    true,					/* assign */
    true,					/* bool or */
    true,					/* bool and */
    true,					/* bool not */
    false, false, false, false, false, true,	/* comparisons */
    false, false, true, 			/* times, divide, rem */
    false, true,				/* binary plus, minus */
    false, false, true, 			/* unPlus, unMinus, abs */
    true,					/* bitOr */
    false, false, false, true,			/* shifts, bitXor, bitAnd */
    true					/* bitNot */
);

[STANDARD_BUILTIN_COUNT] BuiltInList_t BuiltInList := (
    (&BuiltInList[ 1], tk_void	, ty_void),
    (&BuiltInList[ 2], tk_bool	, ty_bool),
    (&BuiltInList[ 3], tk_byte	, ty_ushort),
    (&BuiltInList[ 4], tk_char	, ty_char),
    (&BuiltInList[ 5], tk_string, ty_string),
    (&BuiltInList[ 6], tk_ushort, ty_ushort),
    (&BuiltInList[ 7], tk_short , ty_short),
    (&BuiltInList[ 8], tk_uint	, ty_uint),
    (&BuiltInList[ 9], tk_int	, ty_int),
    (&BuiltInList[10], tk_ulong , ty_ulong),
    (&BuiltInList[11], tk_long	, ty_long),
    (nil,	       tk_float , ty_float)
);

/*
 * ParseStandard - fill in the passed areas with a standard parser setup.
 */

proc ParseStandard(*ParserState_t ps; *Construct_t pConstructs;
		   *Operation_t pOperations;
		   register *OperationList_t pOperationList)void:
    register *PriorityList_t prl;
    register *bool pBreak;
    register uint cnt;

    if pConstructs ~= nil and ps*.ps_constructs ~= nil then
	pretend(pConstructs, *[STANDARD_CONSTRUCT_COUNT] Construct_t)* :=
	    Constructs;
	pretend(ps*.ps_constructs,
		*[STANDARD_CONSTRUCT_COUNT] ConstructList_t)* :=
		    ConstructList;
    fi;
    if pOperations ~= nil and pOperationList ~= nil and
	ps*.ps_operations ~= nil
    then
	pretend(pOperations, *[STANDARD_OPERATION_COUNT] Operation_t)* :=
	    Operations;
	prl := ps*.ps_operations;
	pBreak := &OperationBreak[0];
	prl*.prl_operations := pOperationList;
	for cnt from STANDARD_OPERATION_COUNT - 1 downto 0 do
	    pOperationList*.opl_this := pOperations;
	    if pBreak* then
		pOperationList*.opl_next := nil;
	    else
		pOperationList*.opl_next :=
		    pOperationList + sizeof(OperationList_t);
	    fi;
	    pOperationList := pOperationList + sizeof(OperationList_t);
	    pOperations := pOperations + sizeof(Operation_t);
	    if pBreak* then
		prl*.prl_next := prl + sizeof(PriorityList_t);
		prl := prl + sizeof(PriorityList_t);
		if cnt ~= 0 then
		    prl*.prl_operations := pOperationList;
		fi;
	    fi;
	    pBreak := pBreak + sizeof(bool);
	od;
	(prl - sizeof(PriorityList_t))*.prl_next := nil;
    fi;
    if ps*.ps_builtInTypes ~= nil then
	pretend(ps*.ps_builtInTypes,
		*[STANDARD_BUILTIN_COUNT] BuiltInList_t)* :=
		    BuiltInList;
    fi;
    ps*.ps_currentNode := nil;
    ps*.ps_hadAllocError := false;
    ps*.ps_isNonDeclaration := IsNonDeclaration;
    ps*.ps_isTerminator := ConstructTerminator;
    ps*.ps_ifCaseCompatible := IfCaseCompatible;
    ps*.ps_assignable := Assignable;
    ps*.ps_checkIO := CheckIO;
corp;
