#drinc:exec/memory.g
#drinc:util.g

#//Common/Types.g
#//Common/Nodes.g
#//Lex/Src/Lex.g
#//Lex/Src/LexStandard.g
#//Symbol/Src/Symbol.g
#/Src/Parse.g
#/Src/ParseStandard.g
#//Print/Src/Print.g

[STANDARD_RESERVED_WORD_COUNT] ReservedWord_t ReservedWords;
[STANDARD_OPERATOR_COUNT] Operator_t Operators;

[STANDARD_CONSTRUCT_COUNT] Construct_t Constructs;
[STANDARD_CONSTRUCT_COUNT] ConstructList_t ConstructList;
[STANDARD_OPERATION_COUNT] Operation_t Operations;
[STANDARD_OPERATION_COUNT] OperationList_t OperationList;
[STANDARD_PRIORITY_COUNT] PriorityList_t PriorityList;
[STANDARD_BUILTIN_COUNT] BuiltInList_t BuiltInList;

ParserState_t ParserState;
SymbolState_t SymbolState;
PrintState_t PrintState;
uint Line, Column, PrevLine, PrevColumn, CurrentIndent;
*SymbolTable_t SymbolTable;

proc getChar()char:
    char ch;

    PrevLine := Line;
    PrevColumn := Column;
    if not read(ch) then
	if ioerror() = CH_EOF then
	    ch := '\e';
	else
	    readln();
	    ch := '\n';
	    Column := 0;
	    Line := Line + 1;
	fi;
    else
	if ch = '\t' then
	    Column := (Column + 8) & (-8);
	else
	    Column := Column + 1;
	fi;
    fi;
    ch
corp;

proc err(*char message)void:

    writeln("Line ", PrevLine, ", column ", PrevColumn, " *** ", message);
corp;

proc err3(*char message1, message2, message3)void:

    writeln("Line ", PrevLine, ", column ", PrevColumn, " *** ",
	    message1, message2, message3);
corp;

proc getExpression()ulong:

    GetNextToken(&ParserState.ps_tk);
    case ParserState.ps_tk.tk_kind
    incase tk_ulong:
	ParserState.ps_tk.tk_ulongValue
    incase tk_char:
	ParserState.ps_tk.tk_charValue - '\e'
    default:
	err("illegal escaped character expression");
	0
    esac
corp;

proc doubleBuffer(*TokenizerState_t tk)bool:
    *char newBuffer;

    newBuffer := AllocMem(tk*.tk_bufferLength * 2, 0);
    if newBuffer = nil then
	false
    else
	BlockCopy(newBuffer, tk*.tk_buffer, tk*.tk_bufferPosition);
	FreeMem(tk*.tk_buffer, tk*.tk_bufferLength);
	tk*.tk_buffer := newBuffer;
	tk*.tk_bufferLength := tk*.tk_bufferLength * 2;
	true
    fi
corp;

proc myAlloc(ulong length)arbptr:

    AllocMem(length, 0)
corp;

proc myFree(arbptr p; ulong length)void:

    FreeMem(p, length);
corp;

proc idDeclarer(*char name; ulong nameLen; *TypeValue_t pType)*Symbol_t:
    register *Symbol_t sy;

    sy := sy_enter(&SymbolState, SymbolTable, name, nameLen);
    if sy ~= nil then
	if sy*.sy_kind = syk_entered then
	    sy*.sy_kind := 10;
	    sy*.sy_type.syu_typeNo := pType*.ty_typeNo;
	else
	    err3("symbol '", name, "' is already defined");
	fi;
    fi;
    sy
corp;

proc idHandler(*char name; ulong nameLen)*Symbol_t:
    register *Symbol_t sy;

    sy := sy_lookup(&SymbolState, SymbolTable, name, nameLen);
    if sy = nil then
	err3("symbol '", name, "' is not defined");
	sy := sy_enter(&SymbolState, SymbolTable, name, nameLen);
	if sy ~= nil then
	    sy*.sy_kind := 10;
	    sy*.sy_type.syu_typeNo := ty_error;
	fi;
    fi;
    sy
corp;

proc myDelete(*Symbol_t sy)void:

/*
    writeln("Deleting symbol '", sy*.sy_name, '\'');
*/
corp;

proc writeSpace()void:

    write(' ');
corp;

proc writeNewLine()void:
    register uint i;

    writeln();
    if CurrentIndent ~= 0 then
	for i from CurrentIndent - 1 downto 0 do
	    write(' ');
	od;
    fi;
corp;

proc indent(bool indentIn)void:

    if indentIn then
	CurrentIndent := CurrentIndent + 4;
    elif CurrentIndent ~= 0 then
	CurrentIndent := CurrentIndent - 4;
    fi;
corp;

proc putChar(char ch)void:

    write(ch);
corp;

proc putString(*char st)void:

    write(st);
corp;

proc printSymbol(*Symbol_t sy)void:

    write(sy*.sy_name);
corp;

proc checkDump()void:
    register uint i;
    register *PriorityList_t prl;
    register *OperationList_t opl;

/*
    writeln("Constructs ", &Constructs[0], " Operations ", &Operations[0],
	    " OperationList ", &OperationList[0]);
    writeln("ConstructList ", &ConstructList[0], " PriorityList ",
	    &PriorityList[0]);
    writeln("OperationList:");
    for i from 0 upto STANDARD_OPERATION_COUNT - 1 do
	writeln(i, ": ", OperationList[i].opl_next,
		", ", OperationList[i].opl_this);
    od;
    writeln("PriorityList:");
    for i from 0 upto STANDARD_PRIORITY_COUNT - 1 do
	writeln(i, ": ", PriorityList[i].prl_next,
		", ", PriorityList[i].prl_operations);
    od;
*/
    prl := &PriorityList[0];
    while
	write(prl : 6, ": ");
	opl := prl*.prl_operations;
	while opl ~= nil do
	    write(opl, ' ');
	    opl := opl*.opl_next;
	od;
	writeln();
	prl := prl*.prl_next;
	prl ~= nil
    do
    od;
corp;

proc main()void:
    type PAIR = [2] char;
    register *ParserState_t ps;
    register *TokenizerState_t tk;
    register *PrintState_t prs;
    TypeValue_t typ;
    uint oldLine, oldColumn;

    if OpenlexLibrary(0) = nil then
	return;
    fi;

    SymbolState.sym_alloc := myAlloc;
    SymbolState.sym_free := myFree;
    SymbolState.sym_hashFunction := sy_simpleHash;
    SymbolState.sym_deleteFunction := myDelete;
    SymbolTable := sy_create(&SymbolState, 100);
    if SymbolTable = nil then
	CloselexLibrary();
	return;
    fi;

    ps := &ParserState;
    tk := &ps*.ps_tk;
    prs := &PrintState;
    tk*.tk_buffer := AllocMem(1000, 0);
    if tk*.tk_buffer = nil then
	return;
    fi;
    tk*.tk_bufferLength := 1000;
    tk*.tk_getChar := getChar;
    tk*.tk_getExpression := getExpression;
    tk*.tk_stringError := err;
    tk*.tk_doubleBuffer := doubleBuffer;
    tk*.tk_reservedWords := &ReservedWords[0];
    tk*.tk_operators := &Operators[0];
    LexStandard(tk);

    ps*.ps_stringError := err;
    ps*.ps_stringError3 := err3;
    ps*.ps_alloc := myAlloc;
    ps*.ps_free := myFree;
    ps*.ps_idDeclarer := idDeclarer;
    ps*.ps_idHandler := idHandler;
    ps*.ps_constructs := &ConstructList[0];
    ps*.ps_operations := &PriorityList[0];
    ps*.ps_builtInTypes := &BuiltInList[0];
    ParseStandard(ps, &Constructs[0], &Operations[0], &OperationList[0]);

    prs*.prs_space := writeSpace;
    prs*.prs_newLine := writeNewLine;
    prs*.prs_indent := indent;
    prs*.prs_putChar := putChar;
    prs*.prs_putString := putString;
    prs*.prs_printToken := nil;
    prs*.prs_printSymbol := printSymbol;
    prs*.prs_printChar := nil;
    prs*.prs_tk := tk;

    PrevLine := 1;
    Line := 1;
    PrevColumn := 0;
    Column := 0;
    CurrentIndent := 0;
    GetNextToken(tk);
    while tk*.tk_kind >= tk_void and tk*.tk_kind <= tk_float do
	ParseType(ps, &typ);
	Declare(ps, &typ, nil, tk_endOfFile);
	if ps*.ps_tk.tk_kind = tk_semicolon then
	    GetNextToken(tk);
	fi;
    od;

    while
	tk*.tk_kind ~= tk_endOfFile
    do
	oldLine := Line;
	oldColumn := Column;
	ParseStatementList(ps);
	PrintTree(prs, ps*.ps_currentNode);
	writeln();
	DestroyTree(ps, ps*.ps_currentNode);
	if Line = oldLine and Column = oldColumn then
	    err("unexpected closing token - ignored");
	    GetNextToken(tk);
	fi;
    od;

    FreeMem(tk*.tk_buffer, tk*.tk_bufferLength);
    sy_destroy(&SymbolState, SymbolTable);
    CloselexLibrary();
corp;
