#//Lex/Src/Lex.g
#//Lex/Src/LexStandard.g
#//Common/Types.g
#//Common/Nodes.g
#//Common/NodeTypes.g
#//Common/Constructs.g

#Print.g

/*
 * PrintIf - print an 'if' node.
 */

proc PrintIf(register *PrintState_t prs; *ExpressionNode_t exn)void:
    register *IfNode_t ifn;
    register *IfPart_t ifp;
    bool first;

    ifn := exn*.exn_.exu_con.exc_tree;
    if ifn ~= nil then
	ifp := ifn*.ifn_alternatives;
	first := true;
	while ifp ~= nil do
	    if prs*.prs_printToken ~= nil then
		if first then
		    first := false;
		    prs*.prs_printToken(tk_if);
		else
		    prs*.prs_printToken(tk_elif);
		fi;
	    else
		if first then
		    first := false;
		    PrintToken(prs, tk_if);
		else
		    PrintToken(prs, tk_elif);
		fi;
	    fi;
	    prs*.prs_indent(true);
	    prs*.prs_space();
	    PrintTree(prs, ifp*.ifp_condition);
	    prs*.prs_space();
	    prs*.prs_indent(false);
	    if prs*.prs_printToken ~= nil then
		prs*.prs_printToken(tk_then);
	    else
		PrintToken(prs, tk_then);
	    fi;
	    prs*.prs_indent(true);
	    prs*.prs_newLine();
	    PrintTree(prs, ifp*.ifp_statement);
	    prs*.prs_indent(false);
	    prs*.prs_newLine();
	    ifp := ifp*.ifp_next;
	od;
	if ifn*.ifn_else ~= nil then
	    if prs*.prs_printToken ~= nil then
		prs*.prs_printToken(tk_else);
	    else
		PrintToken(prs, tk_else);
	    fi;
	    prs*.prs_indent(true);
	    prs*.prs_newLine();
	    PrintTree(prs, ifn*.ifn_else);
	    prs*.prs_indent(false);
	    prs*.prs_newLine();
	fi;
	if prs*.prs_printToken ~= nil then
	    prs*.prs_printToken(tk_fi);
	else
	    PrintToken(prs, tk_fi);
	fi;
    fi;
corp;

/*
 * PrintWhile - print out a while node.
 */

proc PrintWhile(register *PrintState_t prs; *ExpressionNode_t exn)void:
    register *WhileNode_t whn;
    register proc(uint token)void printToken;

    whn := exn*.exn_.exu_con.exc_tree;
    if whn ~= nil then
	printToken := prs*.prs_printToken;
	if printToken ~= nil then
	    printToken(tk_while);
	else
	    PrintToken(prs, tk_while);
	fi;
	if whn*.whn_condition ~= nil and
	    (whn*.whn_condition*.exn_kind = exk_list or
	     whn*.whn_condition*.exn_kind = exk_special)
	then
	    prs*.prs_indent(true);
	    prs*.prs_newLine();
	    PrintTree(prs, whn*.whn_condition);
	    prs*.prs_indent(false);
	    prs*.prs_newLine();
	else
	    prs*.prs_space();
	    PrintTree(prs, whn*.whn_condition);
	    prs*.prs_space();
	fi;
	if printToken ~= nil then
	    printToken(tk_do);
	else
	    PrintToken(prs, tk_do);
	fi;
	if whn*.whn_body ~= nil then
	    prs*.prs_indent(true);
	    prs*.prs_newLine();
	    PrintTree(prs, whn*.whn_body);
	    prs*.prs_indent(false);
	fi;
	prs*.prs_newLine();
	if printToken ~= nil then
	    printToken(tk_od);
	else
	    PrintToken(prs, tk_od);
	fi;
    fi;
corp;

/*
 * PrintIO - print a read/write/readln/writeln construct.
 */

proc PrintIO(register *PrintState_t prs; *ExpressionNode_t exn)void:
    register *IOStatement_t ios;
    register *FormatList_t fml;
    register proc(uint token)void printToken;
    register uint token;

    ios := exn*.exn_.exu_con.exc_tree;
    if ios ~= nil then
	if ios*.ios_isRead then
	    if ios*.ios_isLine then
		token := tk_readln;
	    else
		token := tk_read;
	    fi;
	else
	    if ios*.ios_isLine then
		token := tk_writeln;
	    else
		token := tk_write;
	    fi;
	fi;
	printToken := prs*.prs_printToken;
	if printToken ~= nil then
	    printToken(token);
	    printToken(tk_openParen);
	else
	    PrintToken(prs, token);
	    PrintToken(prs, tk_openParen);
	fi;
	fml := ios*.ios_expressions;
	if ios*.ios_channel ~= nil then
	    PrintTree(prs, ios*.ios_channel);
	    if printToken ~= nil then
		printToken(tk_semicolon);
	    else
		PrintToken(prs, tk_semicolon);
	    fi;
	    if fml ~= nil then
		prs*.prs_space();
	    fi;
	fi;
	while fml ~= nil do
	    PrintTree(prs, fml*.fml_expression);
	    if fml*.fml_format ~= ' ' then
		prs*.prs_space();
		if printToken ~= nil then
		    printToken(tk_colon);
		else
		    PrintToken(prs, tk_colon);
		fi;
		prs*.prs_space();
		prs*.prs_putChar(fml*.fml_format);
	    fi;
	    if fml*.fml_fieldLength ~= nil then
		prs*.prs_space();
		if printToken ~= nil then
		    printToken(tk_colon);
		else
		    PrintToken(prs, tk_colon);
		fi;
		prs*.prs_space();
		PrintTree(prs, fml*.fml_fieldLength);
	    fi;
	    if fml*.fml_fieldPrecision ~= nil then
		prs*.prs_space();
		if printToken ~= nil then
		    printToken(tk_colon);
		else
		    PrintToken(prs, tk_colon);
		fi;
		prs*.prs_space();
		PrintTree(prs, fml*.fml_fieldPrecision);
	    fi;
	    fml := fml*.fml_next;
	    if fml ~= nil then
		if printToken ~= nil then
		    printToken(tk_comma);
		else
		    PrintToken(prs, tk_comma);
		fi;
		prs*.prs_space();
	    fi;
	od;
	if printToken ~= nil then
	    printToken(tk_closeParen);
	else
	    PrintToken(prs, tk_closeParen);
	fi;
    fi;
corp;
