#Symbol.g

/*
 * sy_create - create a new symbol table of the given size.
 */

proc sy_create(register *SymbolState_t sym;
	       register ulong entryCount)*SymbolTable_t:
    register *SymbolTable_t syt;
    register *Symbol_t sy;

    if entryCount < 5 then
	return(nil);
    fi;
    syt := sym*.sym_alloc(sizeof(SymbolTable_t));
    if syt ~= nil then
	syt*.syt_entryCount := entryCount;
	syt*.syt_usedCount := 0;
	sy := sym*.sym_alloc(sizeof(Symbol_t) * entryCount);
	if sy ~= nil then
	    syt*.syt_entries := sy;
	    while entryCount ~= 0 do
		entryCount := entryCount - 1;
		sy*.sy_kind := syk_empty;
		sy := sy + sizeof(Symbol_t);
	    od;
	else
	    sym*.sym_free(syt, sizeof(SymbolTable_t));
	    syt := nil;
	fi;
    fi;
    syt
corp;

/*
 * sy_destroy - destroy the indicated symbol table. This is done by calling
 *	'sym_deleteFunction' for each symbol present in the table. Note that
 *	that routine should not free the symbol names as that is done here.
 */

proc sy_destroy(register *SymbolState_t sym; *SymbolTable_t syt)void:
    register ulong n;
    register *Symbol_t sy;

    if syt ~= nil then
	sy := syt*.syt_entries;
	n := syt*.syt_entryCount;
	while n ~= 0 do
	    n := n - 1;
	    if sy*.sy_kind ~= syk_empty and sy*.sy_kind ~= syk_free then
		sym*.sym_deleteFunction(sy);
		sym*.sym_free(sy*.sy_name, sy*.sy_nameLen);
	    fi;
	    sy := sy + sizeof(Symbol_t);
	od;
	sym*.sym_free(syt*.syt_entries,syt*.syt_entryCount * sizeof(Symbol_t));
	sym*.sym_free(syt, sizeof(SymbolTable_t));
    fi;
corp;

/*
 * sy_empty - empty a symbol table by deleting all of its entries.
 */

proc sy_empty(register *SymbolState_t sym; register *SymbolTable_t syt)void:
    register ulong n;
    register *Symbol_t sy;

    if syt ~= nil then
	sy := syt*.syt_entries;
	n := syt*.syt_entryCount;
	while n ~= 0 do
	    n := n - 1;
	    if sy*.sy_kind ~= syk_empty and sy*.sy_kind ~= syk_free then
		sym*.sym_deleteFunction(sy);
		sym*.sym_free(sy*.sy_name, sy*.sy_nameLen);
	    fi;
	    sy*.sy_kind := syk_empty;
	    sy := sy + sizeof(Symbol_t);
	od;
	syt*.syt_usedCount := 0;
    fi;
corp;

/*
 * sy_search - local routine to search for a name or where the name should
 *	be in a table.
 */

proc sy_search(*SymbolState_t sym; *SymbolTable_t syt; *char name;
	       ulong nameLen)*Symbol_t:
    register *Symbol_t sy;
    *Symbol_t found, syEnd, start;
    register *char p, q;
    register ulong hash, len;
    bool first;

    hash := sym*.sym_hashFunction(name, nameLen);
    sy := syt*.syt_entries + hash % syt*.syt_entryCount * sizeof(Symbol_t);
    syEnd := syt*.syt_entries + (syt*.syt_entryCount - 1) * sizeof(Symbol_t);
    found := nil;
    start := sy;
    first := true;
    while
	if sy = start and not first then
	    /* scanned entire table - name not found - use the first of the
	       free slots we passed. */
	    sy := found;
	    false
	elif sy*.sy_kind = syk_empty then
	    /* found an unused slot - name is not in table */
	    false
	elif sy*.sy_kind = syk_free then
	    /* found a slot that used to be used - have to keep looking, but
	       remember this slot for entry if it is the first of them */
	    if found = nil then
		found := sy;
	    fi;
	    true
	elif sy*.sy_hash ~= hash or sy*.sy_nameLen ~= nameLen then
	    /* match not possible */
	    true
	else
	    /* have to actually compare strings */
	    p := name;
	    q := sy*.sy_name;
	    len := nameLen;
	    while len ~= 0 and p* = q* do
		len := len - 1;
		p := p + sizeof(char);
		q := q + sizeof(char);
	    od;
	    len ~= 0
	fi
    do
	if sy = syEnd then
	    sy := syt*.syt_entries;
	else
	    sy := sy + sizeof(Symbol_t);
	fi;
	first := false;
    od;
    if sy*.sy_kind = syk_empty or sy*.sy_kind = syk_free then
	if found ~= nil then
	    sy := found;
	fi;
	sy*.sy_hash := hash;
    fi;
    sy
corp;

/*
 * sy_enter - enter a symbol into the symbol table, and return a pointer to
 *	the Symbol_t structure. Return 'nil' if the name is already in the
 *	table or if we cannot allocate the needed space.
 */

proc sy_enter(*SymbolState_t sym; *SymbolTable_t syt; *char name;
	      ulong nameLen)*Symbol_t:
    register *Symbol_t sy;
    register *char p, q;
    *SymbolTable_t newTable;
    register ulong len;

    if syt = nil or name = nil or nameLen = 0 then
	return(nil);
    fi;
    if syt*.syt_usedCount >= syt*.syt_entryCount * 4 / 5 then
	newTable := sy_create(sym, syt*.syt_entryCount * 2);
	if newTable = nil then
	    return(nil);
	fi;
	newTable*.syt_usedCount := syt*.syt_usedCount;
	len := syt*.syt_entryCount;
	sy := syt*.syt_entries;
	while len ~= 0 do
	    len := len - 1;
	    if sy*.sy_kind ~= syk_empty and sy*.sy_kind ~= syk_free then
		sy_search(sym, newTable, sy*.sy_name, sy*.sy_nameLen)* := sy*;
	    fi;
	    sy := sy + sizeof(Symbol_t);
	od;
	sym*.sym_free(syt*.syt_entries,syt*.syt_entryCount * sizeof(Symbol_t));
	syt* := newTable*;
	sym*.sym_free(newTable, sizeof(SymbolTable_t));
    fi;
    sy := sy_search(sym, syt, name, nameLen);
    if sy*.sy_kind = syk_empty or sy*.sy_kind = syk_free then
	sy*.sy_kind := syk_entered;
	sy*.sy_nameLen := nameLen;
	sy*.sy_name := sym*.sym_alloc(nameLen);
	if sy*.sy_name = nil then
	    return(nil);
	fi;
	p := name;
	q := sy*.sy_name;
	len := nameLen;
	while len ~= 0 do
	    len := len - 1;
	    q* := p*;
	    q := q + sizeof(char);
	    p := p + sizeof(char);
	od;
	syt*.syt_usedCount := syt*.syt_usedCount + 1;
    fi;
    sy
corp;

/*
 * sy_lookup - look a name up in the table - do not enter it if not found.
 */

proc sy_lookup(register *SymbolState_t sym; register *SymbolTable_t syt;
	       *char name; ulong nameLen)*Symbol_t:
    register *Symbol_t sy;

    if syt = nil or name = nil or nameLen = 0 then
	nil
    else
	sy := sy_search(sym, syt, name, nameLen);
	if sy*.sy_kind = syk_empty or sy*.sy_kind = syk_free then
	    sy := nil;
	fi;
	sy
    fi
corp;

/*
 * sy_delete - delete the given symbol from the given symbol table.
 */

proc sy_delete(*SymbolState_t sym; *SymbolTable_t syt; *char name;
	       ulong nameLen)void:
    register *Symbol_t sy;

    if syt ~= nil then
	sy := sy_search(sym, syt, name, nameLen);
	if sy*.sy_kind ~= syk_empty and sy*.sy_kind ~= syk_free then
	    sym*.sym_deleteFunction(sy);
	    sym*.sym_free(sy*.sy_name, sy*.sy_nameLen);
	    sy*.sy_kind := syk_free;
	    syt*.syt_usedCount := syt*.syt_usedCount - 1;
	fi;
    fi;
corp;

/*
 * sy_simpleHash - standard hash function for symbol tables.
 */

proc sy_simpleHash(register *char name; register ulong nameLen)ulong:
    register ulong hash;

    hash := name* - '\e';
    if nameLen > 1 then
	hash := hash << 8;
	hash := hash | ((name + sizeof(char))* - '\e');
	if nameLen > 2 then
	    hash := hash << 8;
	    hash := hash | ((name + (nameLen - 2) * sizeof(char))* - '\e');
	    if nameLen > 3 then
		hash := hash << 8;
		hash := hash |
			((name + (nameLen - 3) * sizeof(char))* - '\e');
	    fi;
	fi;
    fi;
    hash
corp;
