#patternInternal.g

register *PatternState_t PS;

/*
 * nextItem - move up to the next item in the input pattern. Here, an item
 *	is either a single character or an escaped character.
 */

proc nextItem()void:
    register *[2]char pattern;
    register ulong pos;

    pattern := PS*.ps_pattern;
    pos := PS*.ps_position;
    if PS*.ps_char = '\'' then
	if pos ~= PS*.ps_length then
	    PS*.ps_char := pattern*[pos];
	    pos := pos + 1;
	else
	    PS*.ps_char := ' ';
	    PS*.ps_end := true;
	fi;
    fi;
    if pos ~= PS*.ps_length then
	PS*.ps_char := pattern*[pos];
	pos := pos + 1;
    else
	PS*.ps_char := ' ';
	PS*.ps_end := true;
    fi;
    PS*.ps_position := pos;
corp;

/*
 * primary - recursive descent parser for expression primaries. Return the
 *	head of the exit chain within the compiled vector for this primary.
 */

proc primary()ulong:
    extern expression(ulong chain)ulong;
    register *[2]ulong comp;
    register *ulong pComp;
    register ulong startPos, chain;
    register char operator;

    startPos := PS*.ps_position;
    if PS*.ps_end then
	PS*.ps_error := pse_missingPrimary;
    else
	operator := PS*.ps_char;
	nextItem();
	case operator
	incase ')':
	    PS*.ps_error := pse_unexpectedRightParen;
	incase '|':
	    PS*.ps_error := pse_unexpectedOr;
	incase '\#':
	incase '~':
	    chain := primary();
	    /* point all of the exits from the primary to the '#', thus
	       forming the repetition loop */
	    comp := PS*.ps_compiled;
	    while chain ~= 0 do
		pComp := &comp*[chain];
		chain := pComp*;
		pComp* := startPos;
	    od;
	incase '(':
	    /* recurse down to get a nested subpattern */
	    startPos := expression(startPos);
	    if PS*.ps_end or PS*.ps_char ~= ')' then
		PS*.ps_error := pse_missingRightParen;
	    else
		/* skip the ')' */
		nextItem();
	    fi;
	esac;
    fi;
    startPos
corp;

/*
 * expression - recursive descent parser - parse an alternation.
 *	'chain' is the head of a chain in the compiled vector of the
 *	exits that already exist on the same level as this expression. The
 *	same chain, augmented by the alternatives here, is returned.
 */

proc expression(ulong oldChain)ulong:
    register *[2]ulong comp;
    register *ulong pComp;
    register ulong newChain, exits, temp;
    register char operator;

    comp := PS*.ps_compiled;
    exits := 0;
    while true do
	newChain := primary();
	if PS*.ps_end then
	    operator := ' ';
	else
	    operator := PS*.ps_char;
	fi;
	if PS*.ps_end or operator = '|' or operator = ')' then
	    /* join together the exits built from any alternation at
	       this level with those from the primary */
	    if exits = 0 then
		exits := newChain;
	    else
		temp := exits;
		while
		    pComp := &comp*[temp];
		    pComp* ~= 0
		do
		    temp := pComp*;
		od;
		pComp* := newChain;
	    fi;
	    if operator ~= '|' then
		return exits;
	    fi;
	    /* there should be another alterative - make the last
	       '|' or '(' point to this one */
	    comp*[oldChain] := PS*.ps_position;
	    oldChain := PS*.ps_position;
	    nextItem();
	else
	    /* a non-special character - make all of the exits from the
	       primary point to this element as their successors */
	    while newChain ~= 0 do
		pComp := &comp*[newChain];
		temp := pComp*;
		pComp* := PS*.ps_position;
		newChain := temp;
	    od;
	fi;
    od;
    /* we never get here, but the compiler isn't smart enough to know that,
       so we have to return something */
    0
corp;

/*
 * internalCompile - the top-level entry for pattern compilation.
 */

proc internalCompile(*PatternState_t ps)void:
    register *[2]ulong comp;
    register *ulong cPtr;
    register ulong i, temp;

    PS := ps;
    if PS*.ps_length = 0 then
	PS*.ps_error := pse_missingPrimary;
    else
	cPtr := &PS*.ps_compiled*[0];
	for i from PS*.ps_length downto 0 do
	    cPtr* := 0;
	    cPtr := cPtr + sizeof(ulong);
	od;
	PS*.ps_error := pse_ok;
	PS*.ps_end := false;
	PS*.ps_char := PS*.ps_pattern*[0];
	PS*.ps_position := 1;
	i := expression(0);
	if PS*.ps_char = ')' then
	    PS*.ps_error := pse_unexpectedRightParen;
	fi;
	/* Any pointers left coming out of the expression subpattern will
	   be for a successful match. Give them a compiled value of 0, which
	   indicates the final success. */
	comp := PS*.ps_compiled;
	while i ~= 0 do
	    cPtr := &comp*[i];
	    temp := cPtr*;
	    cPtr* := 0;
	    i := temp;
	od;
    fi;
corp;
