#patternInternal.g

register *PatternState_t PS;

/*
 * addState - add the given state to the set of active states.
 */

proc addState(register ulong state)void:
    register *ulong pState;
    register ulong i, count;

    if state = 0 then

	/* state 0 is the special end-of-pattern state. If we see it, then
	   we mark the match as being successful, and if we have run out
	   of subject string, all has gone well */

	PS*.ps_matched := true;
    else

	/* check for duplicates of this state in the set of active states */

	pState := &PS*.ps_activeStates*[0];
	count := PS*.ps_stateCount;
	i := 0;
	while i ~= count and pState* ~= state do
	    pState := pState + sizeof(ulong);
	    i := i + 1;
	od;
	if i = count then

	    /* this state not already in the set - add it */

	    pState := pState + sizeof(ulong);
	    pState* := state;
	    i := i + 1;
	    PS*.ps_stateCount := i;
	fi;
    fi;
corp;

/*
 * internalMatch - match the compiled pattern against a subject string.
 *	Return 'true' if the subject matches the pattern.
 */

proc internalMatch(*PatternState_t ps; register *char subject;
		   ulong subjectLength)bool:
    register *[2]ulong comp;
    *char subjectPointer, subjectEnd;
    register ulong stateNum, state, compValue, temp;
    ulong stateCount;
    register char patternChar @ temp, subjectChar @ compValue;

    PS := ps;
    comp := PS*.ps_compiled;
    PS*.ps_stateCount := 0;	/* no active states yet */
    subjectEnd := subject + subjectLength * sizeof(char);
    PS*.ps_matched := false;

    /* for the first subject character, the only active state is that
       associated with the first pattern character */

    addState(1);

    /* if the pattern has top-level alternation, add the head of that
       alternation list to the set of active states */

    if comp*[0] ~= 0 then
	addState(comp*[0]);
    fi;

    /* loop until we have matched the entire subject string, or we have no
       active states */

    while

	/* each time around, we must first add in all of those states that
	   are the successors to the set of states we have now. In the
	   following loop, note that the value of PS*.ps_stateCount can
	   change inside the loop, so a Draco 'for' will not work. */

	stateNum := 1;
	while stateNum <= PS*.ps_stateCount do
	    state := PS*.ps_activeStates*[stateNum];
	    compValue := comp*[state];
	    case PS*.ps_pattern*[state - 1]
	    incase '\#':

		/* since '#' allows 0 or more, both the '#'ed item and
		   the thing after it are OK now */

		addState(state + 1);
	    fallthrough:
	    incase '%':

		/* if a match-nothing is OK now, then so is whatever comes
		   after it in the pattern */

		addState(compValue);
	    incase '(':
	    incase '|':

		/* part of an alternative chain - this alternative is OK */

		addState(state + 1);
		if compValue ~= 0 then

		    /* this is chaining down a list of alternatives */

		    addState(compValue);
		fi;
	    incase '~':

		/* have to check the notted pattern to see if it succeeds */

		addState(state + 1);
	    esac;
	    stateNum := stateNum + 1;
	od;

	subject < subjectEnd and PS*.ps_stateCount ~= 0
    do

	subjectChar := subject*;
	subject := subject + sizeof(char);
	if PS*.ps_ignoreCase and subjectChar >= 'a' and subjectChar <= 'z' then
	    subjectChar := subjectChar - 32;
	fi;

	/* go through the built-up set of active states, and see which ones
	   are successful, i.e. describe the current subject character. For
	   those that do, add the resulting state to the set of active states.
	   Note: we are re-using the active state vector when we do this. The
	   new set will be no larger than the old in this phase - it is in
	   the previous expansion phase, for the next time around, that it
	   might grow. */

	temp := PS*.ps_stateCount;
	PS*.ps_stateCount := 0;
	PS*.ps_matched := false;
	for stateNum from 1 upto temp do
	    state := PS*.ps_activeStates*[stateNum];
	    patternChar := PS*.ps_pattern*[state - 1];
	    case patternChar
	    incase '\#':
	    incase '|':
	    incase '%':
	    incase '(':
	    incase '~':
		/* just ignore these characters - their effect is handled
		   by the compiled vector */
		;
	    incase '?':
		/* match any character, so just unconditionally add the
		   successor state */
		addState(comp*[state]);
	    incase '\'':
		/* quote - get the quoted character, then handle it just
		   like we do for normal characters */
		patternChar := PS*.ps_pattern*[state];
	    fallthrough:
	    default:
		/* a normal character - if it is the same as the subject
		   character, then add the successor state, otherwise this
		   state is failing and we do nothing for it */
		if PS*.ps_ignoreCase and
		    patternChar >= 'a' and patternChar <= 'z'
		then
		    patternChar := patternChar - 32;
		fi;
		state := comp*[state];
writeln("character ", patternChar, " vs ", subjectChar, " state ", state);
		if PS*.ps_pattern*[state - 1] = '~' then
writeln("~ check");
		    /* was preceeded by a ~ */
		    if subjectChar ~= patternChar and subject ~= subjectEnd
		    then
			addState(state);
		    else
			addState(comp*[state]);
		    fi;
		else
		    if subjectChar = patternChar then
			addState(state);
		    fi;
		fi;
	    esac;
	od;
    od;

    if subject = subjectEnd then
	PS*.ps_matched
    else
	false
    fi
corp;
